import { __extends } from "tslib";
import { ASObject } from '../../avm2/nat/ASObject';
/**
 * Base export class for dispatching events
*
* @export class away.events.EventDispatcher
*
*/
var EventDispatcherBase = /** @class */ (function (_super) {
    __extends(EventDispatcherBase, _super);
    function EventDispatcherBase(target) {
        if (target === void 0) { target = null; }
        var _this = _super.call(this) || this;
        _this._listenerObjects = new Array();
        _this._queuedEvents = [];
        _this._t = target || _this;
        return _this;
    }
    EventDispatcherBase.prototype.toString = function () {
        return "";
    };
    /**
     * Add an event listener
     * @method addEventListener
     * @param {String} Name of event to add a listener for
     * @param {Function} Callback function
     */
    EventDispatcherBase.prototype.addEventListener = function (type, listener) {
        if (!this._listenerObjects)
            return;
        var l = this._listenerObjects[type];
        if (l === undefined)
            l = this._listenerObjects[type] = new ListenerObject();
        l.addEventListener(listener);
    };
    /**
     * Remove an event listener
     * @method removeEventListener
     * @param {String} Name of event to remove a listener for
     * @param {Function} Callback function
     */
    EventDispatcherBase.prototype.removeEventListener = function (type, listener) {
        if (!this._listenerObjects)
            return;
        var l = this._listenerObjects[type];
        if (l) {
            l.removeEventListener(listener);
            if (l.numListeners == 0)
                delete this._listenerObjects[type];
        }
    };
    EventDispatcherBase.prototype.getQueuedEvents = function () {
        if (!this._queuedEvents)
            return null;
        var returnEvents = this._queuedEvents.concat();
        this._queuedEvents.length = 0;
        return returnEvents;
    };
    /**
     * Dispatch an event
     * @method dispatchEvent
     * @param {Event} Event to dispatch
     */
    EventDispatcherBase.prototype.dispatchEvent = function (event) {
        if (!this._listenerObjects) {
            if (!this._queuedEvents)
                this._queuedEvents = [];
            this._queuedEvents.push(event);
            return;
        }
        var l = this._listenerObjects[event.type];
        if (l) {
            event.target = this._t;
            l.dispatchEvent(event);
        }
    };
    /**
     * check if an object has an event listener assigned to it
     * @method hasListener
     * @param {String} Name of event to remove a listener for
     * @param {Function} Callback function
     */
    EventDispatcherBase.prototype.hasEventListener = function (type, listener) {
        if (!this._listenerObjects)
            return;
        if (this._listenerObjects[type] === undefined)
            return false;
        if (listener != null)
            return this._listenerObjects[type].getEventListenerIndex(listener) !== -1;
        return this._listenerObjects[type].numListeners > 0;
    };
    return EventDispatcherBase;
}(ASObject));
export { EventDispatcherBase };
var ListenerObject = /** @class */ (function () {
    function ListenerObject() {
        this._index = 0;
        this._listeners = new Array();
        this.numListeners = 0;
    }
    ListenerObject.prototype.addEventListener = function (listener) {
        //check if listener already added
        if (this.getEventListenerIndex(listener) !== -1)
            return;
        this._listeners.push(listener);
        this.numListeners++;
    };
    ListenerObject.prototype.removeEventListener = function (listener) {
        //check if listener exists
        var index = this.getEventListenerIndex(listener);
        if (index === -1)
            return;
        this._listeners.splice(index, 1);
        //deals with removing a listener mid-way through dispatching listeners
        if (index <= this._index)
            this._index--;
        this.numListeners--;
    };
    ListenerObject.prototype.dispatchEvent = function (event) {
        var len = this.numListeners;
        for (this._index = 0; this._index < len && this._index < this.numListeners; this._index++)
            this._listeners[this._index].call(this._listeners[this._index], event);
    };
    /**
     * get Event Listener Index in array. Returns -1 if no listener is added
     * @method getEventListenerIndex
     * @param {String} Name of event to remove a listener for
     * @param {Function} Callback function
     */
    ListenerObject.prototype.getEventListenerIndex = function (listener) {
        for (var index = 0; index < this.numListeners; index++)
            if (listener == this._listeners[index])
                return index;
        return -1;
    };
    return ListenerObject;
}());
export { ListenerObject };
export default EventDispatcherBase;
