/**
 * Copyright 2014 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
import { __extends } from "tslib";
import { getAwayJSAdaptee, getAwayObjectOrTemplate, getAVM1Object, hasAwayJSAdaptee, initializeAVM1Object, wrapAVM1NativeClass, toTwipFloor, avm2AwayDepth, away2avmDepth } from "./AVM1Utils";
import { alCoerceString, alForEachProperty, alInstanceOf, alIsName, alNewObject, alToBoolean, alToInt32, alToNumber, alToString } from "../runtime";
import { release, assert, Debug, notImplemented, somewhatImplemented, warning } from "../../base/utilities/Debug";
import { isNullOrUndefined } from "../../base/utilities";
import { toAS3BitmapData } from "./AVM1BitmapData";
import { toAS3Matrix } from "./AVM1Matrix";
import { AVM1ArrayNative } from "../natives";
import { FlashNetScript_navigateToURL } from "../AVM1Helpers";
import { copyAS3PointTo, toAS3Point } from "./AVM1Point";
import { MovieClipProperties } from "../interpreter/MovieClipProperties";
import { MovieClip, TextField, Billboard, TextFormat, MouseManager, FrameScriptManager } from "@awayjs/scene";
import { AssetLibrary, Point, WaveAudio, URLRequest, Rectangle } from "@awayjs/core";
import { AVM1TextField } from "./AVM1TextField";
import { LineScaleMode } from "@awayjs/graphics";
import { AVM1SymbolBase } from "./AVM1SymbolBase";
import { AVM1Object } from "../runtime/AVM1Object";
import { AVM1Stage } from "./AVM1Stage";
import { AVM1LoaderHelper } from "./AVM1LoaderHelper";
import { EventsListForMC } from "./AVM1EventHandler";
import { PickGroup } from '@awayjs/view';
import { MethodMaterial } from "@awayjs/materials";
function convertAS3RectangeToBounds(as3Rectange, context) {
    var result = alNewObject(context);
    result.alPut('xMin', as3Rectange.left);
    result.alPut('yMin', as3Rectange.top);
    result.alPut('xMax', as3Rectange.right);
    result.alPut('yMax', as3Rectange.bottom);
    return result;
}
var AVM1MovieClip = /** @class */ (function (_super) {
    __extends(AVM1MovieClip, _super);
    function AVM1MovieClip() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.onLoaded = null;
        _this.executeConstructor = null;
        _this.unregisteredColors = {};
        _this.isDragging = false;
        _this.startDragPoint = new Point();
        _this.startDragMCPosition = new Point();
        return _this;
    }
    AVM1MovieClip.createAVM1Class = function (context) {
        return wrapAVM1NativeClass(context, true, AVM1MovieClip, [], ['$version#', '_alpha#', 'getAwayJSID', 'attachAudio', 'attachBitmap', 'attachMovie',
            'beginFill', 'beginBitmapFill', 'beginGradientFill', 'blendMode#',
            'cacheAsBitmap#', '_callFrame', 'clear', 'createEmptyMovieClip',
            'createTextField', '_currentframe#', 'curveTo', '_droptarget#',
            'duplicateMovieClip', 'enabled#', 'endFill', 'filters#', '_framesloaded#',
            '_focusrect#', 'forceSmoothing#', 'getBounds',
            'getBytesLoaded', 'getBytesTotal', 'getDepth', 'getInstanceAtDepth',
            'getNextHighestDepth', 'getRect', 'getSWFVersion', 'getTextSnapshot',
            'getURL', 'globalToLocal', 'gotoAndPlay', 'gotoAndStop', '_height#',
            '_highquality#', 'hitArea#', 'hitTest', 'lineGradientStyle', 'lineStyle',
            'lineTo', 'loadMovie', 'loadVariables', 'localToGlobal', '_lockroot#',
            'menu#', 'moveTo', '_name#', 'nextFrame', 'opaqueBackground#', '_parent#',
            'play', 'prevFrame', '_quality#', 'removeMovieClip', '_root#', '_rotation#',
            'scale9Grid#', 'scrollRect#', 'setMask', '_soundbuftime#', 'startDrag',
            'stop', 'stopDrag', 'swapDepths', 'tabChildren#', 'tabEnabled#', 'tabIndex#',
            '_target#', '_totalframes#', 'trackAsMenu#', 'toString',
            'unloadMovie', '_url#', 'useHandCursor#', '_visible#', '_width#',
            '_x#', '_xmouse#', '_xscale#', '_y#', '_ymouse#', '_yscale#']);
    };
    AVM1MovieClip.prototype._updateMouseEnabled = function (event, enabled) {
        if (!this.adaptee.isAVMScene) {
            if (event.isMouse) {
                if (enabled) {
                    this._mouseListenerCount++;
                    this.adaptee.mouseEnabled = true;
                    this.adaptee.mouseChildren = false;
                    if (event.isButton) {
                        this._mouseButtonListenerCount++;
                        this.adaptee.buttonMode = true;
                    }
                }
                else {
                    this._mouseListenerCount--;
                    if (this._mouseListenerCount <= 0) {
                        this._mouseListenerCount = 0;
                        this.adaptee.mouseEnabled = false;
                        this.adaptee.mouseChildren = true;
                    }
                    if (event.isButton) {
                        this._mouseButtonListenerCount--;
                        if (this._mouseButtonListenerCount <= 0) {
                            this._mouseButtonListenerCount = 0;
                            this.adaptee.buttonMode = false;
                        }
                    }
                }
            }
        }
        if (!this.enabled) {
            this.adaptee.mouseChildren = false;
        }
    };
    AVM1MovieClip.prototype.clone = function () {
        return getAVM1Object(this.adaptee.clone(), this._avm1Context);
    };
    AVM1MovieClip.prototype.attachCustomConstructor = function () {
        if (this.adaptee) {
            var symbolClass = this.adaptee.avm1Symbol;
            if (symbolClass) {
                this.alPut("__proto__", symbolClass._ownProperties.prototype.value);
                var myThis = this;
                this.executeConstructor = function () {
                    symbolClass.alCall(myThis);
                };
            }
        }
    };
    // this is used for ordering AVM1 Framescripts into correct order
    AVM1MovieClip.prototype.compareAVM1FrameScripts = function (a, b) {
        if (!a.precedence) {
            return !b.precedence ? 0 : -1;
        }
        else if (!b.precedence) {
            return 1;
        }
        var i = 0;
        while (i < a.precedence.length && i < b.precedence.length && a.precedence[i] === b.precedence[i]) {
            i++;
        }
        if (i >= a.precedence.length) {
            return a.precedence.length === b.precedence.length ? 0 : -1;
        }
        else {
            return i >= b.precedence.length ? 1 : a.precedence[i] - b.precedence[i];
        }
    };
    AVM1MovieClip.prototype.executeScript = function (actionsBlocks) {
        AVM1MovieClip.currentMCAssetNameSpace = this.adaptee.assetNamespace;
        AVM1TextField.syncQueedTextfields();
        var name = this.adaptee.name.replace(/[^\w]/g, '');
        if (!actionsBlocks) {
            window.alert("actionsBlocks is empty, can not execute framescript" + name + this.adaptee.currentFrameIndex);
            return;
        }
        var unsortedScripts = [];
        for (var k = 0; k < actionsBlocks.length; k++) {
            var actionsBlock = actionsBlocks[k];
            var script = function (actionsData) {
                this._avm1Context.executeActions(actionsData, this);
            }.bind(this, actionsBlock.data);
            // this uses parents of current scenegraph. so its not possible to easy preset in parser 
            script.precedence = this.adaptee.getScriptPrecedence().concat(actionsBlock.precedence);
            script.context = this.adaptee;
            unsortedScripts.push(script);
        }
        if (unsortedScripts.length) {
            unsortedScripts.sort(this.compareAVM1FrameScripts);
            var sortedFrameScripts = unsortedScripts;
            for (var k = 0; k < sortedFrameScripts.length; k++) {
                var myScript = sortedFrameScripts[k];
                var mc = myScript.context;
                myScript.call(mc);
            }
        }
    };
    AVM1MovieClip.prototype.addScript = function (source, frameIdx) {
        var allscripts = source;
        var translatedScripts = [];
        for (var i = 0; i < allscripts.length; i++) {
            var scripts = allscripts[i];
            for (var s = 0; s < scripts.length; s++) {
                var script = scripts[s];
                var mcName = this.adaptee.name;
                if (typeof mcName != "string") {
                    mcName = mcName.toString();
                }
                script.data = this._avm1Context.actionsDataFactory.createActionsData(script.actionsData, 'script_' + mcName.replace(/[^\w]/g, '') + "_" + this.adaptee.id + '_frame_' + frameIdx + '_idx_' + i);
                translatedScripts[translatedScripts.length] = script;
            }
        }
        return translatedScripts;
    };
    AVM1MovieClip.prototype.stopAllSounds = function () {
        var allProps = this.alGetKeys();
        for (var i = 0; i < allProps.length; i++) {
            var desc = this.alGetProperty(allProps[i]);
            var val = desc ? desc.value : null;
            if (val && val._sound && val._sound.isAsset && val._sound.isAsset(WaveAudio)) {
                val.stop();
            }
        }
        var child;
        i = this.adaptee.numChildren;
        while (i > 0) {
            i--;
            child = this.adaptee.getChildAt(i);
            if (child.isAsset(MovieClip) && child.adapter != child) {
                child.adapter.freeFromScript();
            }
        }
    };
    // called from adaptee whenever the mc gets reset
    AVM1MovieClip.prototype.freeFromScript = function () {
        //this.stopAllSounds();
        this.hasSwappedDepth = false;
        _super.prototype.freeFromScript.call(this);
        this._mouseButtonListenerCount = 0;
    };
    // called from adaptee whenever the mc gets added to a parent
    AVM1MovieClip.prototype.initAdapter = function () {
        this._dropTarget = "";
        this._mouseButtonListenerCount = 0;
        AVM1MovieClip.currentMCAssetNameSpace = this.adaptee.assetNamespace;
        // execute the init-actionscript that is stored on timeline
        for (var key in this.adaptee.timeline.avm1InitActions)
            this.executeScript(this.addScript(this.adaptee.timeline.avm1InitActions[key], ("initActionsData" + key)));
        // execute the init-actionscript that is stored on timeline
        if (this.initEvents) {
            initializeAVM1Object(this.adaptee, this._avm1Context, this.initEvents);
        }
        this.attachCustomConstructor();
        this.initialDepth = this.adaptee._depthID;
        if (this.adaptee.onLoaded || this.executeConstructor) {
            FrameScriptManager.add_loaded_action_to_queue(this.adaptee);
        }
    };
    AVM1MovieClip.prototype.registerScriptObject = function (child, fromTimeline) {
        // 	whenever multiple childs get registered for the same name, the child with the lowest depth wins
        //	this is true for objects added via timeline, attachMovie, duplicateMovieClip and CreateEmptyMovieClip and also when renaming objects
        if (fromTimeline === void 0) { fromTimeline = true; }
        // 	if a avm1 variable for a registered child-name was defined via script, it always wins over the child registration. 
        //	for example, after setting "testMC='something'" in script, trying to get "testMC" will return "something", 
        //	and childs with the name "testMC" will no longer be accessible.
        //	this is true for all datatypes. once a variable is set by script, the value set by script will be returned.
        //	also true if the we do something like "testMC=null"
        //	if a avm1 script creates a variable as reference to a registered child, it behaves like this:
        //	the var is not updated if the child registered for the name changes
        //	if we do something like this:
        //			indirectRef = testMC;
        //			this.attachMovie("somethingFromLib", "testMC", -16380)
        //			testMC._x=100;
        //			testMC._y=100;
        //			indirectRef._x=100;
        //			indirectRef._y=100;
        //	this code will move both objects, because the attachMovie changes the object that is available under the name "testMC",
        //	but the indirectRef still holds the reference to the original object.
        //	however, if we remove the original child via removeMovieCLip or if it is removed via timeline, 
        //	the variable will return a reference to the current object that might now be registered 
        //	under the same name as the object that the variable was created for. 
        // 	if a movieclip was tinted by a color object, and than another mc gets registered for the name, the tinting is applied to the new mc
        if (child.adapter != child)
            child.adapter.setEnabled(true);
        var name = child.name;
        if (name) {
            // 	in AVM1 FP<8, all names ignore case, so we can just convert everything to lower case
            //	for FP>=8, only method names can be handled this way, object names need to keep their case
            name = name.toLowerCase();
            var hasVar = this.alHasOwnProperty(name);
            if (fromTimeline && hasVar) {
                // there exists a avm1 var for this name. Object registration should fail
                return;
            }
            //	only register object if:
            //			- no object is registered for this name,
            //			- a already registered object has no valid parent
            //			- the already registered object has a higher depthID than the new object
            if (!this._childrenByName[name]
                || (this._childrenByName[name].adaptee && this._childrenByName[name].adaptee.parent == null)
                || (this._childrenByName[name].adaptee && this._childrenByName[name].adaptee._depthID > child._depthID)) {
                if (this._childrenByName[name] && this._childrenByName[name].avmColor) {
                    // if a object already is registered for this name, and we are going to replace it, 
                    // we need to check if it is tinted by a color-class, and if so change the target of the color class to the new object		
                    this.unregisteredColors[name] = this._childrenByName[name].avmColor;
                }
                //register new object
                this._childrenByName[name] = getAVM1Object(child, this.context);
                if (!fromTimeline) {
                    this.alPut(name, this._childrenByName[name]);
                }
                // if a object was previous unregistered for this name, we need to check if it was tinted by a color 
                // and if so, we need to apply the tint to the new object
                if (this.unregisteredColors[name]) {
                    this.unregisteredColors[name].changeTarget(child.adapter);
                    this.unregisteredColors[name] = null;
                }
            }
            /* old stuff:
            if(force || !this._childrenByName[child.name] || (this._childrenByName[child.name].adaptee && this._childrenByName[child.name].adaptee.parent==null) ||
                this._childrenByName[child.name].adaptee._depthID>child._depthID){
                
                
                // only replace value if no property exists yet, or if a existing prop was a Away-Object
                var existingValue=this.alGet(child.name);
                if(existingValue && existingValue.adaptee){
                    this.alPut(child.name, child.adapter);
                    if(this.scriptRefsToChilds[child.name]){
                        // update indirectReferences to this prop
                        this.scriptRefsToChilds[child.name].obj.alPut(this.scriptRefsToChilds[child.name].name, child.adapter);
                    }
                }
                else if(existingValue){
                }
                else{
                    this.alPut(child.name, child.adapter);
                    if(this.scriptRefsToChilds[child.name]){
                        // update indirectReferences to this prop
                        this.scriptRefsToChilds[child.name].obj.alPut(this.scriptRefsToChilds[child.name].name, child.adapter);
                    }
                }
                if(this.unregisteredColors[child.name]){
                    this.unregisteredColors[child.name].changeTarget(child.adapter);
                    this.unregisteredColors[child.name]=null;
                }
                // always register the child by name on the parent, no matter what avm1prop existed previously
                this._childrenByName[child.name]=child._adapter;
            }
            */
            /*
            else{
                
                this._unregisteredChilds[child.name]=child;
            }
            */
        }
    };
    AVM1MovieClip.prototype.unregisterScriptObject = function (child) {
        if (child && child.adapter != child)
            child.adapter.alPut("onEnterFrame", null);
        var name = child.name;
        if (name) {
            // 	in AVM1 FP<8, all names ignore case, so we can just convert everything to lower case
            //	for FP>=8, only method names can be handled this way, object names need to keep their case
            name = name.toLowerCase();
            if (this._childrenByName[name] && this._childrenByName[name].adaptee.id == child.id) {
                if (this._childrenByName[name] && this._childrenByName[name].avmColor) {
                    this.unregisteredColors[name] = this._childrenByName[name].avmColor;
                }
                // 	check if there is another child with the same name on the movieclip
                //	and if so, register it instead
                //	if multiple childs match the name, get the one with highest depth
                //	attention: at this point the child that we unregister right now is still child of the mc
                var allChilds = this.adaptee._children;
                var allChildsLen = allChilds.length;
                var tmpChild = null;
                var newChild = null;
                for (var i = 0; i < allChildsLen; i++) {
                    tmpChild = allChilds[i];
                    if (tmpChild != child && tmpChild.name && tmpChild.name.toLowerCase() == name) {
                        if (!newChild || newChild._depthID > tmpChild._depthID) {
                            newChild = tmpChild;
                        }
                    }
                }
                //	if we have a new child to register, we register it
                //	if not, we delete the registration for this name
                if (newChild) {
                    this._childrenByName[name] = getAVM1Object(newChild, this.context);
                    if (this.unregisteredColors[name]) {
                        this.unregisteredColors[name].changeTarget(newChild.adapter);
                        this.unregisteredColors[name] = null;
                    }
                }
                else {
                    delete this._childrenByName[name];
                }
            }
        }
    };
    AVM1MovieClip.prototype.getLatestObjectForName = function (name) {
        var hasVar = this.alHasOwnProperty(name);
        if (hasVar) {
            // there exists a avm1 var for this name. Object registration should fail
            return;
        }
        var allChilds = this.adaptee._children;
        var allChildsLen = allChilds.length;
        var tmpChild = null;
        var newChild = null;
        for (var i = 0; i < allChildsLen; i++) {
            tmpChild = allChilds[i];
            if (tmpChild.name && tmpChild.name.toLowerCase() == name) {
                if (!newChild || newChild._depthID > tmpChild._depthID) {
                    newChild = tmpChild;
                }
            }
        }
        //	if we have a new child to register, we register it
        //	if not, we delete the registration for this name
        if (newChild) {
            this._childrenByName[name] = getAVM1Object(newChild, this.context);
            if (this.unregisteredColors[name]) {
                this.unregisteredColors[name].changeTarget(newChild.adapter);
                this.unregisteredColors[name] = null;
            }
        }
    };
    Object.defineProperty(AVM1MovieClip.prototype, "graphics", {
        get: function () {
            return this.adaptee.graphics;
        },
        enumerable: true,
        configurable: true
    });
    AVM1MovieClip.prototype.initAVM1SymbolInstance = function (context, awayObject) {
        var _this = this;
        this._childrenByName = Object.create(null);
        _super.prototype.initAVM1SymbolInstance.call(this, context, awayObject);
        this.dragListenerDelegate = function (event) { return _this.dragListener(event); };
        this.stopDragDelegate = function (event) { return _this.stopDrag(event); };
        this.dynamicallyCreated = false;
        this.adaptee = awayObject;
        this._initEventsHandlers();
    };
    AVM1MovieClip.prototype._lookupChildByName = function (name) {
        release || assert(alIsName(this.context, name));
        return this._childrenByName[name];
    };
    AVM1MovieClip.prototype._lookupChildInAS3Object = function (name) {
        var lookupOptions = 2 /* INCLUDE_NON_INITIALIZED */;
        if (!this.context.isPropertyCaseSensitive) {
            lookupOptions |= 1 /* IGNORE_CASE */;
        }
        //80pro todo lookupOptions
        var as3Child = this.adaptee.getChildByName(name); //, lookupOptions);
        return getAVM1Object(as3Child, this.context);
    };
    Object.defineProperty(AVM1MovieClip.prototype, "__targetPath", {
        get: function () {
            //return "";
            var target = this.get_target();
            var as3Root = this.adaptee.root;
            release || Debug.assert(as3Root);
            var level = this.context.levelsContainer._getLevelForRoot(as3Root);
            release || Debug.assert(level >= 0);
            var prefix = '_level' + level;
            return target != '/' ? prefix + target.replace(/\//g, '.') : prefix;
        },
        enumerable: true,
        configurable: true
    });
    AVM1MovieClip.prototype.getAwayJSID = function () {
        return this.adaptee.id;
    };
    AVM1MovieClip.prototype.attachAudio = function (id) {
        if (isNullOrUndefined(id)) {
            return; // ignoring all undefined objects, probably nothing to attach
        }
        if (id === false) {
            return; // TODO stop playing all attached audio source (when implemented).
        }
        // TODO implement NetStream and Microphone objects to make this work.
        notImplemented('AVM1MovieClip.attachAudio');
    };
    AVM1MovieClip.prototype.attachBitmap = function (bmp, depth, pixelSnapping, smoothing) {
        if (pixelSnapping === void 0) { pixelSnapping = 'auto'; }
        if (smoothing === void 0) { smoothing = false; }
        pixelSnapping = alCoerceString(this.context, pixelSnapping);
        smoothing = alToBoolean(this.context, smoothing);
        var awayBitmapImage2D = bmp.as3BitmapData;
        awayBitmapImage2D.transparent = true;
        var billboardMaterial = new MethodMaterial(awayBitmapImage2D);
        billboardMaterial.alphaBlending = true;
        billboardMaterial.useColorTransform = true;
        var billboard = new Billboard(billboardMaterial, pixelSnapping, smoothing); // this.context.sec.flash.display.Bitmap.axClass.axConstruct([as3BitmapData, pixelSnapping, smoothing]);
        this._insertChildAtDepth(billboard, depth);
    };
    AVM1MovieClip.prototype._constructMovieClipSymbol = function (symbolId, name) {
        symbolId = alToString(this.context, symbolId);
        name = alToString(this.context, name);
        var symbol = AssetLibrary.getAsset(symbolId, this.adaptee.assetNamespace);
        if (!symbol) {
            return undefined;
        }
        /*
        var props: SpriteSymbol = Object.create(symbol.symbolProps);
        props.avm1Name = name;
        */
        var mc;
        mc = symbol.adaptee.clone(); //constructClassFromSymbol(props, this.context.sec.flash.display.MovieClip.axClass);
        mc.name = name;
        getAVM1Object(mc, this._avm1Context);
        return mc;
    };
    AVM1MovieClip.prototype.get$version = function () {
        return "";
        //return Capabilities.version;
    };
    AVM1MovieClip.prototype.rgbaToArgb = function (float32Color) {
        var r = (float32Color & 0xff000000) >>> 24;
        var g = (float32Color & 0xff0000) >>> 16;
        var b = (float32Color & 0xff00) >>> 8;
        var a = float32Color & 0xff;
        return (a << 24) | (r << 16) | (g << 8) | b;
    };
    AVM1MovieClip.prototype.attachMovie = function (symbolId, name, depth, initObject) {
        if (!this._constructMovieClipSymbol)
            return;
        var mc = this._constructMovieClipSymbol(symbolId, name);
        if (!mc) {
            return undefined;
        }
        depth = alToNumber(this.context, depth);
        var oldAVMMC = this._childrenByName[name.toLowerCase()];
        //console.log("attachMovie", name, avm2AwayDepth(depth));
        var avmMc = this._insertChildAtDepth(mc, avm2AwayDepth(depth));
        if (initObject) {
            avmMc._init(initObject);
        }
        if (oldAVMMC && oldAVMMC.avmColor) {
            oldAVMMC.avmColor.changeTarget(avmMc);
        }
        if (mc.timeline && mc.timeline.isButton) {
            mc.addButtonListeners();
        }
        avmMc.dynamicallyCreated = true;
        this.registerScriptObject(mc, false);
        return avmMc;
    };
    AVM1MovieClip.prototype.beginFill = function (color, alpha) {
        color = alToInt32(this.context, color);
        if (typeof alpha == "undefined") {
            if (arguments.length == 2) {
                // alpha was set with "undefined" variable - should be 0
                alpha = 0;
            }
            else if (arguments.length <= 1) {
                // alpha was not set at all - should be 100
                alpha = 100;
            }
        }
        alpha = alToNumber(this.context, alpha);
        this.graphics.beginFill(color, alpha / 100.0);
    };
    AVM1MovieClip.prototype.beginBitmapFill = function (bmp, matrix, repeat, smoothing) {
        if (matrix === void 0) { matrix = null; }
        if (repeat === void 0) { repeat = false; }
        if (smoothing === void 0) { smoothing = false; }
        if (!alInstanceOf(this.context, bmp, this.context.globals.BitmapData)) {
            return; // skipping operation if first parameter is not a BitmapData.
        }
        var bmpNative = toAS3BitmapData(bmp);
        var matrixNative = isNullOrUndefined(matrix) ? null : toAS3Matrix(matrix);
        repeat = alToBoolean(this.context, repeat);
        smoothing = alToBoolean(this.context, smoothing);
        notImplemented('AVM1MovieClip.beginBitmapFill');
        //this.graphics.beginBitmapFill(bmpNative.adaptee, matrixNative, repeat, smoothing);
    };
    AVM1MovieClip.prototype.beginGradientFill = function (fillType, colors, alphas, ratios, matrix, spreadMethod, interpolationMethod, focalPointRatio) {
        var _this = this;
        if (spreadMethod === void 0) { spreadMethod = 'pad'; }
        if (interpolationMethod === void 0) { interpolationMethod = 'rgb'; }
        if (focalPointRatio === void 0) { focalPointRatio = 0.0; }
        var context = this.context, sec = context.sec;
        // fillType = alToString(this.context, fillType);
        var colorsNative = sec.createArray(AVM1ArrayNative.mapToJSArray(colors, function (item) { return alToInt32(_this.context, item); }));
        var alphasNative = sec.createArray(AVM1ArrayNative.mapToJSArray(alphas, function (item) { return alToNumber(_this.context, item) / 100.0; }));
        var ratiosNative = sec.createArray(AVM1ArrayNative.mapToJSArray(ratios, function (item) { return alToNumber(_this.context, item); }));
        var matrixNative = null;
        if (isNullOrUndefined(matrix)) {
            somewhatImplemented('AVM1MovieClip.beginGradientFill');
        }
        spreadMethod = alToString(this.context, spreadMethod);
        interpolationMethod = alToString(this.context, interpolationMethod);
        focalPointRatio = alToNumber(this.context, focalPointRatio);
        this.graphics.beginGradientFill(fillType, colorsNative, alphasNative, ratiosNative, matrixNative, spreadMethod, interpolationMethod, focalPointRatio);
    };
    AVM1MovieClip.prototype._callFrame = function (frame) {
        var script;
        if (typeof frame === "string")
            script = this.adaptee.timeline.getScriptForLabel(this.adaptee, frame);
        else if (typeof frame === "number")
            script = this.adaptee.timeline.get_script_for_frame(this.adaptee, frame - 1);
        if (script)
            this.executeScript(script);
    };
    AVM1MovieClip.prototype.clear = function () {
        this.graphics.clear();
    };
    AVM1MovieClip.prototype._insertChildAtDepth = function (mc, depth) {
        this.adaptee.addChildAtDepth(mc, depth);
        return getAVM1Object(mc, this._avm1Context);
    };
    AVM1MovieClip.prototype._updateChildName = function (child, oldName, newName) {
        oldName && this._removeChildName(child, oldName);
        newName && this._addChildName(child, newName);
    };
    AVM1MovieClip.prototype._removeChildName = function (child, name) {
        release || assert(name);
        if (!this.context.isPropertyCaseSensitive) {
            name = name.toLowerCase();
        }
        if (!this._childrenByName || !this._childrenByName[name])
            return;
        if (this._childrenByName[name] !== child) {
            return;
        }
        var newChildForName = this._lookupChildInAS3Object(name);
        if (newChildForName) {
            this._childrenByName[name] = newChildForName;
        }
        else {
            delete this._childrenByName[name];
        }
    };
    AVM1MovieClip.prototype._addChildName = function (child, name) {
        release || assert(name);
        if (!this.context.isPropertyCaseSensitive) {
            name = name.toLowerCase();
        }
        release || assert(this._childrenByName[name] !== child);
        var currentChild = this._childrenByName[name];
        if (!currentChild || currentChild.getDepth() > child.getDepth()) {
            this._childrenByName[name] = child;
        }
    };
    AVM1MovieClip.prototype.createEmptyMovieClip = function (name, depth) {
        // 	creates a new empty movieclip.
        //	if a mc already exists for the name, but at different depth, it will create a new movieclip, and also keep the existing alive
        //	if a mc already exists at same depth, it will replace the existing movieclip with the new one
        name = alToString(this.context, name);
        var mc = new MovieClip();
        mc.name = name;
        mc.assetNamespace = this.adaptee.assetNamespace;
        getAVM1Object(mc, this._avm1Context);
        //console.log("createEmptyMovieClip", name, avm2AwayDepth(depth));
        var avmMC = this._insertChildAtDepth(mc, avm2AwayDepth(depth));
        avmMC.dynamicallyCreated = true;
        this.registerScriptObject(mc, false);
        // dynamicallyCreated needs to be set after adding child, otherwise it gets reset
        return avmMC;
    };
    AVM1MovieClip.prototype.createTextField = function (name, depth, x, y, width, height) {
        name = alToString(this.context, name);
        var text = new TextField();
        text.name = name;
        text.textFormat = new TextFormat();
        getAVM1Object(text, this._avm1Context);
        var myTF = this._insertChildAtDepth(text, avm2AwayDepth(depth));
        this.registerScriptObject(text, false);
        text.x = x;
        text.y = y;
        text.width = width;
        text.height = height;
        myTF.dynamicallyCreated = true;
        return myTF;
    };
    AVM1MovieClip.prototype.get_currentframe = function () {
        return this.adaptee.currentFrameIndex + 1;
    };
    AVM1MovieClip.prototype.curveTo = function (controlX, controlY, anchorX, anchorY) {
        controlX = alToNumber(this.context, controlX);
        controlY = alToNumber(this.context, controlY);
        anchorX = alToNumber(this.context, anchorX);
        anchorY = alToNumber(this.context, anchorY);
        this.graphics.curveTo(controlX, controlY, anchorX, anchorY);
    };
    AVM1MovieClip.prototype.setDropTarget = function (dropTarget) {
        if (dropTarget) {
            //console.log((<AVMRaycastPicker>AVM1Stage.stage.view.mousePicker).getDropTarget().name);
            var names = [];
            while (dropTarget) {
                if (dropTarget.isAVMScene) {
                    dropTarget = null;
                }
                else {
                    if (dropTarget.name != null)
                        names.push(dropTarget.name);
                    dropTarget = dropTarget.parent;
                }
            }
            var i = names.length;
            var mc_path = "";
            while (i > 0) {
                i--;
                if (names[i] != "")
                    mc_path += "/";
                mc_path += names[i];
            }
            //console.log(mc_path);
            this._dropTarget = mc_path;
            return;
        }
        this._dropTarget = "";
    };
    AVM1MovieClip.prototype.get_droptarget = function () {
        return this._dropTarget;
    };
    AVM1MovieClip.prototype.duplicateMovieClip = function (name, depth, initObject) {
        name = alToString(this.context, name);
        if (name == this.adaptee.name) {
            return this;
        }
        var parent = this.get_parent();
        if (!parent) {
            warning("AVM1MovieClip.duplicateMovieClip could not get parent");
            parent = this.context.resolveTarget(null);
        }
        var mc;
        if (this.adaptee._symbol) {
            notImplemented('AVM1MovieClip.duplicateMovieClip from symbol');
            //mc = constructClassFromSymbol(nativeAS3Object._symbol, nativeAS3Object.axClass);
        }
        else {
            mc = this.clone().adaptee; //new this.context.sec.flash.display.MovieClip();
        }
        mc.name = name;
        mc.adapter.placeObjectTag = this.placeObjectTag;
        mc.adapter.initEvents = this.initEvents;
        var avmMc = parent._insertChildAtDepth(mc, avm2AwayDepth(depth));
        // dynamicallyCreated needs to be set after adding child, otherwise it gets reset
        avmMc.dynamicallyCreated = true;
        avmMc._avm1Context = this._avm1Context;
        parent.registerScriptObject(mc, false);
        var new_matrix = mc.transform.matrix3D;
        var originalMatrix = this.adaptee.transform.matrix3D._rawData;
        new_matrix._rawData[0] = originalMatrix[0];
        new_matrix._rawData[1] = originalMatrix[1];
        new_matrix._rawData[4] = originalMatrix[4];
        new_matrix._rawData[5] = originalMatrix[5];
        new_matrix._rawData[12] = originalMatrix[12];
        new_matrix._rawData[13] = originalMatrix[13];
        mc.transform.invalidateComponents();
        mc.alpha = this.adaptee.alpha;
        mc.blendMode = this.adaptee.blendMode;
        mc.cacheAsBitmap = this.adaptee.cacheAsBitmap;
        if (initObject) {
            avmMc._init(initObject);
        }
        return avmMc;
    };
    AVM1MovieClip.prototype.endFill = function () {
        this.graphics.endFill();
    };
    AVM1MovieClip.prototype.getForceSmoothing = function () {
        notImplemented('AVM1MovieClip.getForceSmoothing');
        return false;
    };
    AVM1MovieClip.prototype.setForceSmoothing = function (value) {
        value = alToBoolean(this.context, value);
        notImplemented('AVM1MovieClip.setForceSmoothing');
    };
    AVM1MovieClip.prototype.get_framesloaded = function () {
        notImplemented('AVM1MovieClip.get_framesloaded');
        return 0; //this.adaptee.framesLoaded;
    };
    AVM1MovieClip.prototype.getBounds = function (bounds) {
        var obj = getAwayJSAdaptee(bounds);
        if (!obj) {
            return undefined;
        }
        return convertAS3RectangeToBounds(PickGroup.getInstance(this.context.globals.Stage._awayAVMStage.scene.renderer.view).getBoundsPicker(this.adaptee.partition).getBoxBounds(obj, true), this.context);
    };
    AVM1MovieClip.prototype.getBytesLoaded = function () {
        //var loaderInfo = this.adaptee.loaderInfo;
        return this.adaptee.currentFrameIndex >= 0 ? 100 : -1; //loaderInfo.bytesLoaded;
    };
    AVM1MovieClip.prototype.getBytesTotal = function () {
        //var loaderInfo = this.adaptee.loaderInfo;
        return 100; //loaderInfo.bytesTotal;
    };
    AVM1MovieClip.prototype.getInstanceAtDepth = function (depth) {
        //Debug.notImplemented('AVM1MovieClip.getInstanceAtDepth');
        // 80pro: why does this always return movieclip ?
        // todo: check if in as3 this could be a textfield
        var child = this.adaptee.getChildAtDepth(avm2AwayDepth(depth));
        if (!child) {
            return null;
        }
        if (child.isAsset(Billboard)) {
            return this;
        }
        else if (child.isAsset(MovieClip)) {
            return this;
        }
        return getAVM1Object(child, this.context);
        /*
        var symbolDepth = alCoerceNumber(this.context, depth) + DEPTH_OFFSET;
        var nativeObject = this.adaptee;
        var lookupChildOptions = LookupChildOptions.INCLUDE_NON_INITIALIZED;
        for (var i = 0, numChildren = nativeObject.numChildren; i < numChildren; i++) {
            var child = nativeObject._lookupChildByIndex(i, lookupChildOptions);
            // child is null if it hasn't been constructed yet. This can happen in InitActionBlocks.
            if (child && child._depth === symbolDepth) {
                // Somewhat absurdly, this method returns the mc if a bitmap is at the given depth.
                if (this.context.sec.flash.display.Bitmap.axIsType(child)) {
                    return this;
                }
                return <AVM1MovieClip>getAVM1Object(child, this.context);
            }
        }
        return undefined;
        */
    };
    AVM1MovieClip.prototype.getNextHighestDepth = function () {
        return away2avmDepth(this.adaptee.getNextHighestDepth());
    };
    AVM1MovieClip.prototype.getRect = function (bounds) {
        var obj = getAwayJSAdaptee(bounds);
        if (!obj) {
            return undefined;
        }
        return convertAS3RectangeToBounds(PickGroup.getInstance(this.context.globals.Stage._awayAVMStage.scene.renderer.view).getBoundsPicker(this.adaptee.partition).getBoxBounds(obj), this.context);
    };
    AVM1MovieClip.prototype.getSWFVersion = function () {
        var loaderInfo = this.adaptee.loaderInfo;
        return 0; //loaderInfo.swfVersion;
    };
    AVM1MovieClip.prototype.getTextSnapshot = function () {
        notImplemented('AVM1MovieClip.getTextSnapshot');
    };
    AVM1MovieClip.prototype.getURL = function (url, window, method) {
        var request = new URLRequest(url);
        if (method) {
            request.method = method;
        }
        FlashNetScript_navigateToURL(request, window);
    };
    AVM1MovieClip.prototype.globalToLocal = function (pt) {
        if (!pt)
            return;
        var tmp = toAS3Point(pt);
        this.adaptee.transform.globalToLocal(tmp, tmp);
        copyAS3PointTo(tmp, pt);
    };
    AVM1MovieClip.prototype.gotoAndPlay = function (frame) {
        if (this.protoTypeChanged || frame == null)
            return;
        if (Array.isArray(frame)) {
            if (frame.length == 0)
                return;
            frame = frame[0];
        }
        if (frame instanceof AVM1ArrayNative) {
            if (!frame.value || frame.value.length == 0)
                return;
            frame = frame.value[0];
        }
        if (typeof frame === "string") {
            if (this.adaptee.timeline._labels[frame.toLowerCase()] == null) {
                frame = parseInt(frame);
                if (!isNaN(frame)) {
                    this.adaptee.currentFrameIndex = frame - 1;
                    this.adaptee.play();
                }
                return;
            }
        }
        if (typeof frame === "number" && frame <= 0)
            return;
        this.adaptee.play();
        this._gotoFrame(frame);
    };
    AVM1MovieClip.prototype.gotoAndStop = function (frame) {
        if (this.protoTypeChanged || frame == null)
            return;
        if (Array.isArray(frame)) {
            if (frame.length == 0)
                return;
            frame = frame[0];
        }
        if (frame instanceof AVM1ArrayNative) {
            if (!frame.value || frame.value.length == 0)
                return;
            frame = frame.value[0];
        }
        if (typeof frame === "number" && frame <= 0)
            return;
        this.adaptee.stop();
        this._gotoFrame(frame);
    };
    AVM1MovieClip.prototype._gotoFrame = function (frame) {
        if (typeof frame === "number") {
            if (frame % 1 !== 0) {
                frame = frame.toString();
            }
        }
        if (typeof frame === "string") {
            if (this.adaptee.timeline._labels[frame.toLowerCase()] == null) {
                frame = parseInt(frame);
                if (!isNaN(frame)) {
                    this.adaptee.currentFrameIndex = frame - 1;
                }
                return;
            }
            this.adaptee.jumpToLabel(frame.toLowerCase());
        }
        else {
            this.adaptee.currentFrameIndex = frame - 1;
        }
    };
    AVM1MovieClip.prototype.getHitArea = function () {
        return this._hitArea;
    };
    AVM1MovieClip.prototype.setHitArea = function (value) {
        // The hitArea getter always returns exactly the value set here, so we have to store that.
        this._hitArea = value;
        var obj = value ? getAwayJSAdaptee(value) : null;
        if (obj && !obj.isAsset(MovieClip))
            obj = null;
        // 	MA_GBR_0700AAx0100 is the first lesson encountered that makes use of hitArea
        // 	if the hitArea is set, the mouse-interactions on the ducks stop working
        //	this.adaptee.hitArea=obj;
    };
    AVM1MovieClip.prototype.hitTest = function (x, y, shapeFlag) {
        if (arguments.length <= 1) {
            // Alternative method signature: hitTest(target: AVM1Object): boolean
            var target = arguments[0];
            if (typeof target === "string") {
                target = this.context.resolveTarget(target);
            }
            if (isNullOrUndefined(target) || !hasAwayJSAdaptee(target)) {
                return false; // target is undefined or not a AVM1 display object, returning false.
            }
            return PickGroup.getInstance(this.context.globals.Stage._awayAVMStage.scene.renderer.view).getBoundsPicker(this.adaptee.partition).hitTestObject(PickGroup.getInstance(this.context.globals.Stage._awayAVMStage.scene.renderer.view).getBoundsPicker(getAwayJSAdaptee(target).partition));
        }
        x = alToNumber(this.context, x);
        y = alToNumber(this.context, y);
        x += this.get_root().get_x();
        y += this.get_root().get_y();
        shapeFlag = alToBoolean(this.context, shapeFlag);
        return PickGroup.getInstance(this.context.globals.Stage._awayAVMStage.scene.renderer.view).getBoundsPicker(this.adaptee.partition).hitTestPoint(x, y, shapeFlag);
    };
    AVM1MovieClip.prototype.lineGradientStyle = function (fillType, colors, alphas, ratios, matrix, spreadMethod, interpolationMethod, focalPointRatio) {
        var _this = this;
        if (spreadMethod === void 0) { spreadMethod = 'pad'; }
        if (interpolationMethod === void 0) { interpolationMethod = 'rgb'; }
        if (focalPointRatio === void 0) { focalPointRatio = 0.0; }
        var context = this.context, sec = context.sec;
        // fillType = alToString(this.context, fillType);
        var colorsNative = sec.createArray(AVM1ArrayNative.mapToJSArray(colors, function (item) { return alToInt32(_this.context, item); }));
        var alphasNative = sec.createArray(AVM1ArrayNative.mapToJSArray(alphas, function (item) { return alToNumber(_this.context, item) / 100.0; }));
        var ratiosNative = sec.createArray(AVM1ArrayNative.mapToJSArray(ratios, function (item) { return alToNumber(_this.context, item); }));
        var matrixNative = null;
        if (isNullOrUndefined(matrix)) {
            somewhatImplemented('AVM1MovieClip.lineGradientStyle');
        }
        spreadMethod = alToString(this.context, spreadMethod);
        interpolationMethod = alToString(this.context, interpolationMethod);
        focalPointRatio = alToNumber(this.context, focalPointRatio);
        this.graphics.lineGradientStyle(fillType, colorsNative, alphasNative, ratiosNative, matrixNative, spreadMethod, interpolationMethod, focalPointRatio);
    };
    AVM1MovieClip.prototype.lineStyle = function (thickness, rgb, alpha, pixelHinting, noScale, capsStyle, jointStyle, miterLimit) {
        if (thickness === void 0) { thickness = NaN; }
        if (rgb === void 0) { rgb = 0x000000; }
        if (alpha === void 0) { alpha = 100; }
        if (pixelHinting === void 0) { pixelHinting = false; }
        if (noScale === void 0) { noScale = 'normal'; }
        if (capsStyle === void 0) { capsStyle = 'round'; }
        if (jointStyle === void 0) { jointStyle = 'round'; }
        if (miterLimit === void 0) { miterLimit = 3; }
        thickness = alToNumber(this.context, thickness);
        rgb = alToInt32(this.context, rgb);
        pixelHinting = alToBoolean(this.context, pixelHinting);
        noScale = alToString(this.context, noScale);
        var capsStyleInt = AVM1MovieClip.capStyleMapStringToInt[alToString(this.context, capsStyle)];
        var jointStyleInt = AVM1MovieClip.jointStyleMapStringToInt[alToString(this.context, jointStyle)];
        miterLimit = alToNumber(this.context, miterLimit);
        this.graphics.lineStyle(thickness, rgb, alpha / 100.0, pixelHinting, AVM1MovieClip.noScaleDictionary[noScale], capsStyleInt, jointStyleInt, miterLimit);
    };
    AVM1MovieClip.prototype.lineTo = function (x, y) {
        x = toTwipFloor(alToNumber(this.context, x));
        y = toTwipFloor(alToNumber(this.context, y));
        this.graphics.lineTo(x, y);
    };
    AVM1MovieClip.prototype.loadMovie = function (url, method) {
        var loaderHelper = new AVM1LoaderHelper(this.context);
        loaderHelper.load(url, method).then(function () {
            if (loaderHelper.content == null) {
                warning("loadMovie - content is null");
                return;
            }
            this.adaptee.timeline = loaderHelper.content.timeline;
            this.adaptee.assetNamespace = loaderHelper.content.assetNamespace;
            this.adaptee.reset(true);
        }.bind(this));
    };
    AVM1MovieClip.prototype.loadVariables = function (url, method) {
        this.context.actions._loadVariables(this, url, method);
    };
    AVM1MovieClip.prototype.localToGlobal = function (pt) {
        if (!pt) {
            return;
        }
        var tmp = toAS3Point(pt);
        this.adaptee.transform.localToGlobal(tmp, tmp);
        copyAS3PointTo(tmp, pt);
    };
    AVM1MovieClip.prototype.get_lockroot = function () {
        return this._lockroot;
    };
    AVM1MovieClip.prototype.set_lockroot = function (value) {
        somewhatImplemented('AVM1MovieClip._lockroot');
        this._lockroot = alToBoolean(this.context, value);
    };
    AVM1MovieClip.prototype.moveTo = function (x, y) {
        x = toTwipFloor(alToNumber(this.context, x));
        y = toTwipFloor(alToNumber(this.context, y));
        this.graphics.moveTo(x, y);
    };
    AVM1MovieClip.prototype.nextFrame = function () {
        ++this.adaptee.currentFrameIndex;
    };
    AVM1MovieClip.prototype.nextScene = function () {
        notImplemented('AVM1MovieClip.nextScene');
    };
    AVM1MovieClip.prototype.play = function () {
        this.adaptee.play();
    };
    AVM1MovieClip.prototype.prevFrame = function () {
        --this.adaptee.currentFrameIndex;
        this.adaptee.stop();
    };
    AVM1MovieClip.prototype.prevScene = function () {
        notImplemented('AVM1MovieClip.prevScene');
    };
    AVM1MovieClip.prototype.setMask = function (mc) {
        if (mc == null) {
            // Cancel a mask.
            this.adaptee.mask = null;
            return;
        }
        var mask = this.context.resolveTarget(mc);
        if (mask) {
            this.adaptee.mask = getAwayJSAdaptee(mask);
        }
    };
    AVM1MovieClip.prototype.startDrag = function (lock, left, top, right, bottom) {
        if (AVM1MovieClip.currentDraggedMC) {
            AVM1MovieClip.currentDraggedMC.stopDrag();
        }
        AVM1MovieClip.currentDraggedMC = this;
        lock = alToBoolean(this.context, lock);
        this._dragBounds = null;
        if (left > right) {
            var tmp = right;
            right = left;
            left = tmp;
        }
        if (top > bottom) {
            var tmp = bottom;
            bottom = top;
            top = tmp;
        }
        if (arguments.length > 1) {
            left = alToNumber(this.context, left);
            top = alToNumber(this.context, top);
            right = alToNumber(this.context, right);
            bottom = alToNumber(this.context, bottom);
            //console.log("left", left,"top", top, "right", right, "bottom", bottom );
            this._dragBounds = new Rectangle(left, top, right - left, bottom - top);
        } //todo: listen on stage
        if (!this.isDragging) {
            this.isDragging = true;
            this.startDragPoint = this.adaptee.parent.transform.globalToLocal(new Point(this.context.globals.Stage._awayAVMStage.mouseX, this.context.globals.Stage._awayAVMStage.mouseY));
            if (lock) {
                this.adaptee.x = this.startDragPoint.x;
                this.adaptee.y = this.startDragPoint.y;
            }
            if (this._dragBounds)
                this.checkBounds();
            this.startDragMCPosition.x = this.adaptee.x;
            this.startDragMCPosition.y = this.adaptee.y;
            AVM1Stage.stage.addEventListener("mouseMove3d", this.dragListenerDelegate);
            //window.addEventListener("mouseup", this.stopDragDelegate);
            //window.addEventListener("touchend", this.stopDragDelegate);
            AVM1Stage.stage.scene.mousePicker.dragEntity = this.adaptee;
            MouseManager.getInstance(AVM1Stage.stage.scene.renderer.renderGroup.pickGroup).isAVM1Dragging = true;
        }
    };
    AVM1MovieClip.prototype.dragListener = function (e) {
        //console.log("drag", e);
        //console.log("mouseX", (<any>this.context.globals.Stage)._awayAVMStage.mouseX);
        //console.log("mouseY", (<any>this.context.globals.Stage)._awayAVMStage.mouseY);
        if (this.adaptee.parent) {
            var tmpPoint = this.adaptee.parent.transform.globalToLocal(new Point(this.context.globals.Stage._awayAVMStage.mouseX, this.context.globals.Stage._awayAVMStage.mouseY));
            this.adaptee.x = this.startDragMCPosition.x + (tmpPoint.x - this.startDragPoint.x);
            this.adaptee.y = this.startDragMCPosition.y + (tmpPoint.y - this.startDragPoint.y);
            if (this._dragBounds)
                this.checkBounds();
        }
    };
    AVM1MovieClip.prototype.checkBounds = function () {
        if (this.adaptee.x < (this._dragBounds.left)) {
            this.adaptee.x = this._dragBounds.left;
        }
        if (this.adaptee.x > (this._dragBounds.right)) {
            this.adaptee.x = (this._dragBounds.right);
        }
        if (this.adaptee.y < this._dragBounds.top) {
            this.adaptee.y = this._dragBounds.top;
        }
        if (this.adaptee.y > (this._dragBounds.bottom)) {
            this.adaptee.y = this._dragBounds.bottom;
        }
    };
    AVM1MovieClip.prototype.stop = function () {
        return this.adaptee.stop();
    };
    AVM1MovieClip.prototype.stopDrag = function (e) {
        if (e === void 0) { e = null; }
        if (AVM1MovieClip.currentDraggedMC && AVM1MovieClip.currentDraggedMC != this) {
            AVM1MovieClip.currentDraggedMC.stopDrag();
        }
        this.isDragging = false;
        AVM1MovieClip.currentDraggedMC = null;
        AVM1Stage.stage.scene.mousePicker.dragEntity = null;
        MouseManager.getInstance(AVM1Stage.stage.scene.renderer.renderGroup.pickGroup).isAVM1Dragging = true;
        AVM1Stage.stage.removeEventListener("mouseMove3d", this.dragListenerDelegate);
        window.removeEventListener("mouseup", this.stopDragDelegate);
        window.removeEventListener("touchend", this.stopDragDelegate);
    };
    AVM1MovieClip.prototype.swapDepths = function (target) {
        //return;
        // if this is the scene, or if no parent exists, we do not want to do anything
        if (this.adaptee.isAVMScene || !this.get_parent()) {
            return;
        }
        var parent = getAwayJSAdaptee(this.get_parent());
        if (!parent) {
            warning("AVM1MovieClip.swapDepth called for object with no parent");
            return;
        }
        if (typeof target === 'undefined') {
            warning("AVM1MovieClip.swapDepth called with undefined as target depth");
            return;
        }
        if (typeof target === 'number') {
            //console.log("swap to number", this.adaptee.name, target);
            parent.swapDepths(this.adaptee, avm2AwayDepth(target));
        }
        else if (target.adaptee) {
            target._blockedByScript = true;
            target._ctBlockedByScript = true;
            //console.log("swap to children", this.adaptee.name, target.adaptee.name);
            parent.swapChildren(this.adaptee, target.adaptee);
        }
        this._blockedByScript = true;
        this._ctBlockedByScript = true;
        if (this.adaptee._depthID == this.initialDepth) {
            this.hasSwappedDepth = false;
        }
        else {
            this.hasSwappedDepth = true;
        }
        if (this.adaptee.name && parent && parent.adapter) {
            // the object has a name, and might be registered for access via script
            // we need to check if that registration must be updated to another mc that has a higher depth
            parent.adapter.getLatestObjectForName(this.adaptee.name.toLowerCase());
        }
    };
    AVM1MovieClip.prototype.getTabChildren = function () {
        return getAwayObjectOrTemplate(this).tabChildren;
    };
    AVM1MovieClip.prototype.setTabChildren = function (value) {
        getAwayObjectOrTemplate(this).tabChildren = alToBoolean(this.context, value);
    };
    AVM1MovieClip.prototype.get_totalframes = function () {
        return this.adaptee.numFrames;
    };
    AVM1MovieClip.prototype.getTrackAsMenu = function () {
        notImplemented("AVM1MovieClip.getTrackAsMenu()");
        return getAwayObjectOrTemplate(this).trackAsMenu;
    };
    AVM1MovieClip.prototype.setTrackAsMenu = function (value) {
        notImplemented("AVM1MovieClip.setTrackAsMenu()");
        getAwayObjectOrTemplate(this).trackAsMenu = alToBoolean(this.context, value);
    };
    AVM1MovieClip.prototype.unloadMovie = function () {
        var nativeObject = this.adaptee;
        this.adaptee.constructedKeyFrameIndex = 0;
        this.adaptee.stop();
        nativeObject.removeChildren(0, nativeObject.numChildren);
        if ((this.dynamicallyCreated || this.hasSwappedDepth) && nativeObject.parent && nativeObject.parent.adapter) {
            nativeObject.parent.adapter.unregisterScriptObject(nativeObject);
        }
    };
    AVM1MovieClip.prototype.getUseHandCursor = function () {
        return this.adaptee.useHandCursor;
    };
    AVM1MovieClip.prototype.setUseHandCursor = function (value) {
        if (!this.adaptee)
            return;
        this.adaptee.useHandCursor = value;
    };
    AVM1MovieClip.prototype.setParameters = function (parameters) {
        for (var paramName in parameters) {
            if (!this.alHasProperty(paramName)) {
                this.alPut(paramName, parameters[paramName]);
            }
        }
    };
    // Special and children names properties resolutions
    AVM1MovieClip.prototype._resolveLevelNProperty = function (name) {
        release || assert(alIsName(this.context, name));
        if (name === '_level0') {
            return this.context.resolveLevel(0);
        }
        else if (name === '_root') {
            return this.context.resolveRoot();
        }
        else if (name.indexOf('_level') === 0) {
            var level = name.substring(6);
            var levelNum = level | 0;
            if (levelNum > 0 && level == levelNum) {
                return this.context.resolveLevel(levelNum);
            }
        }
        return null;
    };
    AVM1MovieClip.prototype._getCachedPropertyResult = function (value) {
        if (!this._cachedPropertyResult) {
            this._cachedPropertyResult = {
                flags: 64 /* DATA */ | 1 /* DONT_ENUM */, value: value
            };
        }
        else {
            this._cachedPropertyResult.value = value;
        }
        return this._cachedPropertyResult;
    };
    AVM1MovieClip.prototype.alGetOwnProperty = function (name) {
        var desc = _super.prototype.alGetOwnProperty.call(this, name);
        if (desc) {
            return desc;
        }
        if (name[0] === '_') {
            if ((name[1] === 'l' && name.indexOf('_level') === 0 ||
                name[1] === 'r' && name.indexOf('_root') === 0)) {
                var level = this._resolveLevelNProperty(name);
                if (level) {
                    return this._getCachedPropertyResult(level);
                }
            }
            else if (name.toLowerCase() in MovieClipProperties) {
                // For MovieClip's properties that start from '_' case does not matter.
                return _super.prototype.alGetOwnProperty.call(this, name.toLowerCase());
            }
        }
        if (hasAwayJSAdaptee(this)) {
            var child = this._lookupChildByName(name);
            if (child) {
                return this._getCachedPropertyResult(child);
            }
        }
        return undefined;
    };
    AVM1MovieClip.prototype.alGetOwnPropertiesKeys = function () {
        var keys = _super.prototype.alGetOwnPropertiesKeys.call(this);
        // if it's a movie listing the children as well
        if (!hasAwayJSAdaptee(this)) {
            return keys; // not initialized yet
        }
        var as3MovieClip = this.adaptee;
        if (as3MovieClip._children.length === 0) {
            return keys; // no children
        }
        var processed = Object.create(null);
        var keysLength = keys.length;
        var i = 0;
        for (i = 0; i < keysLength; i++) {
            processed[keys[i]] = true;
        }
        var numChilds = as3MovieClip._children.length;
        var child = null;
        var name = null;
        var normalizedName = null;
        for (i = 0; i < numChilds; i++) {
            child = as3MovieClip._children[i];
            name = child.name;
            normalizedName = name; // TODO something like this._unescapeProperty(this._escapeProperty(name));
            processed[normalizedName] = true;
        }
        return Object.getOwnPropertyNames(processed);
    };
    AVM1MovieClip.prototype._init = function (initObject) {
        var _this = this;
        if (initObject instanceof AVM1Object) {
            alForEachProperty(initObject, function (name) {
                _this.alPut(name, initObject.alGet(name));
            }, null);
        }
    };
    AVM1MovieClip.prototype._initEventsHandlers = function () {
        this.bindEvents(EventsListForMC);
    };
    AVM1MovieClip.currentMCAssetNameSpace = "";
    AVM1MovieClip.currentDraggedMC = null;
    // if a stop-action occurs, we check if a child with this name is present, and if so, we execute the function provided
    AVM1MovieClip.pokiSDKonStopActionChildName = null;
    AVM1MovieClip.pokiSDKonStopAction = null;
    AVM1MovieClip.capStyleMapStringToInt = { "none": 0, "round": 1, "square": 2 };
    AVM1MovieClip.jointStyleMapStringToInt = { "round": 0, "bevel": 1, "miter": 2 };
    AVM1MovieClip.noScaleDictionary = {
        'normal': LineScaleMode.NORMAL,
        'none': LineScaleMode.NONE,
        'vertical': LineScaleMode.VERTICAL,
        'horizontal': LineScaleMode.HORIZONTAL
    };
    return AVM1MovieClip;
}(AVM1SymbolBase));
export { AVM1MovieClip };
