import { __extends } from "tslib";
import { assert } from "@awayjs/graphics";
import { release, notImplemented } from "../../base/utilities/Debug";
import { DisplayObject } from "./DisplayObject";
import { Sprite } from './Sprite';
import { MovieClip } from './MovieClip';
/**
 * Copyright 2014 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * AVM1Movie is the reflection of AVM1 SWFs loaded into AVM2 content. Since AVM1 content is
 * completely opaque to AVM2 content, it's not a DisplayObjectContainer, even though it contains
 * nested children. This is because the two worlds are completely separated from each other[1], and
 * each AVM1 SWF is entirely isolated from everything else.
 *
 * This causes a few headaches because we implement the AVM1 display list in terms of the AVM2
 * display list: each AVM1 MovieClip is a wrapper around an AVM2 MovieClip instance, which is
 * what's actually on stage. Theoretically, the top-most AVM2 MovieClip for an AVM1 SWF isn't
 * supposed to have a parent. However, we need it to be part of the stage's display tree in order
 * to take part in rendering.
 *
 * Therefore, the AVM2 MovieClip wrapped by an AVM1Movie gets the latter set as its parent, even
 * though AVM1Movie isn't a DisplayObjectContainer. We borrow methods from that and generally
 * pretend that AVM1Movie is a container in some places to pull that off.
 *
 * [1]: If you ignore the undocumented `call` and `addCallback` methods for a moment.
 */
var AVM1Movie = /** @class */ (function (_super) {
    __extends(AVM1Movie, _super);
    function AVM1Movie(level0) {
        var _this = _super.call(this) || this;
        _this._content = Sprite.axClass.axConstruct();
        _this._children = [];
        _this._children[0] = _this._content;
        // Pretend we're a DisplayObjectContainer and can have children. See comment at the top.
        _this._content._setParent(_this, 0);
        _this._setDirtyFlags(2 /* DirtyChildren */);
        _this._invalidateFillAndLineBounds(true, true);
        _this.sec.flash.display.DisplayObject.axClass._advancableInstances.push(_this);
        _this._constructed = false;
        // Setting _level0 root.
        _this._content.addTimelineObjectAtDepth(level0, 0);
        return _this;
    }
    AVM1Movie.prototype.call = function (functionName) {
        notImplemented('AVM1Movie#call');
    };
    AVM1Movie.prototype.addCallback = function (functionName, closure) {
        notImplemented('AVM1Movie#call');
    };
    AVM1Movie.prototype._addFrame = function (frame) {
        this._content._children[0]._addFrame(frame);
    };
    AVM1Movie.prototype._initFrame = function (advance) {
        // Empty implementation: AVM1Movie doesn't have frames, and the contained MovieClip
        // adds itself to the IAdvancables list.
    };
    AVM1Movie.prototype._constructFrame = function () {
        // On custructFrame we need to fully construct the roots container.
        // Once constructed, its children (which are IAdvancable type) will be
        // receiving their own _constructFrame events.
        if (!this._constructed) {
            this._constructed = true;
            this._content._constructChildren();
        }
    };
    AVM1Movie.prototype._enqueueFrameScripts = function () {
        this._removeFlags(16384 /* ContainsFrameScriptPendingChildren */);
        this._content._enqueueFrameScripts();
    };
    AVM1Movie.prototype._propagateFlagsDown = function (flags) {
        if (this._hasFlags(flags)) {
            return;
        }
        this._setFlags(flags);
        this._content._propagateFlagsDown(flags);
    };
    /**
     * AVM1Movie only takes the AVM1 content into consideration when testing points against
     * bounding boxes, not otherwise.
     */
    AVM1Movie.prototype._containsPoint = function (globalX, globalY, localX, localY, testingType, objects) {
        if (testingType === 3 /* Mouse */) {
            return this._content._containsPoint(globalX, globalY, localX, localY, testingType, objects);
        }
        if (testingType !== 0 /* HitTestBounds */ ||
            !this._getContentBounds().contains(localX, localY)) {
            return 0 /* None */;
        }
        return 1 /* Bounds */;
    };
    /**
     * Override of DisplayObject#_getChildBounds that retrieves the AVM1 content's bounds.
     */
    AVM1Movie.prototype._getChildBounds = function (bounds, includeStrokes) {
        var childBounds = this._content._getContentBounds(includeStrokes).clone();
        // Always apply the SimpleButton's matrix.
        this._getConcatenatedMatrix().transformBounds(childBounds);
        bounds.unionInPlace(childBounds);
    };
    AVM1Movie.prototype._getLevelForRoot = function (root) {
        release || assert(root.parent === this._content);
        return root._depth;
    };
    AVM1Movie.prototype._getRootForLevel = function (level) {
        return this._content.getTimelineObjectAtDepth(level);
    };
    AVM1Movie.prototype._addRoot = function (level, root) {
        release || assert(MovieClip.axClass.axIsType(root));
        this._removeRoot(level);
        release || assert(!this._content.getTimelineObjectAtDepth(level));
        this._content.addTimelineObjectAtDepth(root, level);
    };
    AVM1Movie.prototype._removeRoot = function (level) {
        var root = this._content.getTimelineObjectAtDepth(level);
        if (!root) {
            return false;
        }
        this._content.removeChild(root);
        return true;
    };
    AVM1Movie.classInitializer = null;
    AVM1Movie.classSymbols = null;
    AVM1Movie.instanceSymbols = null;
    return AVM1Movie;
}(DisplayObject));
export { AVM1Movie };
