import { __extends } from "tslib";
/**
 * Copyright 2014 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
// Class: BitmapData
import { release, error, assert, notImplemented, somewhatImplemented } from "../../base/utilities/Debug";
import { DisplaySymbol } from "../symbol";
import { ImageType, Bounds } from "../../base/utilities";
import { ASObject } from "../../avm2/nat/ASObject";
import { Rectangle } from "../geom/Rectangle";
import { ColorUtilitites, unpremultiplyARGB, RGBAToARGB, premultiplyARGB } from "../../base/utilities/ColorUtilities";
import { Errors } from "../../avm2/errors";
import { DisplayObject } from "./DisplayObject";
import { ByteArray } from "../../avm2/natives/byteArray";
import { DataBuffer } from "@awayjs/graphics";
import { indexOf } from "../../base/utilities/ArrayUtilities";
import { swap32 } from "../../base/utilities/IntegerUtilities";
import { axCoerceString } from "../../avm2/run/axCoerceString";
import { constructClassFromSymbol } from "../../avm2/constructClassFromSymbol";
/**
 * Holds blobs of bitmap data in various formats and lets you do basic pixel operations. When
 * data is unpacked, it is stored as premultiplied ARGB since it's what the SWF encodes bitmaps
 * as.  This way we don't have to do unecessary byte conversions.
 */
var BitmapData = /** @class */ (function (_super) {
    __extends(BitmapData, _super);
    function BitmapData(width /*int*/, height /*int*/, transparent, fillColorARGB) {
        if (transparent === void 0) { transparent = true; }
        if (fillColorARGB === void 0) { fillColorARGB = 4294967295; }
        var _this = this;
        width = width | 0;
        height = height | 0;
        transparent = !!transparent;
        fillColorARGB = fillColorARGB | 0;
        _this = _super.call(this) || this;
        if (_this._symbol) {
            _this.applySymbol();
            return;
        }
        if (width > BitmapData.MAXIMUM_WIDTH || width <= 0 ||
            height > BitmapData.MAXIMUM_HEIGHT || height <= 0 ||
            width * height > BitmapData.MAXIMUM_DIMENSION) {
            _this.sec.throwError('ArgumentError', Errors.InvalidBitmapData);
        }
        _this._rect = new Rectangle(0, 0, width, height);
        _this._transparent = transparent;
        _this._id = DisplayObject.getNextSyncID();
        _this._setData(new Uint8Array(width * height * 4), ImageType.PremultipliedAlphaARGB);
        var alpha = fillColorARGB >> 24;
        if (alpha === 0 && transparent) {
            // No need to do an initial fill since this would all be zeros anyway.
            _this._solidFillColorPBGRA = 0;
        }
        else {
            _this.fillRect(_this._rect, fillColorARGB);
        }
        _this._bitmapReferrers = [];
        release || assert(_this._isDirty === !!_this._data);
        release || assert(_this._isRemoteDirty === !_this._data);
        return _this;
    }
    BitmapData.classInitializer = function () {
        this._temporaryRectangle = new Rectangle();
    };
    BitmapData.prototype.applySymbol = function () {
        release || assert(this._symbol);
        var symbol = this._symbol;
        release || assert(symbol.syncId);
        this._rect = new Rectangle(0, 0, symbol.width, symbol.height);
        this._transparent = true;
        this._id = symbol.syncId;
        if (symbol.type === ImageType.PremultipliedAlphaARGB ||
            symbol.type === ImageType.StraightAlphaARGB ||
            symbol.type === ImageType.StraightAlphaRGBA) {
            release || assert(symbol.data);
            this._setData(symbol.data, symbol.type);
        }
        else {
            this._isDirty = false;
            this._isRemoteDirty = true;
        }
        this._solidFillColorPBGRA = null;
        this._bitmapReferrers = [];
    };
    BitmapData.prototype._setData = function (data, type) {
        // Alpha images are parsed to Uint8Clamped array. Easiest to just handle here.
        if (data instanceof Uint8ClampedArray) {
            data = new Uint8Array(data.buffer);
        }
        release || assert(data instanceof Uint8Array);
        this._data = data;
        this._type = type;
        this._view = new Int32Array(data.buffer);
        this._dataBuffer = DataBuffer.FromArrayBuffer(data.buffer);
        this._isDirty = true;
        this._isRemoteDirty = false;
    };
    BitmapData.prototype._addBitmapReferrer = function (bitmap) {
        var index = indexOf(this._bitmapReferrers, bitmap);
        release || assert(index < 0);
        this._bitmapReferrers.push(bitmap);
    };
    BitmapData.prototype._removeBitmapReferrer = function (bitmap) {
        var index = indexOf(this._bitmapReferrers, bitmap);
        release || assert(index >= 0);
        this._bitmapReferrers[index] = null;
    };
    /**
     * Called whenever the contents of this bitmap data changes.
     */
    BitmapData.prototype._invalidate = function () {
        if (this._isDirty) {
            return;
        }
        this._isDirty = true;
        this._isRemoteDirty = false;
        // TODO: We probably don't need to propagate any flags if |_locked| is true.
        for (var i = 0; i < this._bitmapReferrers.length; i++) {
            var bitmap = this._bitmapReferrers[i];
            if (bitmap) {
                bitmap._setDirtyFlags(16 /* DirtyBitmapData */);
            }
        }
    };
    BitmapData.prototype._getTemporaryRectangleFrom = function (rect) {
        var r = BitmapData.axClass._temporaryRectangle;
        r.copyFrom(rect);
        return r;
    };
    BitmapData.prototype.getDataBuffer = function () {
        return this._dataBuffer;
    };
    BitmapData.prototype._getContentBounds = function () {
        return Bounds.FromRectangle(this._rect);
    };
    /**
     * TODO: Not tested.
     */
    BitmapData.prototype._getPixelData = function (rect) {
        var r = this._getTemporaryRectangleFrom(this._rect).intersectInPlace(rect);
        if (r.isEmpty()) {
            return;
        }
        var xMin = r.x;
        var xMax = r.x + r.width;
        var yMin = r.y;
        var yMax = r.y + r.height;
        this._ensureBitmapData();
        var view = this._view;
        var width = this._rect.width;
        var output = new Int32Array(r.area);
        var p = 0;
        for (var y = yMin; y < yMax; y++) {
            var offset = y * width;
            for (var x = xMin; x < xMax; x++) {
                var colorBGRA = view[offset + x];
                var alpha = colorBGRA & 0xff;
                var colorBGR = colorBGRA >>> 8;
                colorBGRA = ((255 * colorBGR) / alpha) << 8 | alpha;
                output[p++] = colorBGRA;
            }
        }
        return output;
    };
    /**
     * TODO: Not tested.
     */
    BitmapData.prototype._putPixelData = function (rect, input) {
        var r = this._getTemporaryRectangleFrom(this._rect).intersectInPlace(rect);
        if (r.isEmpty()) {
            return;
        }
        var xMin = r.x;
        var xMax = r.x + r.width;
        var yMin = r.y;
        var yMax = r.y + r.height;
        this._ensureBitmapData();
        var view = this._view;
        var width = this._rect.width;
        var p = (rect.width * rect.height - r.height) + (xMin - rect.x);
        var padding = rect.width - r.width;
        var alphaMask = this._transparent ? 0x00 : 0xff;
        for (var y = yMin; y < yMax; y++) {
            var offset = y * width;
            for (var x = xMin; x < xMax; x++) {
                var colorBGRA = input[p++];
                var alpha = colorBGRA & alphaMask;
                var colorBGR = colorBGRA >>> 8;
                view[offset + x] = (((colorBGR * alpha + 254) / 255) & 0x00ffffff) << 8 | alpha;
            }
            p += padding;
        }
        this._invalidate();
    };
    Object.defineProperty(BitmapData.prototype, "width", {
        get: function () {
            return this._rect.width;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BitmapData.prototype, "height", {
        get: function () {
            return this._rect.height;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BitmapData.prototype, "rect", {
        get: function () {
            return this._rect.clone();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BitmapData.prototype, "transparent", {
        get: function () {
            return this._transparent;
        },
        enumerable: true,
        configurable: true
    });
    BitmapData.prototype.clone = function () {
        var bd = Object.create(BitmapData.axClass.tPrototype);
        bd._rect = this._rect.clone();
        bd._transparent = this._transparent;
        bd._solidFillColorPBGRA = this._solidFillColorPBGRA;
        bd._bitmapReferrers = [];
        // TODO: clone without reading back bitmapdata. There's no need for that.
        this._ensureBitmapData();
        bd._id = DisplayObject.getNextSyncID();
        bd._setData(new Uint8Array(this._data), this._type);
        return bd;
    };
    /**
     * Returns an straight alpha RGB pixel value 0x00RRGGBB.
     */
    BitmapData.prototype.getPixel = function (x /*int*/, y /*int*/) {
        x = x | 0;
        y = y | 0;
        return this.getPixel32(x, y) & 0x00ffffff;
    };
    /**
     * Returns an straight alpha ARGB pixel value 0xAARRGGBB.
     */
    BitmapData.prototype.getPixel32 = function (x /*int*/, y /*int*/) {
        x = x | 0;
        y = y | 0;
        if (!this._rect.contains(x, y)) {
            return 0;
        }
        this._ensureBitmapData();
        var value = this._view[y * this._rect.width + x];
        switch (this._type) {
            case ImageType.PremultipliedAlphaARGB:
                var pARGB = swap32(value);
                var uARGB = unpremultiplyARGB(pARGB);
                return uARGB >>> 0;
            case ImageType.StraightAlphaRGBA:
                return RGBAToARGB(swap32(value));
            default:
                notImplemented(ImageType[this._type]);
                return 0;
        }
    };
    BitmapData.prototype.setPixel = function (x /*int*/, y /*int*/, uARGB /*uint*/) {
        // TODO: implement this as a wrapper for setPixel32.
        x = x | 0;
        y = y | 0;
        uARGB = uARGB | 0;
        if (!this._rect.contains(x, y)) {
            return;
        }
        this._ensureBitmapData();
        var i = y * this._rect.width + x;
        var a = this._view[i] & 0xff;
        uARGB = uARGB & 0x00ffffff | a << 24;
        var pARGB = premultiplyARGB(uARGB);
        this._view[i] = swap32(pARGB);
        this._invalidate();
        this._solidFillColorPBGRA = null;
    };
    BitmapData.prototype.setPixel32 = function (x /*int*/, y /*int*/, uARGB /*uint*/) {
        x = x | 0;
        y = y | 0;
        if (!this._rect.contains(x, y)) {
            return;
        }
        this._ensureBitmapData();
        var a = uARGB >>> 24;
        var uRGB = uARGB & 0x00ffffff;
        if (this._transparent) {
            var uARGB = uRGB | a << 24;
            var pARGB = premultiplyARGB(uARGB);
        }
        else {
            var pARGB = uRGB | 0xff000000;
        }
        this._view[y * this._rect.width + x] = swap32(pARGB);
        this._invalidate();
        this._solidFillColorPBGRA = null;
    };
    BitmapData.prototype.applyFilter = function (sourceBitmapData, sourceRect, destPoint, filter) {
        sourceBitmapData = sourceBitmapData;
        sourceRect = sourceRect;
        destPoint = destPoint;
        filter = filter;
        release || somewhatImplemented("public flash.display.BitmapData::applyFilter " + filter);
        return;
    };
    BitmapData.prototype.colorTransform = function (rect, colorTransform) {
        rect = rect;
        colorTransform = colorTransform;
        release || somewhatImplemented("public flash.display.BitmapData::colorTransform");
        return;
    };
    BitmapData.prototype.compare = function (otherBitmapData) {
        otherBitmapData = otherBitmapData;
        release || notImplemented("public flash.display.BitmapData::compare");
        return;
    };
    BitmapData.prototype.copyChannel = function (sourceBitmapData, sourceRect, destPoint, sourceChannel /*uint*/, destChannel
    /*uint*/ ) {
        sourceBitmapData = sourceBitmapData;
        sourceRect = sourceRect;
        destPoint = destPoint;
        sourceChannel = sourceChannel >>> 0;
        destChannel = destChannel >>> 0;
        release || somewhatImplemented("public flash.display.BitmapData::copyChannel");
        return;
    };
    /**
     * Copies a rectangular region of pixels into the current bitmap data.
     */
    BitmapData.prototype.copyPixels = function (sourceBitmapData, sourceRect, destPoint, alphaBitmapData, alphaPoint, mergeAlpha) {
        if (alphaBitmapData === void 0) { alphaBitmapData = null; }
        if (alphaPoint === void 0) { alphaPoint = null; }
        mergeAlpha = !!mergeAlpha;
        if (alphaBitmapData || alphaPoint) {
            release || somewhatImplemented("public flash.display.BitmapData::copyPixels - Alpha");
            return;
        }
        // Deal with fractional pixel coordinates, looks like Flash "rounds" the corners of
        // the source rect, however a width of |0.5| rounds down rather than up so we're not
        // quite correct here.
        var sRect;
        if (sourceRect) {
            sRect = this._getTemporaryRectangleFrom(sourceRect).roundInPlace();
        }
        else {
            sRect = BitmapData.axClass._temporaryRectangle.setEmpty();
        }
        var tBRect = this._rect;
        var sBRect = sourceBitmapData._rect;
        // Clip sRect against SBRect.
        var sL = Math.max(sRect.x, 0);
        var sT = Math.max(sRect.y, 0);
        var sR = Math.min(sRect.x + sRect.width, sBRect.width);
        var sB = Math.min(sRect.y + sRect.height, sBRect.height);
        // Compute source rect offsets (in case the source rect had negative x, y coordinates).
        var oX = sL - sRect.x;
        var oY = sT - sRect.y;
        var tL = (destPoint.x | 0) + oX;
        var tT = (destPoint.y | 0) + oY;
        if (tL < 0) {
            sL -= tL;
            tL = 0;
        }
        if (tT < 0) {
            sT -= tT;
            tT = 0;
        }
        var tW = Math.min(sR - sL, tBRect.width - tL);
        var tH = Math.min(sB - sT, tBRect.height - tT);
        if (tW <= 0 || tH <= 0) {
            return;
        }
        var sX = sL;
        var sY = sT;
        var tX = tL;
        var tY = tT;
        var sStride = sourceBitmapData._rect.width;
        var tStride = this._rect.width;
        this._ensureBitmapData();
        sourceBitmapData._ensureBitmapData();
        var s = sourceBitmapData._view;
        var t = this._view;
        if (sourceBitmapData._type !== this._type) {
            release || somewhatImplemented("public flash.display.BitmapData::copyPixels - Color Format Conversion");
        }
        if (mergeAlpha && this._type !== ImageType.PremultipliedAlphaARGB) {
            release || somewhatImplemented("public flash.display.BitmapData::copyPixels - Merge Alpha");
            return;
        }
        // No reason to copy pixels since since both source and target are the same solid fill,
        // regardless of alpha blending. (TODO: I think the math works out for mergeAlpha also.)
        if (this._solidFillColorPBGRA !== null &&
            this._solidFillColorPBGRA === sourceBitmapData._solidFillColorPBGRA) {
            return;
        }
        // Source has a solid fill but is fully opaque, we can get away without alpha blending here.
        if (sourceBitmapData._solidFillColorPBGRA !== null &&
            (sourceBitmapData._solidFillColorPBGRA & 0xFF) === 0xFF) {
            mergeAlpha = false;
        }
        // Finally do the copy. All the math above is needed just so we don't do any branches inside
        // this hot loop.
        if (mergeAlpha) {
            this._copyPixelsAndMergeAlpha(s, sX, sY, sStride, t, tX, tY, tStride, tW, tH);
        }
        else {
            var sP = (sY * sStride + sX) | 0;
            var tP = (tY * tStride + tX) | 0;
            if ((tW & 3) === 0) {
                for (var y = 0; y < tH; y = y + 1 | 0) {
                    for (var x = 0; x < tW; x = x + 4 | 0) {
                        t[(tP + x + 0) | 0] = s[(sP + x + 0) | 0];
                        t[(tP + x + 1) | 0] = s[(sP + x + 1) | 0];
                        t[(tP + x + 2) | 0] = s[(sP + x + 2) | 0];
                        t[(tP + x + 3) | 0] = s[(sP + x + 3) | 0];
                    }
                    sP = sP + sStride | 0;
                    tP = tP + tStride | 0;
                }
            }
            else {
                for (var y = 0; y < tH; y = y + 1 | 0) {
                    for (var x = 0; x < tW; x = x + 1 | 0) {
                        t[tP + x | 0] = s[sP + x | 0];
                    }
                    sP = sP + sStride | 0;
                    tP = tP + tStride | 0;
                }
            }
        }
        this._solidFillColorPBGRA = null;
        this._invalidate();
    };
    BitmapData.prototype._copyPixelsAndMergeAlpha = function (s, sX, sY, sStride, t, tX, tY, tStride, tW, tH) {
        var sP = (sY * sStride + sX) | 0;
        var tP = (tY * tStride + tX) | 0;
        for (var y = 0; y < tH; y = y + 1 | 0) {
            for (var x = 0; x < tW; x = x + 1 | 0) {
                var spBGRA = s[sP + x | 0];
                var sA = spBGRA & 0xff;
                // Optimize for the case where the source pixel is fully opaque or transparent. This
                // pays off if the source image has many such pixels but slows down the normal case.
                if (sA === 0xff) {
                    t[tP + x | 0] = spBGRA; // Opaque, just copy value over.
                }
                else if (sA === 0) {
                    // Transparent, don't do anything.
                }
                else {
                    // Compute the blending equation: src.rgb + (dst.rgb * (1 - src.a)). The trick here
                    // is to compute GA and BR at the same time without pulling apart each channel.
                    // We use the "double blend trick" (http://stereopsis.com/doubleblend.html) to
                    // compute GA and BR without unpacking them.
                    var sGA = spBGRA & 0x00ff00ff;
                    var sBR = spBGRA >> 8 & 0x00ff00ff;
                    var tpBGRA = t[tP + x | 0];
                    var tGA = tpBGRA & 0x00ff00ff;
                    var tBR = tpBGRA >> 8 & 0x00ff00ff;
                    var A = 256 - sA;
                    tGA = Math.imul(tGA, A) >> 8;
                    tBR = Math.imul(tBR, A) >> 8;
                    // TODO: Not sure if target alpha is computed correctly.
                    t[tP + x | 0] = ((sBR + tBR & 0x00ff00ff) << 8) | (sGA + tGA & 0x00ff00ff);
                }
            }
            sP = sP + sStride | 0;
            tP = tP + tStride | 0;
        }
    };
    BitmapData.prototype.dispose = function () {
        this._rect.setEmpty();
        this._view = null;
        this._invalidate();
    };
    BitmapData.prototype.draw = function (source, matrix, colorTransform, blendMode, clipRect, smoothing) {
        if (matrix === void 0) { matrix = null; }
        if (colorTransform === void 0) { colorTransform = null; }
        if (blendMode === void 0) { blendMode = null; }
        if (clipRect === void 0) { clipRect = null; }
        if (smoothing === void 0) { smoothing = false; }
        blendMode = axCoerceString(blendMode);
        smoothing = !!smoothing;
        release || somewhatImplemented("public flash.display.BitmapData::draw");
        var serializer = this.sec.player;
        if (matrix) {
            matrix = matrix.clone().toTwipsInPlace();
        }
        serializer.drawToBitmap(this, source, matrix, colorTransform, blendMode, clipRect, smoothing);
        this._isRemoteDirty = true;
    };
    BitmapData.prototype.drawWithQuality = function (source, matrix, colorTransform, blendMode, clipRect, smoothing, quality) {
        if (matrix === void 0) { matrix = null; }
        if (colorTransform === void 0) { colorTransform = null; }
        if (blendMode === void 0) { blendMode = null; }
        if (clipRect === void 0) { clipRect = null; }
        if (smoothing === void 0) { smoothing = false; }
        if (quality === void 0) { quality = null; }
        quality = axCoerceString(quality);
        release || somewhatImplemented("public flash.display.BitmapData::drawWithQuality");
        this.draw(source, matrix, colorTransform, blendMode, clipRect, smoothing);
    };
    BitmapData.prototype.fillRect = function (rect, uARGB /*uint*/) {
        this._ensureBitmapData();
        // TODO: what guarantees this, and why do we even need it?
        release || assert(this._type === ImageType.PremultipliedAlphaARGB);
        if (this._transparent) {
            var pARGB = premultiplyARGB(uARGB);
        }
        else {
            var pARGB = uARGB | 0xff000000;
        }
        var pBGRA = swap32(pARGB);
        var r = this._getTemporaryRectangleFrom(this._rect).intersectInPlace(rect);
        if (r.isEmpty()) {
            return;
        }
        // Filling with the same color?
        if (this._solidFillColorPBGRA === pBGRA) {
            return;
        }
        var view = this._view;
        // If we are filling the entire buffer, we can do a little better ~ 25% faster.
        if (r.equals(this._rect)) {
            var length = view.length | 0;
            // Unroll 4 iterations, ~ 5% faster.
            if ((length & 0x3) === 0) {
                for (var i = 0; i < length; i += 4) {
                    view[i] = pBGRA;
                    view[i + 1] = pBGRA;
                    view[i + 2] = pBGRA;
                    view[i + 3] = pBGRA;
                }
            }
            else {
                for (var i = 0; i < length; i++) {
                    view[i] = pBGRA;
                }
            }
            this._solidFillColorPBGRA = pBGRA;
        }
        else {
            var xMin = r.x | 0;
            var xMax = r.x + r.width | 0;
            var yMin = r.y | 0;
            var yMax = r.y + r.height | 0;
            var width = this._rect.width | 0;
            for (var y = yMin; y < yMax; y++) {
                var offset = y * width | 0;
                for (var x = xMin; x < xMax; x++) {
                    view[offset + x] = pBGRA;
                }
            }
            this._solidFillColorPBGRA = null;
        }
        this._invalidate();
    };
    BitmapData.prototype.floodFill = function (x /*int*/, y /*int*/, color /*uint*/) {
        x = x | 0;
        y = y | 0;
        color = color >>> 0;
        release || somewhatImplemented("public flash.display.BitmapData::floodFill");
        return;
    };
    BitmapData.prototype.generateFilterRect = function (sourceRect, filter) {
        sourceRect = sourceRect;
        filter = filter;
        release || somewhatImplemented("public flash.display.BitmapData::generateFilterRect");
        return;
    };
    BitmapData.prototype.getColorBoundsRect = function (mask /*uint*/, color /*uint*/, findColor) {
        if (findColor === void 0) { findColor = true; }
        mask = mask >>> 0;
        color = color >>> 0;
        findColor = !!findColor;
        release || somewhatImplemented("public flash.display.BitmapData::getColorBoundsRect");
        return new Rectangle();
    };
    BitmapData.prototype.getPixels = function (rect) {
        var outputByteArray = new ByteArray();
        this.copyPixelsToByteArray(rect, outputByteArray);
        return outputByteArray;
    };
    BitmapData.prototype.copyPixelsToByteArray = function (rect, data) {
        var pixelData = this._getPixelData(rect);
        if (!pixelData) {
            return;
        }
        data.writeRawBytes(new Uint8Array(pixelData));
    };
    BitmapData.prototype.getVector = function (rect) {
        var outputVector = new this.sec.Uint32Vector(pixelData.length);
        var pixelData = this._getPixelData(rect);
        if (!pixelData) {
            return outputVector;
        }
        outputVector.length = pixelData.length;
        outputVector._view().set(pixelData);
        return outputVector;
    };
    BitmapData.prototype.hitTest = function (firstPoint, firstAlphaThreshold /*uint*/, secondObject, secondBitmapDataPoint, secondAlphaThreshold) {
        if (secondBitmapDataPoint === void 0) { secondBitmapDataPoint = null; }
        if (secondAlphaThreshold === void 0) { secondAlphaThreshold = 1; }
        firstPoint = firstPoint;
        firstAlphaThreshold = firstAlphaThreshold >>> 0;
        secondObject = secondObject;
        secondBitmapDataPoint = secondBitmapDataPoint;
        secondAlphaThreshold = secondAlphaThreshold >>> 0;
        release || somewhatImplemented("public flash.display.BitmapData::hitTest");
        return true;
    };
    BitmapData.prototype.merge = function (sourceBitmapData, sourceRect, destPoint, redMultiplier /*uint*/, greenMultiplier
    /*uint*/ , blueMultiplier /*uint*/, alphaMultiplier /*uint*/) {
        sourceBitmapData = sourceBitmapData;
        sourceRect = sourceRect;
        destPoint = destPoint;
        redMultiplier = redMultiplier >>> 0;
        greenMultiplier = greenMultiplier >>> 0;
        blueMultiplier = blueMultiplier >>> 0;
        alphaMultiplier = alphaMultiplier >>> 0;
        release || somewhatImplemented("public flash.display.BitmapData::merge");
    };
    BitmapData.prototype.noise = function (randomSeed /*int*/, low, high, channelOptions, grayScale) {
        if (low === void 0) { low = 0; }
        if (high === void 0) { high = 255; }
        if (channelOptions === void 0) { channelOptions = 7; }
        if (grayScale === void 0) { grayScale = false; }
        randomSeed = randomSeed | 0;
        low = low >>> 0;
        high = high >>> 0;
        channelOptions = channelOptions >>> 0;
        grayScale = !!grayScale;
        release || somewhatImplemented("public flash.display.BitmapData::noise");
    };
    BitmapData.prototype.paletteMap = function (sourceBitmapData, sourceRect, destPoint, redArray, greenArray, blueArray, alphaArray) {
        if (redArray === void 0) { redArray = null; }
        if (greenArray === void 0) { greenArray = null; }
        if (blueArray === void 0) { blueArray = null; }
        if (alphaArray === void 0) { alphaArray = null; }
        sourceBitmapData = sourceBitmapData;
        sourceRect = sourceRect;
        destPoint = destPoint;
        redArray = redArray;
        greenArray = greenArray;
        blueArray = blueArray;
        alphaArray = alphaArray;
        release || somewhatImplemented("public flash.display.BitmapData::paletteMap");
    };
    BitmapData.prototype.perlinNoise = function (baseX, baseY, numOctaves /*uint*/, randomSeed
    /*int*/ , stitch, fractalNoise, channelOptions, grayScale, offsets) {
        if (channelOptions === void 0) { channelOptions = 7; }
        if (grayScale === void 0) { grayScale = false; }
        if (offsets === void 0) { offsets = null; }
        baseX = +baseX;
        baseY = +baseY;
        numOctaves = numOctaves >>> 0;
        randomSeed = randomSeed | 0;
        stitch = !!stitch;
        fractalNoise = !!fractalNoise;
        channelOptions = channelOptions >>> 0;
        grayScale = !!grayScale;
        offsets = offsets;
        release || somewhatImplemented("public flash.display.BitmapData::perlinNoise");
    };
    BitmapData.prototype.pixelDissolve = function (sourceBitmapData, sourceRect, destPoint, randomSeed, numPixels, fillColor) {
        if (randomSeed === void 0) { randomSeed = 0; }
        if (numPixels === void 0) { numPixels = 0; }
        if (fillColor === void 0) { fillColor = 0; }
        sourceBitmapData = sourceBitmapData;
        sourceRect = sourceRect;
        destPoint = destPoint;
        randomSeed = randomSeed | 0;
        numPixels = numPixels | 0;
        fillColor = fillColor >>> 0;
        release || somewhatImplemented("public flash.display.BitmapData::pixelDissolve");
        return;
    };
    BitmapData.prototype.scroll = function (x /*int*/, y /*int*/) {
        x = x | 0;
        y = y | 0;
        release || somewhatImplemented("public flash.display.BitmapData::scroll");
        return;
    };
    BitmapData.prototype.setPixels = function (rect, inputByteArray) {
        console.log("todo: BitmapData.setPixels");
        //this._putPixelData(rect, new Int32Array(inputByteArray.readRawBytes()));
    };
    BitmapData.prototype.setVector = function (rect, inputVector) {
        console.log("todo: BitmapData.setVector");
        //this._putPixelData(rect, inputVector._view());
    };
    BitmapData.prototype.threshold = function (sourceBitmapData, sourceRect, destPoint, operation, threshold /*uint*/, color, mask, copySource) {
        if (color === void 0) { color = 0; }
        if (mask === void 0) { mask = 4294967295; }
        if (copySource === void 0) { copySource = false; }
        sourceBitmapData = sourceBitmapData;
        sourceRect = sourceRect;
        destPoint = destPoint;
        operation = axCoerceString(operation);
        threshold = threshold >>> 0;
        color = color >>> 0;
        mask = mask >>> 0;
        copySource = !!copySource;
        release || somewhatImplemented("public flash.display.BitmapData::threshold");
        return;
    };
    BitmapData.prototype.lock = function () {
        this._locked = true;
    };
    BitmapData.prototype.unlock = function (changeRect) {
        if (changeRect === void 0) { changeRect = null; }
        //changeRect = changeRect;
        this._locked = false;
    };
    BitmapData.prototype.histogram = function (hRect) {
        if (hRect === void 0) { hRect = null; }
        hRect = hRect;
        release || notImplemented("public flash.display.BitmapData::histogram");
        return;
    };
    BitmapData.prototype.encode = function (rect, compressor, byteArray) {
        if (byteArray === void 0) { byteArray = null; }
        rect = rect;
        compressor = compressor;
        byteArray = byteArray;
        release || notImplemented("public flash.display.BitmapData::encode");
        return;
    };
    /**
     * Ensures that we have the most up-to-date version of the bitmap data. If a call to
     * |BitmpaData.draw| was made since the last time this method was called, then we need to send
     * a synchronous message to the GFX remote requesting the latest image data.
     *
     * Here we also normalize the image format to |ImageType.StraightAlphaRGBA|. We only need the
     * normalized pixel data for pixel operations, so we defer image decoding as late as possible.
     */
    BitmapData.prototype._ensureBitmapData = function () {
        if (this._isRemoteDirty) {
            var data = this.sec.player.requestBitmapData(this);
            this._setData(data.getBytes(), ImageType.StraightAlphaRGBA);
            this._isRemoteDirty = false;
            this._isDirty = false;
            this._solidFillColorPBGRA = null;
        }
        release || assert(!(this._type === ImageType.JPEG || this._type === ImageType.PNG ||
            this._type === ImageType.GIF));
        if (this._type !== ImageType.PremultipliedAlphaARGB) {
            ColorUtilitites.convertImage(this._type, ImageType.PremultipliedAlphaARGB, this._view, this._view);
            this._type = ImageType.PremultipliedAlphaARGB;
            this._solidFillColorPBGRA = null;
        }
        release || assert(this._data);
        release || assert(this._dataBuffer);
        release || assert(this._view);
    };
    BitmapData.MAXIMUM_WIDTH = 8191;
    BitmapData.MAXIMUM_HEIGHT = 8191;
    BitmapData.MAXIMUM_DIMENSION = 16777215;
    return BitmapData;
}(ASObject));
export { BitmapData };
var BitmapSymbol = /** @class */ (function (_super) {
    __extends(BitmapSymbol, _super);
    function BitmapSymbol(data, sec) {
        var _this = _super.call(this, data, BitmapData.axClass, false) || this;
        _this.ready = false;
        return _this;
    }
    BitmapSymbol.FromData = function (data, loaderInfo) {
        var symbol = new BitmapSymbol(data, loaderInfo.sec);
        // For non-decoded images, we don't yet have dimensions.
        symbol.width = data.width || -1;
        symbol.height = data.height || -1;
        symbol.syncId = DisplayObject.axClass.getNextSyncID();
        symbol.data = data.data;
        switch (data.mimeType) {
            case "application/octet-stream":
                symbol.type = data.dataType;
                symbol.ready = true;
                break;
            case "image/jpeg":
                symbol.type = ImageType.JPEG;
                break;
            case "image/png":
                symbol.type = ImageType.PNG;
                break;
            case "image/gif":
                symbol.type = ImageType.GIF;
                break;
            default:
                notImplemented(data.mimeType);
        }
        return symbol;
    };
    BitmapSymbol.prototype.getSharedInstance = function () {
        return this.sharedInstance || this.createSharedInstance();
    };
    BitmapSymbol.prototype.createSharedInstance = function () {
        release || assert(this.ready);
        return this.sharedInstance = constructClassFromSymbol(this, this.symbolClass);
    };
    Object.defineProperty(BitmapSymbol.prototype, "resolveAssetCallback", {
        get: function () {
            return this._unboundResolveAssetCallback.bind(this);
        },
        enumerable: true,
        configurable: true
    });
    BitmapSymbol.prototype._unboundResolveAssetCallback = function (data) {
        release || assert(!this.ready);
        this.ready = true;
        if (!data) {
            release || error("Error while decoding image");
            return;
        }
        release || assert(data.width);
        release || assert(data.height);
        this.width = data.width;
        this.height = data.height;
    };
    return BitmapSymbol;
}(DisplaySymbol));
export { BitmapSymbol };
