import { __extends } from "tslib";
import { WeakList, Bounds, isInteger, IndentingWriter } from "../../base/utilities";
import { EventDispatcher } from "../events/EventDispatcher";
import { Event } from "../events/Event";
import { checkNullParameter } from "../../avm2/run/checkNullParameter";
import { assert, axCoerceString, isNullOrUndefined } from "@awayjs/graphics";
import { release, somewhatImplemented, notImplemented } from "../../base/utilities/Debug";
import { Matrix } from "../geom/Matrix";
import { Transform } from "../geom/Transform";
import { ColorTransform } from "../geom/ColorTransform";
import { Rectangle } from "../geom/Rectangle";
import { Errors } from "../../avm2/errors";
import { StringUtilities } from "../../base/utilities/StringUtilities";
import { BitmapData } from "./BitmapData";
//import { Bitmap } from "./Bitmap";
import { constructClassFromSymbol } from "../../avm2/constructClassFromSymbol";
import { Stage } from "./Stage";
import { MovieClip } from "./MovieClip";
import { BlendMode } from "./BlendMode";
import { DisplayObjectContainer } from "./DisplayObjectContainer";
import { Graphics } from "./Graphics";
import { BitmapFilter } from "../filters/BitmapFilter";
import { DropShadowFilter } from "../filters/DropShadowFilter";
import { BlurFilter } from "../filters/BlurFilter";
import { BevelFilter } from "../filters/BevelFilter";
import { GlowFilter } from "../filters/GlowFilter";
import { GradientGlowFilter } from "../filters/GradientGlowFilter";
import { ConvolutionFilter } from "../filters/ConvolutionFilter";
import { ColorMatrixFilter } from "../filters/ColorMatrixFilter";
import { GradientBevelFilter } from "../filters/GradientBevelFilter";
import { ShapeSymbol } from "./Shape";
import { TextSymbol } from "../text/TextField";
import { StaticText } from "../text/StaticText";
import { InteractiveObject } from "./InteractiveObject";
import { Mouse } from "../ui/Mouse";
import { enterTimeline, leaveTimeline } from "../../avm2/module";
var displayObjectSyncID = 0;
var DisplayObject = /** @class */ (function (_super) {
    __extends(DisplayObject, _super);
    function DisplayObject() {
        var _this = _super.call(this) || this;
        if (!_this._fieldsInitialized) {
            _this._initializeFields();
        }
        _this._addReference();
        _this._setFlags(256 /* Constructed */);
        return _this;
    }
    /**
     * Every displayObject is assigned an unique integer ID.
     */
    DisplayObject.getNextSyncID = function () {
        return displayObjectSyncID++;
    };
    // Called whenever the class is initialized.
    DisplayObject.classInitializer = function () {
        this.reset();
    };
    DisplayObject.reset = function () {
        this._advancableInstances = new WeakList();
    };
    /**
     * Creates a new display object from a symbol and initializes its animated display properties.
     * Calling its constructor is optional at this point, since that can happen in a later frame
     * phase.
     */
    DisplayObject.prototype.createAnimatedDisplayObject = function (symbol, placeObjectTag, callConstructor) {
        var symbolClass = symbol.symbolClass;
        var bitmapDataClass = BitmapData.axClass;
        if (symbolClass === bitmapDataClass ||
            bitmapDataClass.dPrototype.isPrototypeOf(symbolClass.dPrototype)) {
            console.log("80pro todo circular dependency in DisplayObject <-> Bitmap");
            //symbolClass = Bitmap.axClass;
        }
        var instance = constructClassFromSymbol(symbol, symbolClass);
        if (placeObjectTag.flags & 32 /* HasName */) {
            instance._name = placeObjectTag.name;
        }
        instance._setFlags(4096 /* AnimatedByTimeline */);
        instance._animate(placeObjectTag);
        if (callConstructor) {
            instance.axInitializer();
        }
        return instance;
    };
    /**
     * Runs one full turn of the frame events cycle.
     *
     * Frame navigation methods on MovieClip can trigger nested frame events cycles. These nested
     * cycles do everything the outermost cycle does, except for broadcasting the ENTER_FRAME
     * event.
     *
     * If runScripts is true, no events are dispatched and Movieclip frame scripts are run. This
     * is true for nested cycles, too. (We keep static state for that.)
     */
    DisplayObject.performFrameNavigation = function (mainLoop, runScripts) {
        if (mainLoop) {
            this._runScripts = runScripts;
        }
        else {
            runScripts = this._runScripts;
        }
        release || assert(this._advancableInstances.length < 1024 * 16, "Too many advancable instances.");
        // Step 1: Remove timeline objects that don't exist on new frame, update existing ones with
        // new properties, and declare, but not create, new ones, update numChildren.
        // NOTE: the Order Of Operations senocular article is wrong on this: timeline objects are
        // removed from stage at the beginning of a frame, just as new objects are declared at that
        // point.
        // Also, changed properties of existing objects are updated here instead of during frame
        // construction after ENTER_FRAME.
        // Thus, all these can be done together.
        enterTimeline("DisplayObject.InitFrame");
        this._advancableInstances.forEach(function (value) {
            value._initFrame(mainLoop);
        });
        leaveTimeline();
        // Step 2: Dispatch ENTER_FRAME, only called in outermost invocation.
        enterTimeline("DisplayObject.EnterFrame");
        if (mainLoop && runScripts) {
            this._broadcastFrameEvent(Event.ENTER_FRAME);
        }
        leaveTimeline();
        // Step 3: Create new timeline objects.
        enterTimeline("DisplayObject.ConstructFrame");
        this._advancableInstances.forEach(function (value) {
            value._constructFrame();
        });
        leaveTimeline();
        // Step 4: Dispatch FRAME_CONSTRUCTED.
        if (runScripts) {
            enterTimeline("DisplayObject.FrameConstructed");
            this._broadcastFrameEvent(Event.FRAME_CONSTRUCTED);
            leaveTimeline();
            // Step 5: Run frame scripts
            // Flash seems to enqueue all frame scripts recursively, starting at the root of each
            // independent object graph. That can be the stage or a container that isn't itself on
            // stage, but has (grand-)children.
            // The order in which these independent graphs are processed seems not to follow a
            // specific system: in some testing scenarios all independent graphs are processes before
            // the stage, in others the first-created such graph is processes *after* the stage, all
            // others before the stage. There might be other permutations of this, but it seems
            // doubtful anybody could reasonably rely on the exact details of all this.
            // Of course, nothing guarantees that there isn't content that accidentally does, so it'd
            // be nice to eventually get this right.
            enterTimeline("DisplayObject.EnqueueFrameScripts");
            var displayObjectContainerClass = DisplayObjectContainer.axClass;
            this._advancableInstances.forEach(function (value) {
                var container = value;
                if (displayObjectContainerClass.axIsType(container) && !container.parent) {
                    container._enqueueFrameScripts();
                }
            });
            this._stage._enqueueFrameScripts();
            leaveTimeline();
            enterTimeline("DisplayObject.RunFrameScript");
            var movieClipClass = MovieClip.axClass;
            if (movieClipClass.frameNavigationModel === 1 /* SWF1 */) {
                movieClipClass.runAvm1FrameScripts();
            }
            else {
                movieClipClass.runFrameScripts();
            }
            leaveTimeline();
            // Step 6: Dispatch EXIT_FRAME.
            enterTimeline("DisplayObject.ExitFrame");
            this._broadcastFrameEvent(Event.EXIT_FRAME);
            leaveTimeline();
        }
        else {
            MovieClip.reset();
        }
        if (mainLoop) {
            this._runScripts = true;
        }
    };
    /**
     * Dispatches a frame event on all instances of DisplayObjects.
     */
    DisplayObject._broadcastFrameEvent = function (type) {
        var event = Event.axClass.getBroadcastInstance(type);
        EventDispatcher.axClass.broadcastEventDispatchQueue.dispatchEvent(event);
    };
    DisplayObject.prototype._initializeFields = function () {
        _super.prototype._initializeFields.call(this, this);
        this._id = DisplayObject.axClass.getNextSyncID();
        this._flags = 1 /* Visible */ |
            2 /* InvalidLineBounds */ |
            4 /* InvalidFillBounds */ |
            32 /* InvalidConcatenatedMatrix */ |
            64 /* InvalidInvertedConcatenatedMatrix */ |
            536870912 /* DirtyDescendents */;
        this._dirtyFlags = 4 /* DirtyGraphics */ |
            1 /* DirtyMatrix */ |
            64 /* DirtyColorTransform */ |
            128 /* DirtyMask */ |
            256 /* DirtyClipDepth */ |
            512 /* DirtyMiscellaneousProperties */;
        this._root = null;
        this._stage = null;
        this._setInitialName();
        this._parent = null;
        this._mask = null;
        this._z = 0;
        this._scaleX = 1;
        this._scaleY = 1;
        this._skewX = 0;
        this._skewY = 0;
        this._scaleZ = 1;
        this._rotation = 0;
        this._rotationX = 0;
        this._rotationY = 0;
        this._rotationZ = 0;
        this._width = 0;
        this._height = 0;
        this._opaqueBackground = null;
        this._scrollRect = null;
        this._filters = null;
        this._blendMode = BlendMode.NORMAL;
        // No need to take ownership: scale9Grid is never changed.
        this._scale9Grid = this._symbol ? this._symbol.scale9Grid : null;
        this._loaderInfo = null;
        this._accessibilityProperties = null;
        this._fillBounds = new Bounds(0, 0, 0, 0);
        this._lineBounds = new Bounds(0, 0, 0, 0);
        this._clipDepth = -1;
        this._concatenatedMatrix = new Matrix();
        this._invertedConcatenatedMatrix = new Matrix();
        this._matrix = new Matrix();
        this._invertedMatrix = new Matrix();
        this._matrix3D = null;
        this._perspectiveProjectionFOV = 55 /* FOV */;
        this._perspectiveProjectionCenterX = 250 /* CenterX */;
        this._perspectiveProjectionCenterY = 250 /* CenterY */;
        this._colorTransform = new ColorTransform();
        this._concatenatedColorTransform = new ColorTransform();
        this._depth = -1;
        this._ratio = 0;
        this._index = -1;
        this._maskedObject = null;
        this._mouseOver = false;
        this._mouseDown = false;
        this._graphics = null;
        this._children = null;
        this._referenceCount = 0;
    };
    /**
     * Sets the object's initial name to adhere to the 'instanceN' naming scheme.
     */
    DisplayObject.prototype._setInitialName = function () {
        this._name = 'instance' + (DisplayObject.axClass._instanceID++);
    };
    DisplayObject.prototype._setParent = function (parent, depth) {
        var oldParent = this._parent;
        release || assert(parent !== this);
        this._parent = parent;
        this._setDepth(depth);
        if (parent) {
            this._addReference();
            var bubblingFlags = 0 /* None */;
            if (this._hasFlags(8192 /* HasFrameScriptPending */)) {
                bubblingFlags |= 16384 /* ContainsFrameScriptPendingChildren */;
            }
            if (this._hasAnyDirtyFlags(1023 /* Dirty */)) {
                bubblingFlags |= 536870912 /* DirtyDescendents */;
            }
            if (this._hasAnyFlags(536920064 /* Bubbling */)) {
                bubblingFlags |= this._flags & 536920064 /* Bubbling */;
            }
            if (bubblingFlags) {
                parent._propagateFlagsUp(bubblingFlags);
            }
        }
        if (oldParent) {
            this._removeReference();
        }
    };
    DisplayObject.prototype._setDepth = function (value) {
        if (value > -1) {
            this._setFlags(2048 /* OwnedByTimeline */);
        }
        else {
            this._removeFlags(2048 /* OwnedByTimeline */);
        }
        this._depth = value;
    };
    DisplayObject.prototype._setFillAndLineBoundsFromWidthAndHeight = function (width, height) {
        this._fillBounds.width = width;
        this._fillBounds.height = height;
        this._lineBounds.width = width;
        this._lineBounds.height = height;
        this._removeFlags(2 /* InvalidLineBounds */ | 4 /* InvalidFillBounds */);
        this._invalidateParentFillAndLineBounds(true, true);
    };
    DisplayObject.prototype._setFillAndLineBoundsFromSymbol = function (symbol) {
        release || assert(symbol.fillBounds || symbol.lineBounds, "Fill or Line bounds are not defined in the symbol.");
        if (symbol.fillBounds) {
            this._fillBounds.copyFrom(symbol.fillBounds);
            this._removeFlags(4 /* InvalidFillBounds */);
        }
        if (symbol.lineBounds) {
            this._lineBounds.copyFrom(symbol.lineBounds);
            this._removeFlags(2 /* InvalidLineBounds */);
        }
        this._invalidateParentFillAndLineBounds(!!symbol.fillBounds, !!symbol.lineBounds);
    };
    DisplayObject.prototype._setFlags = function (flags) {
        this._flags |= flags;
    };
    /**
     * Use this to set dirty flags so that we can also propagate the dirty child bit.
     */
    DisplayObject.prototype._setDirtyFlags = function (flags) {
        this._dirtyFlags |= flags;
        if (this._parent) {
            // Notify parent that it has a dirty descendent.
            this._parent._propagateFlagsUp(536870912 /* DirtyDescendents */);
        }
    };
    DisplayObject.prototype._removeDirtyFlags = function (flags) {
        this._dirtyFlags &= ~flags;
    };
    DisplayObject.prototype._hasDirtyFlags = function (flags) {
        return (this._dirtyFlags & flags) === flags;
    };
    DisplayObject.prototype._hasAnyDirtyFlags = function (flags) {
        return !!(this._dirtyFlags & flags);
    };
    DisplayObject.prototype._toggleFlags = function (flags, on) {
        if (on) {
            this._flags |= flags;
        }
        else {
            this._flags &= ~flags;
        }
    };
    DisplayObject.prototype._removeFlags = function (flags) {
        this._flags &= ~flags;
    };
    DisplayObject.prototype._hasFlags = function (flags) {
        return (this._flags & flags) === flags;
    };
    DisplayObject.prototype._hasAnyFlags = function (flags) {
        return !!(this._flags & flags);
    };
    /**
     * Propagates flags up the display list. Propagation stops if all flags are already set.
     */
    DisplayObject.prototype._propagateFlagsUp = function (flags) {
        if (this._hasFlags(flags)) {
            return;
        }
        this._setFlags(flags);
        var parent = this._parent;
        if (parent) {
            parent._propagateFlagsUp(flags);
        }
    };
    /**
     * Propagates flags down the display list. Non-containers just set the flags on themselves.
     *
     * Overridden in DisplayObjectContainer.
     */
    DisplayObject.prototype._propagateFlagsDown = function (flags) {
        this._setFlags(flags);
    };
    /**
     * Finds the nearest ancestor with a given set of flags that are either turned on or off.
     */
    DisplayObject.prototype._findNearestAncestor = function (flags, on) {
        var node = this;
        while (node) {
            if (node._hasFlags(flags) === on) {
                return node;
            }
            node = node._parent;
        }
        return null;
    };
    DisplayObject.prototype._findFurthestAncestorOrSelf = function () {
        var node = this;
        while (node) {
            if (!node._parent) {
                return node;
            }
            node = node._parent;
        }
    };
    /**
     * Tests if this display object is an ancestor of the specified display object.
     */
    DisplayObject.prototype._isAncestor = function (child) {
        var node = child;
        while (node) {
            if (node === this) {
                return true;
            }
            node = node._parent;
        }
        return false;
    };
    /**
     * Clamps the rotation value to the range (-180, 180).
     */
    DisplayObject._clampRotation = function (value) {
        value %= 360;
        if (value > 180) {
            value -= 360;
        }
        else if (value < -180) {
            value += 360;
        }
        return value;
    };
    /**
     * Return's a list of ancestors excluding the |last|, the return list is reused.
     */
    DisplayObject._getAncestors = function (node, last) {
        var path = DisplayObject._path;
        path.length = 0;
        while (node && node !== last) {
            path.push(node);
            node = node._parent;
        }
        release || assert(node === last, "Last ancestor is not an ancestor.");
        return path;
    };
    /**
     * Computes the combined transformation matrixes of this display object and all of its parents.
     * It is not the same as |transform.concatenatedMatrix|, the latter also includes the screen
     * space matrix.
     */
    DisplayObject.prototype._getConcatenatedMatrix = function () {
        if (this._hasFlags(32 /* InvalidConcatenatedMatrix */)) {
            if (this._parent) {
                this._parent._getConcatenatedMatrix().preMultiplyInto(this._getMatrix(), this._concatenatedMatrix);
            }
            else {
                this._concatenatedMatrix.copyFrom(this._getMatrix());
            }
            this._removeFlags(32 /* InvalidConcatenatedMatrix */);
        }
        return this._concatenatedMatrix;
    };
    DisplayObject.prototype._getInvertedConcatenatedMatrix = function () {
        if (this._hasFlags(64 /* InvalidInvertedConcatenatedMatrix */)) {
            this._getConcatenatedMatrix().invertInto(this._invertedConcatenatedMatrix);
            this._removeFlags(64 /* InvalidInvertedConcatenatedMatrix */);
        }
        return this._invertedConcatenatedMatrix;
    };
    DisplayObject.prototype._setMatrix = function (matrix, toTwips) {
        if (!toTwips && this._matrix.equals(matrix)) {
            // No need to dirty the matrix if it's equal to the current matrix.
            return;
        }
        var m = this._matrix;
        m.copyFrom(matrix);
        if (toTwips) {
            m.toTwipsInPlace();
        }
        this._scaleX = m.getScaleX();
        this._scaleY = m.getScaleY();
        this._skewX = matrix.getSkewX();
        this._skewY = matrix.getSkewY();
        this._rotation = DisplayObject._clampRotation(this._skewY * 180 / Math.PI);
        this._removeFlags(8 /* InvalidMatrix */);
        this._setFlags(16 /* InvalidInvertedMatrix */);
        this._setDirtyFlags(1 /* DirtyMatrix */);
        this._invalidatePosition();
    };
    /**
     * Returns an updated matrix if the current one is invalid.
     */
    DisplayObject.prototype._getMatrix = function () {
        if (this._hasFlags(8 /* InvalidMatrix */)) {
            this._matrix.updateScaleAndRotation(this._scaleX, this._scaleY, this._skewX, this._skewY);
            this._removeFlags(8 /* InvalidMatrix */);
        }
        return this._matrix;
    };
    DisplayObject.prototype._getInvertedMatrix = function () {
        if (this._hasFlags(16 /* InvalidInvertedMatrix */)) {
            this._getMatrix().invertInto(this._invertedMatrix);
            this._removeFlags(16 /* InvalidInvertedMatrix */);
        }
        return this._invertedMatrix;
    };
    /**
     * Computes the combined transformation color matrixes of this display object and all of its
     * ancestors.
     */
    DisplayObject.prototype._getConcatenatedColorTransform = function () {
        if (!this.stage) {
            return this._colorTransform.clone();
        }
        // Compute the concatenated color transforms for this node and all of its ancestors.
        if (this._hasFlags(128 /* InvalidConcatenatedColorTransform */)) {
            var ancestor = this._findNearestAncestor(128 /* InvalidConcatenatedColorTransform */, false);
            var path = DisplayObject._getAncestors(this, ancestor);
            var i = path.length - 1;
            if (Stage.axIsType(path[i])) {
                i--;
            }
            var m = ancestor && !Stage.axIsType(ancestor) ?
                ancestor._concatenatedColorTransform.clone() :
                new ColorTransform();
            while (i >= 0) {
                ancestor = path[i--];
                release || assert(ancestor._hasFlags(128 /* InvalidConcatenatedColorTransform */));
                m.preMultiply(ancestor._colorTransform);
                m.convertToFixedPoint();
                ancestor._concatenatedColorTransform.copyFrom(m);
                ancestor._removeFlags(128 /* InvalidConcatenatedColorTransform */);
            }
        }
        return this._concatenatedColorTransform;
    };
    DisplayObject.prototype._setColorTransform = function (colorTransform) {
        if (this._colorTransform.equals(colorTransform)) {
            return;
        }
        this._colorTransform.copyFrom(colorTransform);
        this._colorTransform.convertToFixedPoint();
        this._propagateFlagsDown(128 /* InvalidConcatenatedColorTransform */);
        this._setDirtyFlags(64 /* DirtyColorTransform */);
    };
    /**
     * Invalidates the fill- and lineBounds of this display object along with all of its ancestors.
     */
    DisplayObject.prototype._invalidateFillAndLineBounds = function (fill, line) {
        /* TODO: We should only propagate this bit if the bounds are actually changed. We can do the
          * bounds computation eagerly if the number of children is low. If there are no changes in the
          * bounds we don't need to propagate the bit. */
        this._propagateFlagsUp((line ? 2 /* InvalidLineBounds */ : 0) |
            (fill ? 4 /* InvalidFillBounds */ : 0));
    };
    DisplayObject.prototype._invalidateParentFillAndLineBounds = function (fill, line) {
        if (this._parent) {
            this._parent._invalidateFillAndLineBounds(fill, line);
        }
    };
    /**
     * Computes the bounding box for all of this display object's content, its graphics and all of
     * its children.
     */
    DisplayObject.prototype._getContentBounds = function (includeStrokes) {
        if (includeStrokes === void 0) { includeStrokes = true; }
        // Tobias: What about filters?
        var invalidFlag;
        var bounds;
        if (includeStrokes) {
            invalidFlag = 2 /* InvalidLineBounds */;
            bounds = this._lineBounds;
        }
        else {
            invalidFlag = 4 /* InvalidFillBounds */;
            bounds = this._fillBounds;
        }
        if (this._hasFlags(invalidFlag)) {
            var graphics = this._getGraphics();
            if (graphics) {
                bounds.copyFrom(graphics._getContentBounds(includeStrokes));
            }
            else {
                bounds.setToSentinels();
            }
            this._getChildBounds(bounds, includeStrokes);
            this._removeFlags(invalidFlag);
        }
        return bounds;
    };
    /**
     * Empty base case: DisplayObject cannot have children, but several distinct subclasses can.
     * Overridden in DisplayObjectContainer, SimpleButton, and AVM1Movie.
     */
    DisplayObject.prototype._getChildBounds = function (bounds, includeStrokes) {
        // TSLint thinks empty methods are uncool. I think TSLint is uncool.
    };
    /**
     * Gets the bounds of this display object relative to another coordinate space. The
     * transformation matrix from the local coordinate space to the target coordinate space is
     * computed using:
     *
     *   this.concatenatedMatrix * inverse(target.concatenatedMatrix)
     *
     * If the |targetCoordinateSpace| is |null| then assume the identity coordinate space.
     */
    DisplayObject.prototype._getTransformedBounds = function (targetCoordinateSpace, includeStroke) {
        var bounds = this._getContentBounds(includeStroke).clone();
        if (targetCoordinateSpace === this || bounds.isEmpty()) {
            return bounds;
        }
        var m;
        if (targetCoordinateSpace) {
            m = Matrix.axClass.TEMP_MATRIX;
            var invertedTargetMatrix = targetCoordinateSpace._getInvertedConcatenatedMatrix();
            invertedTargetMatrix.preMultiplyInto(this._getConcatenatedMatrix(), m);
        }
        else {
            m = this._getConcatenatedMatrix();
        }
        m.transformBounds(bounds);
        return bounds;
    };
    /**
     * Detaches this object from being animated by the timeline. This happens whenever a display
     * property of this object is changed by user code.
     */
    DisplayObject.prototype._stopTimelineAnimation = function () {
        this._removeFlags(4096 /* AnimatedByTimeline */);
    };
    /**
     * Marks this object as having its matrix changed.
     *
     * Propagates flags both up- and (via invalidatePosition) downwards, so is quite costly.
     * TODO: check if we can usefully combine all upwards-propagated flags here.
     */
    DisplayObject.prototype._invalidateMatrix = function () {
        this._setDirtyFlags(1 /* DirtyMatrix */);
        this._setFlags(8 /* InvalidMatrix */ | 16 /* InvalidInvertedMatrix */);
        this._invalidatePosition();
    };
    /**
     * Marks this object as having been moved in its parent display object.
     */
    DisplayObject.prototype._invalidatePosition = function () {
        this._propagateFlagsDown(32 /* InvalidConcatenatedMatrix */ |
            64 /* InvalidInvertedConcatenatedMatrix */);
        this._invalidateParentFillAndLineBounds(true, true);
    };
    /**
     * Animates this object's display properties.
     */
    DisplayObject.prototype._animate = function (placeObjectTag) {
        release || assert(this._hasFlags(4096 /* AnimatedByTimeline */));
        var reset = !(placeObjectTag.flags & 1 /* Move */) &&
            placeObjectTag.flags & 2 /* HasCharacter */;
        var matrixClass = Matrix.axClass;
        if (placeObjectTag.flags & 4 /* HasMatrix */) {
            matrixClass.TEMP_MATRIX.copyFromUntyped(placeObjectTag.matrix);
            this._setMatrix(matrixClass.TEMP_MATRIX, false);
        }
        else if (reset) {
            this._setMatrix(matrixClass.FROZEN_IDENTITY_MATRIX, false);
        }
        var colorTransformClass = ColorTransform.axClass;
        if (placeObjectTag.flags & 8 /* HasColorTransform */) {
            colorTransformClass.TEMP_COLOR_TRANSFORM.copyFromUntyped(placeObjectTag.cxform);
            this._setColorTransform(colorTransformClass.TEMP_COLOR_TRANSFORM);
        }
        else if (reset) {
            this._setColorTransform(colorTransformClass.FROZEN_IDENTITY_COLOR_TRANSFORM);
        }
        if (placeObjectTag.flags & 16 /* HasRatio */ || reset) {
            var ratio = placeObjectTag.ratio | 0;
            if (ratio !== this._ratio) {
                release || assert(ratio >= 0 && ratio <= 0xffff);
                this._ratio = ratio;
                this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
            }
        }
        if (placeObjectTag.flags & 64 /* HasClipDepth */ || reset) {
            var clipDepth = placeObjectTag.clipDepth === undefined ? -1 : placeObjectTag.clipDepth;
            if (clipDepth !== this._clipDepth) {
                this._clipDepth = clipDepth;
                this._setDirtyFlags(256 /* DirtyClipDepth */);
            }
        }
        if (placeObjectTag.flags & 256 /* HasFilterList */) {
            var filters = [];
            var swfFilters = placeObjectTag.filters;
            for (var i = 0; i < swfFilters.length; i++) {
                var obj = swfFilters[i];
                var filter;
                switch (obj.type) {
                    case 0:
                        filter = DropShadowFilter.axClass.FromUntyped(obj);
                        break;
                    case 1:
                        filter = BlurFilter.axClass.FromUntyped(obj);
                        break;
                    case 2:
                        filter = GlowFilter.axClass.FromUntyped(obj);
                        break;
                    case 3:
                        filter = BevelFilter.axClass.FromUntyped(obj);
                        break;
                    case 4:
                        filter = GradientGlowFilter.axClass.FromUntyped(obj);
                        break;
                    case 5:
                        filter = ConvolutionFilter.axClass.FromUntyped(obj);
                        break;
                    case 6:
                        filter = ColorMatrixFilter.axClass.FromUntyped(obj);
                        break;
                    case 7:
                        filter = GradientBevelFilter.axClass.FromUntyped(obj);
                        break;
                    default:
                        release || assert(filter, "Unknown filter type.");
                }
                filters.push(filter);
            }
            this._filters = filters;
            this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
        }
        else if (reset && this._filters) {
            this._filters = null;
            this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
        }
        if (placeObjectTag.flags & 512 /* HasBlendMode */ || reset) {
            var blendMode = BlendMode.fromNumber(placeObjectTag.blendMode === undefined ?
                1 : placeObjectTag.blendMode);
            if (blendMode !== this._blendMode) {
                this._blendMode = blendMode;
                this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
            }
        }
        if (placeObjectTag.flags & 1024 /* HasCacheAsBitmap */ || reset) {
            var cacheAsBitmap = placeObjectTag.bmpCache > 0;
            if (cacheAsBitmap !== this._hasFlags(65536 /* CacheAsBitmap */)) {
                this._toggleFlags(65536 /* CacheAsBitmap */, cacheAsBitmap);
                this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
            }
        }
        if (placeObjectTag.flags & 8192 /* HasVisible */ || reset) {
            var visible = placeObjectTag.visibility === undefined || placeObjectTag.visibility;
            if (visible !== this._hasFlags(1 /* Visible */)) {
                this._toggleFlags(1 /* Visible */, visible);
                this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
            }
        }
    };
    /**
     * Dispatches an event on this object and all its descendants.
     */
    DisplayObject.prototype._propagateEvent = function (event) {
        this.visit(function (node) {
            node.dispatchEvent(event);
            return 0 /* Continue */;
        }, 0 /* None */);
    };
    Object.defineProperty(DisplayObject.prototype, "x", {
        get: function () {
            return this._getX();
        },
        set: function (value) {
            value = (value * 20) | 0;
            this._stopTimelineAnimation();
            if (this._canHaveTextContent()) {
                var bounds = this._getContentBounds();
                value -= bounds.xMin;
            }
            if (value === this._matrix.tx) {
                return;
            }
            this._matrix.tx = value;
            this._invertedMatrix.tx = -value;
            this._invalidatePosition();
            this._setDirtyFlags(1 /* DirtyMatrix */);
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype._getX = function () {
        var value = this._matrix.tx;
        if (this._canHaveTextContent()) {
            var bounds = this._getContentBounds();
            value += bounds.xMin;
        }
        return value / 20;
    };
    Object.defineProperty(DisplayObject.prototype, "y", {
        get: function () {
            return this._getY();
        },
        set: function (value) {
            value = (value * 20) | 0;
            this._stopTimelineAnimation();
            if (this._canHaveTextContent()) {
                var bounds = this._getContentBounds();
                value -= bounds.yMin;
            }
            if (value === this._matrix.ty) {
                return;
            }
            this._matrix.ty = value;
            this._invertedMatrix.ty = -value;
            this._invalidatePosition();
            this._setDirtyFlags(1 /* DirtyMatrix */);
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype._getY = function () {
        var value = this._matrix.ty;
        if (this._canHaveTextContent()) {
            var bounds = this._getContentBounds();
            value += bounds.yMin;
        }
        return value / 20;
    };
    Object.defineProperty(DisplayObject.prototype, "scaleX", {
        /**
         * In Flash player, this always returns a positive number for some reason. This however, is not
         * the case for scaleY.
         */
        get: function () {
            return Math.abs(this._scaleX);
        },
        set: function (value) {
            value = +value;
            this._stopTimelineAnimation();
            if (value === this._scaleX) {
                return;
            }
            this._scaleX = value;
            this._invalidateMatrix();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "scaleY", {
        get: function () {
            return this._scaleY;
        },
        set: function (value) {
            value = +value;
            this._stopTimelineAnimation();
            if (value === this._scaleY) {
                return;
            }
            this._scaleY = value;
            this._invalidateMatrix();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "scaleZ", {
        get: function () {
            return this._scaleZ;
        },
        set: function (value) {
            value = +value;
            release || somewhatImplemented("public DisplayObject::set scaleZ");
            this._scaleZ = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "rotation", {
        get: function () {
            return this._rotation;
        },
        set: function (value) {
            value = +value;
            this._stopTimelineAnimation();
            value = DisplayObject._clampRotation(value);
            if (value === this._rotation) {
                return;
            }
            var delta = value - this._rotation;
            var angle = delta / 180 * Math.PI;
            this._skewX += angle;
            this._skewY += angle;
            this._rotation = value;
            this._invalidateMatrix();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "rotationX", {
        get: function () {
            return this._rotationX;
        },
        set: function (value) {
            value = +value;
            release || somewhatImplemented("public DisplayObject::set rotationX");
            this._rotationZ = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "rotationY", {
        get: function () {
            return this._rotationY;
        },
        set: function (value) {
            value = +value;
            release || somewhatImplemented("public DisplayObject::set rotationY");
            this._rotationY = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "rotationZ", {
        get: function () {
            return this._rotationZ;
        },
        set: function (value) {
            value = +value;
            release || somewhatImplemented("public DisplayObject::set rotationZ");
            this._rotationZ = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "width", {
        /**
         * The width of this display object in its parent coordinate space.
         */
        get: function () {
            return this._getWidth();
        },
        /**
         * Attempts to change the width of this display object by changing its scaleX / scaleY
         * properties. The scaleX property is set to the specified |width| value / baseWidth
         * of the object in its parent cooridnate space with rotation applied.
         */
        set: function (value) {
            this._setWidth(value);
        },
        enumerable: true,
        configurable: true
    });
    // `get width` is overriden in `Stage` and has to be able to call this.
    DisplayObject.prototype._getWidth = function () {
        return this._getTransformedBounds(this._parent, true).width / 20;
    };
    // `set width` is overriden in `Stage` and has to be able to call this.
    DisplayObject.prototype._setWidth = function (value) {
        value = (value * 20) | 0;
        this._stopTimelineAnimation();
        if (value < 0) {
            return;
        }
        var contentBounds = this._getContentBounds(true);
        if (this._canHaveTextContent()) {
            var bounds = this._getContentBounds();
            this._setFillAndLineBoundsFromWidthAndHeight(value, contentBounds.height);
            return;
        }
        var bounds = this._getTransformedBounds(this._parent, true);
        var angle = this._rotation / 180 * Math.PI;
        var baseWidth = contentBounds.getBaseWidth(angle);
        if (!baseWidth) {
            return;
        }
        var baseHeight = contentBounds.getBaseHeight(angle);
        this._scaleY = bounds.height / baseHeight;
        this._scaleX = value / baseWidth;
        this._invalidateMatrix();
    };
    Object.defineProperty(DisplayObject.prototype, "height", {
        /**
         * The height of this display object in its parent coordinate space.
         */
        get: function () {
            return this._getHeight();
        },
        /**
         * Attempts to change the height of this display object by changing its scaleY / scaleX
         * properties. The scaleY property is set to the specified |height| value / baseHeight
         * of the object in its parent cooridnate space with rotation applied.
         */
        set: function (value) {
            this._setHeight(value);
        },
        enumerable: true,
        configurable: true
    });
    // `get height` is overriden in `Stage` and has to be able to call this.
    DisplayObject.prototype._getHeight = function () {
        var bounds = this._getTransformedBounds(this._parent, true);
        return bounds.height / 20;
    };
    // `set height` is overriden in `Stage` and has to be able to call this.
    DisplayObject.prototype._setHeight = function (value) {
        value = (value * 20) | 0;
        this._stopTimelineAnimation();
        if (value < 0) {
            return;
        }
        var contentBounds = this._getContentBounds(true);
        if (this._canHaveTextContent()) {
            var bounds = this._getContentBounds();
            this._setFillAndLineBoundsFromWidthAndHeight(contentBounds.width, value);
            return;
        }
        var bounds = this._getTransformedBounds(this._parent, true);
        var angle = this._rotation / 180 * Math.PI;
        var baseHeight = contentBounds.getBaseHeight(angle);
        if (!baseHeight) {
            return;
        }
        var baseWidth = contentBounds.getBaseWidth(angle);
        this._scaleY = value / baseHeight;
        this._scaleX = bounds.width / baseWidth;
        this._invalidateMatrix();
    };
    Object.defineProperty(DisplayObject.prototype, "mask", {
        get: function () {
            return this._mask;
        },
        /**
         * Sets the mask for this display object. This does not affect the bounds.
         */
        set: function (value) {
            if (this._mask === value || value === this) {
                return;
            }
            if (value && value._maskedObject) {
                value._maskedObject.mask = null;
            }
            if (this._mask) {
                this._mask._maskedObject = null;
            }
            this._mask = value;
            if (value) {
                value._maskedObject = this;
            }
            this._setDirtyFlags(128 /* DirtyMask */);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "transform", {
        get: function () {
            return this._getTransform();
        },
        set: function (value) {
            this._stopTimelineAnimation();
            if (value.matrix3D) {
                this._matrix3D = value.matrix3D;
            }
            else {
                this._setMatrix(value.matrix, true);
            }
            this._setColorTransform(value.colorTransform);
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype._getTransform = function () {
        return new Transform(this);
    };
    DisplayObject.prototype.destroy = function () {
        this._setFlags(512 /* Destroyed */);
    };
    Object.defineProperty(DisplayObject.prototype, "root", {
        /**
         * Walks up the tree to find this display object's root. An object is classified
         * as a root if its _root property points to itself. Root objects are the Stage,
         * the main timeline object and a Loader's content.
         */
        get: function () {
            var node = this;
            do {
                if (node._root === node) {
                    return node;
                }
                node = node._parent;
            } while (node);
            return null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "stage", {
        /**
         * Walks up the tree to find this display object's stage, the first object whose
         * |_stage| property points to itself.
         */
        get: function () {
            var node = this;
            do {
                if (node._stage === node) {
                    release || assert(Stage.axIsType(node));
                    return node;
                }
                node = node._parent;
            } while (node);
            return null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "name", {
        get: function () {
            return this._name;
        },
        set: function (value) {
            checkNullParameter(value, "name", this.sec);
            if (this._hasFlags(2048 /* OwnedByTimeline */)) {
                // In AVM2, setting the name of a timline-placed DisplayObject throws.
                if (this._symbol && !this._symbol.isAVM1Object) { // fail only in AVM2
                    this.sec.throwError('IllegalOperationError', Errors.TimelineObjectNameSealedError);
                }
                return;
            }
            this._name = axCoerceString(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "parent", {
        get: function () {
            return this._parent;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "alpha", {
        get: function () {
            return this._colorTransform.alphaMultiplier;
        },
        set: function (value) {
            this._stopTimelineAnimation();
            value = +value;
            if (value === this._colorTransform.alphaMultiplier) {
                return;
            }
            this._colorTransform.alphaMultiplier = value;
            this._colorTransform.convertToFixedPoint();
            this._propagateFlagsDown(128 /* InvalidConcatenatedColorTransform */);
            this._setDirtyFlags(64 /* DirtyColorTransform */);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "blendMode", {
        get: function () {
            return this._blendMode;
        },
        set: function (value) {
            this._stopTimelineAnimation();
            value = axCoerceString(value);
            if (value === this._blendMode) {
                return;
            }
            if (BlendMode.toNumber(value) < 0) {
                this.sec.throwError("ArgumentError", Errors.InvalidEnumError, "blendMode");
            }
            this._blendMode = value;
            this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "scale9Grid", {
        get: function () {
            return this._getScale9Grid();
        },
        set: function (innerRectangle) {
            this._stopTimelineAnimation();
            this._scale9Grid = Bounds.FromRectangle(innerRectangle);
            this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype._getScale9Grid = function () {
        var rectangleClass = Rectangle.axClass;
        return this._scale9Grid ? rectangleClass.FromBounds(this._scale9Grid) : null;
    };
    Object.defineProperty(DisplayObject.prototype, "cacheAsBitmap", {
        /**
         * This is always true if a filter is applied.
         */
        get: function () {
            return this._getCacheAsBitmap();
        },
        set: function (value) {
            if (this._hasFlags(65536 /* CacheAsBitmap */)) {
                return;
            }
            this._toggleFlags(65536 /* CacheAsBitmap */, !!value);
            this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype._getCacheAsBitmap = function () {
        return (this._filters && this._filters.length > 0) || this._hasFlags(65536 /* CacheAsBitmap */);
    };
    Object.defineProperty(DisplayObject.prototype, "filters", {
        /**
         * References to the internal |_filters| array and its BitmapFilter objects are never leaked
         * outside of this class. The get/set filters accessors always return deep clones of this
         * array.
         */
        get: function () {
            return this._getFilters();
        },
        set: function (value_) {
            var value = value_ ? value_.value : null;
            if (!this._filters) {
                this._filters = [];
            }
            var changed = false;
            if (isNullOrUndefined(value)) {
                changed = this._filters.length > 0;
                this._filters.length = 0;
            }
            else {
                var bitmapFilterClass = BitmapFilter.axClass;
                this._filters = value.map(function (x) {
                    if (!bitmapFilterClass.axIsType(x)) {
                        this.sec.throwError('TypeError', Errors.ParamTypeError, '0', 'Filter');
                    }
                    return x.clone();
                });
                changed = true;
            }
            if (changed) {
                this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
            }
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype._getFilters = function () {
        var filters = this._filters ? this._filters.map(function (x) {
            return x.clone();
        }) : [];
        return this.sec.createArray(filters);
    };
    Object.defineProperty(DisplayObject.prototype, "visible", {
        get: function () {
            return this._hasFlags(1 /* Visible */);
        },
        /**
         * Marks this display object as visible / invisible. This does not affect the bounds.
         */
        set: function (value) {
            value = !!value;
            if (value === this._hasFlags(1 /* Visible */)) {
                return;
            }
            this._toggleFlags(1 /* Visible */, value);
            this._setDirtyFlags(512 /* DirtyMiscellaneousProperties */);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "z", {
        get: function () {
            return this._z;
        },
        set: function (value) {
            value = +value;
            this._z = value;
            release || somewhatImplemented("public DisplayObject::set z");
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype.getBounds = function (targetCoordinateSpace) {
        targetCoordinateSpace = targetCoordinateSpace || this;
        var rectangleClass = Rectangle.axClass;
        return rectangleClass.FromBounds(this._getTransformedBounds(targetCoordinateSpace, true));
    };
    DisplayObject.prototype.getRect = function (targetCoordinateSpace) {
        targetCoordinateSpace = targetCoordinateSpace || this;
        var rectangleClass = Rectangle.axClass;
        return rectangleClass.FromBounds(this._getTransformedBounds(targetCoordinateSpace, false));
    };
    /**
     * Converts a point from the global coordinate space into the local coordinate space.
     */
    DisplayObject.prototype.globalToLocal = function (point) {
        var m = this._getInvertedConcatenatedMatrix();
        var p = m.transformPointInPlace(point.clone().toTwips()).round();
        return p.toPixels();
    };
    /**
     * Converts a point form the local coordinate sapce into the global coordinate space.
     */
    DisplayObject.prototype.localToGlobal = function (point) {
        var m = this._getConcatenatedMatrix();
        var p = m.transformPointInPlace(point.clone().toTwips()).round();
        return p.toPixels();
    };
    DisplayObject.prototype.globalToLocal3D = function (point) {
        release || notImplemented("public DisplayObject::globalToLocal3D");
        return null;
    };
    DisplayObject.prototype.localToGlobal3D = function (point) {
        release || notImplemented("public DisplayObject::localToGlobal3D");
        return null;
    };
    DisplayObject.prototype.local3DToGlobal = function (point3d) {
        release || notImplemented("public DisplayObject::local3DToGlobal");
        return null;
    };
    /**
     * Tree visitor that lets you skip nodes or return early.
     */
    DisplayObject.prototype.visit = function (visitor, visitorFlags, displayObjectFlags) {
        if (displayObjectFlags === void 0) { displayObjectFlags = 0 /* None */; }
        var stack;
        var displayObject;
        var displayObjectContainer;
        var frontToBack = visitorFlags & 8 /* FrontToBack */;
        stack = [this];
        while (stack.length > 0) {
            displayObject = stack.pop();
            var flags = 0 /* None */;
            if (visitorFlags & 16 /* Filter */ && !displayObject._hasAnyFlags(displayObjectFlags)) {
                flags = 2 /* Skip */;
            }
            else {
                flags = visitor(displayObject);
            }
            if (flags === 0 /* Continue */) {
                var children = displayObject._children;
                if (children) {
                    var length = children.length;
                    for (var i = 0; i < length; i++) {
                        var child = children[frontToBack ? i : length - 1 - i];
                        stack.push(child);
                    }
                }
            }
            else if (flags === 1 /* Stop */) {
                return;
            }
        }
    };
    Object.defineProperty(DisplayObject.prototype, "loaderInfo", {
        /**
         * Returns the loader info for this display object's root.
         */
        get: function () {
            var root = this.root;
            if (root) {
                release || assert(root._loaderInfo, "No LoaderInfo object found on root.");
                return root._loaderInfo;
            }
            return null;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Only these objects can have graphics.
     */
    DisplayObject.prototype._canHaveGraphics = function () {
        return false;
    };
    /**
     * Gets the graphics object of this object. Shapes, MorphShapes, and Sprites override this.
     */
    DisplayObject.prototype._getGraphics = function () {
        return null;
    };
    /**
     * Only these objects can have text content.
     */
    DisplayObject.prototype._canHaveTextContent = function () {
        return false;
    };
    /**
     * Gets the text content of this object. StaticTexts and TextFields override this.
     */
    DisplayObject.prototype._getTextContent = function () {
        return null;
    };
    /**
     * Lazily construct a graphics object.
     */
    DisplayObject.prototype._ensureGraphics = function () {
        release || assert(this._canHaveGraphics());
        if (this._graphics) {
            return this._graphics;
        }
        this._graphics = new Graphics();
        this._graphics._setParent(this);
        this._invalidateFillAndLineBounds(true, true);
        this._setDirtyFlags(4 /* DirtyGraphics */);
        return this._graphics;
    };
    /**
     * Sets this object's graphics or text content. Happens when an animated Shape or StaticText
     * object is initialized from a symbol or replaced by a timeline command using the same symbol
     * as this object was initialized from.
     */
    DisplayObject.prototype._setStaticContentFromSymbol = function (symbol) {
        release || assert(!symbol.dynamic);
        if (symbol instanceof ShapeSymbol) {
            release || assert(this._canHaveGraphics());
            var newGraphics = symbol.graphics;
            if (this._graphics === newGraphics) {
                return;
            }
            this._graphics = newGraphics;
            this._setDirtyFlags(4 /* DirtyGraphics */);
        }
        else if (symbol instanceof TextSymbol) {
            release || assert(StaticText.axIsType(this));
            var newTextContent = symbol.textContent;
            if (this._textContent === newTextContent) {
                return;
            }
            this._textContent = newTextContent;
            this._setDirtyFlags(8 /* DirtyTextContent */);
        }
        this._symbol = symbol;
        this._setFillAndLineBoundsFromSymbol(symbol);
    };
    /**
     * Checks if the bounding boxes of two display objects overlap, this happens in the global
     * coordinate coordinate space.
     *
     * Two objects overlap even if one or both are not on the stage, as long as their bounds
     * in the global coordinate space overlap.
     */
    DisplayObject.prototype.hitTestObject = function (other) {
        release || assert(other && DisplayObject.axIsType(other));
        var a = this, b = other;
        var aBounds = a._getContentBounds(false).clone();
        var bBounds = b._getContentBounds(false).clone();
        a._getConcatenatedMatrix().transformBounds(aBounds);
        b._getConcatenatedMatrix().transformBounds(bBounds);
        return aBounds.intersects(bBounds);
    };
    /**
     * The |globalX| and |globalY| arguments are in global coordinates. The |shapeFlag| indicates
     * whether the hit test should be on the actual shape of the object or just its bounding box.
     *
     * Note: shapeFlag is optional, but the type coercion will do the right thing for it, so we
     * don't need to take the overhead from being explicit about that.
     */
    DisplayObject.prototype.hitTestPoint = function (globalX, globalY, shapeFlag) {
        globalX = +globalX * 20 | 0;
        globalY = +globalY * 20 | 0;
        shapeFlag = !!shapeFlag;
        var testingType = shapeFlag ?
            2 /* HitTestShape */ :
            0 /* HitTestBounds */;
        return !!this._containsGlobalPoint(globalX, globalY, testingType, null);
    };
    /**
     * Internal implementation of all point intersection checks.
     *
     * _containsPoint is used for
     *  - mouse/drop target finding
     *  - getObjectsUnderPoint
     *  - hitTestPoint
     *
     * Mouse/Drop target finding and getObjectsUnderPoint require checking against the exact shape,
     * and making sure that the checked coordinates aren't hidden through masking or clipping.
     *
     * hitTestPoint never checks for clipping, and masking only for testingType HitTestShape.
     *
     * The `objects` object is used for collecting objects for `getObjectsUnderPoint` or looking
     * for a drop target. If it is supplied, objects for which `_containsPointDirectly` is true are
     * added to it.
     *
     * Overridden in DisplayObjectContainer, Sprite and SimpleButton.
     */
    DisplayObject.prototype._containsPoint = function (globalX, globalY, localX, localY, testingType, objects) {
        var result = this._boundsAndMaskContainPoint(globalX, globalY, localX, localY, testingType);
        // We're done if either we don't have a hit, or if we're only interested in matching bounds
        // or bounds + mask. That is true for HitTestPoint without shapeFlag set.
        if (result === 0 /* None */ || testingType < 2 /* HitTestShape */) {
            return result;
        }
        var containsPoint = this._containsPointDirectly(localX, localY, globalX, globalY);
        if (containsPoint && objects) {
            if (testingType === 5 /* Drop */) {
                // For Drop, replace previous hit with current one.
                objects[0] = this;
            }
            else if (testingType === 4 /* ObjectsUnderPoint */ ||
                InteractiveObject.axIsType(this) &&
                    this._mouseEnabled) {
                // For getObjectsUnderPoint, push all direct hits, for mouse target finding
                // InteractiveObjects only.
                objects.push(this);
            }
        }
        return containsPoint ? 2 /* Shape */ : result;
    };
    DisplayObject.prototype._containsGlobalPoint = function (globalX, globalY, testingType, objects) {
        var matrix = this._getInvertedConcatenatedMatrix();
        var localX = matrix.transformX(globalX, globalY);
        var localY = matrix.transformY(globalX, globalY);
        return this._containsPoint(globalX, globalY, localX, localY, testingType, objects);
    };
    /**
     * Fast check if a point can intersect the receiver object. Returns true if
     * - the object is visible OR hit testing is performed for one of the `hitTest{Point,Object}`
     *   methods.
     * - the point is within the receiver's bounds
     * - for testingType values other than HitTestBounds, the point intersects with the a mask,
     *   if the object has one.
     *
     * Note that the callers are expected to have both local and global coordinates available
     * anyway, so _boundsAndMaskContainPoint takes both to avoid recalculating them.
     */
    DisplayObject.prototype._boundsAndMaskContainPoint = function (globalX, globalY, localX, localY, testingType) {
        if (testingType >= 1 /* HitTestBoundsAndMask */ &&
            this._hasFlags(32768 /* ContainsMorph */)) {
            // If this display object is a MorphShape or contains at least one descendant that is, then
            // bailing out too early might lead to a wrong hit test result, since the reported bounds
            // of MorphShapes are always the one of their start shapes and don't take the current morph
            // ratio into account. We have to make sure we always hit test MorphShape instances on
            // graphics level.
            return 1 /* Bounds */;
        }
        if (testingType >= 3 /* Mouse */ && !this._hasFlags(1 /* Visible */) ||
            !this._getContentBounds().contains(localX, localY)) {
            return 0 /* None */;
        }
        if (testingType === 0 /* HitTestBounds */ || !this._mask) {
            return 1 /* Bounds */;
        }
        return this._mask._containsGlobalPoint(globalX, globalY, 1 /* HitTestBoundsAndMask */, null);
    };
    /**
     * Tests if the receiver's own visual content intersects with the given point.
     * In the base implementation, this just returns false, because not all DisplayObjects can
     * ever match.
     * Overridden in Shape, MorphShape, Sprite, Bitmap, Video, and TextField.
     */
    DisplayObject.prototype._containsPointDirectly = function (localX, localY, globalX, globalY) {
        return false;
    };
    Object.defineProperty(DisplayObject.prototype, "scrollRect", {
        get: function () {
            return this._getScrollRect();
        },
        set: function (value) {
            value = value;
            this._scrollRect = value ? value.clone() : null;
            /* TODO: Figure out how to deal with the bounds and hit testing when scroll rects are applied.
              * The Flash implementation appears to be broken. */
            release || somewhatImplemented("public DisplayObject::set scrollRect");
            return;
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype._getScrollRect = function () {
        return this._scrollRect ? this._scrollRect.clone() : null;
    };
    Object.defineProperty(DisplayObject.prototype, "opaqueBackground", {
        get: function () {
            return this._opaqueBackground;
        },
        /**
         * Sets the opaque background color. By default this is |null|, which indicates that no opaque
         * color is set. Otherwise this is an unsinged number.
         */
        set: function (value) {
            release || assert(value === null || isInteger(value));
            this._opaqueBackground = value;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Returns the distance between this object and a given ancestor.
     */
    DisplayObject.prototype._getDistance = function (ancestor) {
        var d = 0;
        var node = this;
        while (node && node !== ancestor) {
            d++;
            node = node._parent;
        }
        return d;
    };
    /**
     * Finds the nearest common ancestor with a given node.
     */
    DisplayObject.prototype.findNearestCommonAncestor = function (node) {
        if (!node) {
            return null;
        }
        var ancestor = this;
        var d1 = ancestor._getDistance(null);
        var d2 = node._getDistance(null);
        while (d1 > d2) {
            ancestor = ancestor._parent;
            d1--;
        }
        while (d2 > d1) {
            node = node._parent;
            d2--;
        }
        while (ancestor !== node) {
            ancestor = ancestor._parent;
            node = node._parent;
        }
        return ancestor;
    };
    /**
     * Returns the current mouse position relative to this object.
     */
    DisplayObject.prototype._getLocalMousePosition = function () {
        var position = Mouse.axClass._currentPosition;
        return this.globalToLocal(position);
    };
    Object.defineProperty(DisplayObject.prototype, "mouseX", {
        get: function () {
            return this._getLocalMousePosition().x;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "mouseY", {
        get: function () {
            return this._getLocalMousePosition().y;
        },
        enumerable: true,
        configurable: true
    });
    DisplayObject.prototype.debugName = function (withFlags) {
        if (withFlags === void 0) { withFlags = false; }
        var name = this._id + " [" + this._depth + "]: (" + this._referenceCount + ") " + this;
        if (withFlags) {
            var flagNames = [];
            for (var i = 0; i < 32; i++) {
                if (this._hasFlags(1 << i)) {
                    flagNames.push(String([1 << i]));
                }
            }
            name += " " + flagNames.join("|");
        }
        return name;
    };
    DisplayObject.prototype.debugNameShort = function () {
        return "[" + this._depth + ":" + this._id + "]: (" + this._referenceCount + ") {" + this._flags + "} " + this;
    };
    DisplayObject.prototype.hashCode = function () {
        return this.getBounds(null).hashCode();
    };
    DisplayObject.prototype.getAncestorCount = function () {
        return 0;
    };
    DisplayObject.prototype.debugTrace = function (writer, maxDistance, name) {
        if (writer === void 0) { writer = new IndentingWriter(); }
        if (maxDistance === void 0) { maxDistance = 1024; }
        if (name === void 0) { name = ""; }
        var self = this;
        this.visit(function (node) {
            var distance = node._getDistance(self);
            if (distance > maxDistance) {
                return 2 /* Skip */;
            }
            var prefix = name + StringUtilities.multiple(" ", distance);
            writer.writeLn(prefix + node.debugNameShort() + ", bounds: " + node.getBounds(null).toString());
            return 0 /* Continue */;
        }, 0 /* None */);
    };
    DisplayObject.prototype._addReference = function () {
        this._referenceCount++;
    };
    DisplayObject.prototype._removeReference = function () {
        // TODO: Uncomment this assertion once we're sure reference counting works correctly.
        // assert (this._referenceCount > 0);
        this._referenceCount--;
        if (this._referenceCount !== 0 || !this._children) {
            return;
        }
        var children = this._children;
        for (var i = 0; i < children.length; i++) {
            children[i]._removeReference();
        }
    };
    /**
     * Returns script precedence sequence based on placeObjectTag. Creates every
     * time a new array, so it's safe to modify it.
     * @private
     */
    DisplayObject.prototype._getScriptPrecedence = function () {
        if (!this._parent) {
            return [];
        }
        var result = this._parent._getScriptPrecedence();
        if (this._placeObjectTag) {
            result.push(this._placeObjectTag.actionBlocksPrecedence);
        }
        return result;
    };
    Object.defineProperty(DisplayObject.prototype, "accessibilityProperties", {
        get: function () {
            return this._accessibilityProperties;
        },
        set: function (value) {
            // In Flash this does not do copying.
            // TODO: coerce to the correct type.
            this._accessibilityProperties = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayObject.prototype, "blendShader", {
        set: function (value /* flash.display.Shader */) {
            release || somewhatImplemented("public DisplayObject::set blendShader");
        },
        enumerable: true,
        configurable: true
    });
    /**
     * DisplayObject#name is set to an initial value of 'instanceN', where N is auto-incremented.
     * This is true for all DisplayObjects except for Stage, so it happens in an overrideable
     * method.
     */
    DisplayObject._instanceID = 1;
    // List of static symbols to link.
    DisplayObject.classSymbols = null; // [];
    // List of instance symbols to link.
    DisplayObject.instanceSymbols = null; // ["hitTestObject", "hitTestPoint"];
    DisplayObject._runScripts = true;
    /**
     * Used as a temporary array to avoid allocations.
     */
    DisplayObject._path = [];
    return DisplayObject;
}(EventDispatcher));
export { DisplayObject };
