import { __extends } from "tslib";
import { DisplayObject } from "./DisplayObject";
import { Sprite } from "./Sprite";
import { assert, release, assertUnreachable } from "../../base/utilities/Debug";
import { Scene } from "./Scene";
import { FrameLabel } from "./FrameLabel";
import { axCoerceString } from "../../avm2/run/axCoerceString";
import { Errors } from "../../avm2/errors";
import { SoundStart } from "../symbol";
import { Event } from "../events/Event";
import { MovieClipSoundStream } from "./MovieClipSoundStream";
import { constructClassFromSymbol } from "../../avm2/constructClassFromSymbol";
import { enterTimeline, leaveTimeline } from "../../avm2/module";
import { getAVM1Object } from '../../avm1/lib/AVM1Utils';
var MovieClipSoundsManager = /** @class */ (function () {
    function MovieClipSoundsManager(mc) {
        this._mc = mc;
        this._startSoundRegistrations = null;
        this._soundStream = null;
    }
    MovieClipSoundsManager.prototype.registerStartSounds = function (frameNum, soundStartInfo) {
        if (this._startSoundRegistrations === null) {
            this._startSoundRegistrations = {};
        }
        this._startSoundRegistrations[frameNum] = soundStartInfo;
    };
    MovieClipSoundsManager.prototype.initSoundStream = function (streamInfo) {
        this._soundStream = new MovieClipSoundStream(streamInfo, this._mc);
    };
    MovieClipSoundsManager.prototype.addSoundStreamBlock = function (frameNum, streamBlock) {
        if (this._soundStream) {
            this._soundStream.appendBlock(frameNum, streamBlock);
        }
    };
    MovieClipSoundsManager.prototype._startSounds = function (frameNum) {
        var starts = this._startSoundRegistrations[frameNum];
        if (starts) {
            var sounds = this._soundClips || (this._soundClips = {});
            var loaderInfo = this._mc.loaderInfo;
            for (var i = 0; i < starts.length; i++) {
                var start = starts[i];
                var symbolId = start.soundId;
                var info = start.soundInfo;
                var sound = sounds[symbolId];
                if (!sound) {
                    var symbolInfo = loaderInfo.getSymbolById(symbolId);
                    if (!symbolInfo) {
                        continue;
                    }
                    var symbolClass = symbolInfo.symbolClass;
                    var soundObj = constructClassFromSymbol(symbolInfo, symbolClass);
                    sounds[symbolId] = sound = { object: soundObj };
                }
                var stop = info.flags & 32 /* Stop */;
                if (sound.channel && stop) {
                    sound.channel.stop();
                    sound.channel = null;
                }
                if (!stop && (!sound.channel || !sound.channel.playing)) {
                    // TODO envelope, in/out point
                    var loops = info.flags & 4 /* HasLoops */ ? info.loopCount : 0;
                    sound.channel = sound.object.play(0, loops);
                }
            }
        }
    };
    MovieClipSoundsManager.prototype.syncSounds = function (frameNum) {
        if (this._startSoundRegistrations !== null) {
            this._startSounds(frameNum);
        }
        if (this._soundStream) {
            this._soundStream.playFrame(frameNum);
        }
    };
    return MovieClipSoundsManager;
}());
function compareFrameScripts(a, b) {
    if (!a.precedence) {
        return !b.precedence ? 0 : -1;
    }
    else if (!b.precedence) {
        return 1;
    }
    var i = 0;
    while (i < a.precedence.length && i < b.precedence.length &&
        a.precedence[i] === b.precedence[i]) {
        i++;
    }
    if (i >= a.precedence.length) {
        return a.precedence.length === b.precedence.length ? 0 : -1;
    }
    else {
        return i >= b.precedence.length ? 1 : a.precedence[i] - b.precedence[i];
    }
}
var MovieClip = /** @class */ (function (_super) {
    __extends(MovieClip, _super);
    function MovieClip() {
        var _this = _super.call(this) || this;
        if (!_this._fieldsInitialized) {
            _this._initializeFields();
        }
        return _this;
    }
    // Called whenever the class is initialized.
    MovieClip.classInitializer = function () {
        this.reset();
    };
    MovieClip.reset = function () {
        this.frameNavigationModel = 10 /* SWF10 */;
        this._callQueue = [];
    };
    MovieClip.runFrameScripts = function () {
        enterTimeline("MovieClip.executeFrame");
        var movieClipClass = MovieClip.axClass;
        var displayObjectClass = DisplayObject.axClass;
        var queue = movieClipClass._callQueue;
        movieClipClass._callQueue = [];
        for (var i = 0; i < queue.length; i++) {
            var instance = queue[i];
            instance._allowFrameNavigation = false;
            instance.callFrame(instance._currentFrame);
            instance._allowFrameNavigation = true;
            // If the destination frame isn't the same as before the `callFrame` operation, a frame
            // navigation has happened inside the frame script. In that case, we didn't immediately
            // run frame navigation as described in `_gotoFrameAbs`. Instead, we have to do it here.
            if (instance._nextFrame !== instance._currentFrame) {
                if (movieClipClass.frameNavigationModel === 9 /* SWF9 */) {
                    instance._advanceFrame();
                    instance._constructFrame();
                    instance._removeFlags(8192 /* HasFrameScriptPending */);
                    instance.callFrame(instance._currentFrame);
                }
                else {
                    displayObjectClass.performFrameNavigation(false, true);
                }
            }
        }
        leaveTimeline();
    };
    MovieClip.runAvm1FrameScripts = function () {
        enterTimeline("MovieClip.runAvm1FrameScripts");
        var movieClipClass = MovieClip.axClass;
        var displayObjectClass = DisplayObject.axClass;
        var queue = movieClipClass._callQueue;
        movieClipClass._callQueue = [];
        var unsortedScripts = [];
        for (var i = 0; i < queue.length; i++) {
            var instance = queue[i];
            instance.queueAvm1FrameScripts(instance._currentFrame, unsortedScripts);
        }
        if (unsortedScripts.length) {
            unsortedScripts.sort(compareFrameScripts);
            for (var i = 0; i < queue.length; i++) {
                var instance = queue[i];
                instance._allowFrameNavigation = false;
            }
            var frameScripts = unsortedScripts;
            for (var i = 0; i < frameScripts.length; i++) {
                var script = frameScripts[i];
                var mc = script.context;
                release || assert(mc);
                script.call(mc);
            }
            for (var i = 0; i < queue.length; i++) {
                var instance = queue[i];
                instance._allowFrameNavigation = true;
                if (instance._nextFrame !== instance._currentFrame) {
                    displayObjectClass.performFrameNavigation(false, true);
                }
            }
        }
        leaveTimeline();
    };
    MovieClip.prototype.applySymbol = function () {
        _super.prototype.applySymbol.call(this);
        DisplayObject.axClass._advancableInstances.push(this);
        var symbol = this._symbol;
        this._totalFrames = symbol.numFrames;
        this._currentFrame = 1;
        if (!symbol.isRoot) {
            this.addScene('', symbol.labels, 0, symbol.numFrames);
        }
        this._frames = symbol.frames;
        if (symbol.isAVM1Object) {
            if (symbol.avm1Name) {
                this.name = symbol.avm1Name;
            }
        }
    };
    // This method is called when children are being constructed and AVM1 scripts
    // is about to be run.
    MovieClip.prototype._initAvm1Data = function () {
        var symbol = this._symbol;
        var frames = symbol.frames;
        if (frames) {
            for (var i = 0; i < frames.length; i++) {
                var frameInfo = frames[i];
                this._initAvm1FrameData(i, frameInfo);
            }
        }
    };
    // Adds missing AVM1 scripts data to the AS3 object frameScripts and events.
    MovieClip.prototype._initAvm1FrameData = function (frameIndex, frameInfo) {
        var symbol = this._symbol;
        var avm1Context = symbol.avm1Context;
        if (frameInfo.exports) {
            var exports = frameInfo.exports;
            for (var i = 0; i < exports.length; i++) {
                var asset = exports[i];
                avm1Context.addAsset(asset.className, asset.symbolId, null);
            }
        }
        var initActionBlocks = frameInfo.initActionBlocks;
        if (initActionBlocks) {
            this._addAvm1InitActionBlocks(frameIndex, initActionBlocks);
        }
        var actionBlocks = frameInfo.actionBlocks;
        if (actionBlocks) {
            this._addAvm1FrameScripts(frameIndex, actionBlocks);
        }
    };
    MovieClip.prototype._addAvm1FrameScripts = function (frameIndex, actionsBlocks) {
        for (var i = 0; i < actionsBlocks.length; i++) {
            var actionsBlock = actionsBlocks[i];
            var symbol = this._symbol;
            var avm1Context = symbol.avm1Context;
            var actionsData = avm1Context.actionsDataFactory.createActionsData(actionsBlock.actionsData, 's' + symbol.id + 'f' + frameIndex + 'i' + i);
            var script = function (actionsData) {
                var as2MovieClip = getAVM1Object(this, avm1Context);
                avm1Context.executeActions(actionsData, as2MovieClip);
            }.bind(this, actionsData);
            script.precedence = this._getScriptPrecedence().concat(actionsBlock.precedence);
            this.addFrameScript(frameIndex, script);
        }
    };
    /**
     * AVM1 InitActionBlocks are executed once, before the children are initialized for a frame.
     * That matches AS3's enterFrame event, so we can add an event listener that just bails
     * as long as the target frame isn't reached, and executes the InitActionBlock once it is.
     *
     * After that, the listener removes itself.
     */
    MovieClip.prototype._addAvm1InitActionBlocks = function (frameIndex, actionsBlocks) {
        function executeInitActions() {
            var symbol = self._symbol;
            var avm1Context = symbol.avm1Context;
            // 80pro: todo
            /*var as2MovieClip = AVM1.Lib.getAVM1Object(self, avm1Context);
            for (var i = 0; i < actionsBlocks.length; i++) {
              var actionsData = avm1Context.actionsDataFactory.createActionsData(
                actionsBlocks[i].actionsData, 's' + symbol.id + 'f' + frameIndex + 'i' + i);
              avm1Context.executeActions(actionsData, as2MovieClip);
            }*/
        }
        var self = this;
        if (this.currentFrame === frameIndex + 1) {
            executeInitActions();
            return;
        }
        var enterFrameListener = function () {
            if (self.currentFrame !== frameIndex + 1) {
                return;
            }
            self.removeEventListener('enterFrame', enterFrameListener);
            executeInitActions();
        };
        this.addEventListener('enterFrame', enterFrameListener);
    };
    MovieClip.prototype.removeChildAt = function (index) {
        var child = _super.prototype.removeChildAt.call(this, index);
        if (this._as2Object && child._name) {
            var avm1Child = getAVM1Object(child, this._as2Object.context);
            // Not all display objects are reflected in AVM1, so not all need to be removed.
            avm1Child && this._as2Object._removeChildName(avm1Child, child._name);
        }
        return child;
    };
    MovieClip.prototype._initializeFields = function () {
        _super.prototype._initializeFields.call(this);
        this._currentFrame = 0;
        this._totalFrames = 1;
        this._trackAsMenu = false;
        this._scenes = [];
        if (!this._symbol) {
            this.addScene('', [], 0, 1);
        }
        this._enabled = true;
        this._isPlaying = false;
        this._frames = [];
        this._frameScripts = [];
        this._nextFrame = 1;
        this._stopped = false;
        this._allowFrameNavigation = true;
        this._sounds = null;
        this._buttonFrames = Object.create(null);
        this._currentButtonState = null;
    };
    MovieClip.prototype._addFrame = function (frameInfo) {
        var spriteSymbol = this._symbol;
        var frames = spriteSymbol.frames;
        frames.push(frameInfo);
        if (frameInfo.labelName) {
            // Frame indices are 1-based, so use frames.length after pushing the frame.
            this.addFrameLabel(frameInfo.labelName, frames.length);
        }
        if (frameInfo.soundStreamHead) {
            this._initSoundStream(frameInfo.soundStreamHead);
        }
        if (frameInfo.soundStreamBlock) {
            // Frame indices are 1-based, so use frames.length after pushing the frame.
            this._addSoundStreamBlock(frames.length, frameInfo.soundStreamBlock);
        }
        if (spriteSymbol.isAVM1Object && this._hasFlags(256 /* Constructed */)) {
            var avm1Context = spriteSymbol.avm1Context;
            this._initAvm1FrameData(frames.length - 1, frameInfo);
        }
        if (frames.length === 1) {
            this._initializeChildren(frames[0]);
        }
    };
    MovieClip.prototype._initFrame = function (advance) {
        if (advance) {
            if (this.buttonMode) {
                var state = null;
                if (this._mouseOver) {
                    state = this._mouseDown ? '_down' : '_over';
                }
                else if (this._currentButtonState !== null) {
                    state = '_up';
                }
                if (state !== this._currentButtonState && this._buttonFrames[state]) {
                    this.stop();
                    this._gotoFrame(state, null);
                    this._currentButtonState = state;
                    this._advanceFrame();
                    return;
                }
            }
            if (this._totalFrames > 1 && !this._stopped &&
                this._hasFlags(256 /* Constructed */)) {
                this._nextFrame++;
            }
        }
        this._advanceFrame();
        if (this._symbol && this._symbol.isAVM1Object &&
            !this._hasFlags(256 /* Constructed */)) {
            this._initAvm1Data();
        }
    };
    MovieClip.prototype._constructFrame = function () {
        this._constructChildren();
    };
    MovieClip.prototype._enqueueFrameScripts = function () {
        var addToCallQueue = false;
        if (this._hasFlags(1024 /* NeedsLoadEvent */)) {
            addToCallQueue = true;
        }
        if (this._hasFlags(8192 /* HasFrameScriptPending */)) {
            this._removeFlags(8192 /* HasFrameScriptPending */);
            addToCallQueue = true;
        }
        if (addToCallQueue) {
            MovieClip.axClass._callQueue.push(this);
        }
        _super.prototype._enqueueFrameScripts.call(this);
    };
    Object.defineProperty(MovieClip.prototype, "currentFrame", {
        get: function () {
            var frame = this._currentFrame;
            if (!this._allowFrameNavigation &&
                MovieClip.axClass.frameNavigationModel === 1 /* SWF1 */) {
                // AVM1 needs to return a frame we already navigated during scripts execution.
                frame = this._nextFrame;
            }
            return frame - this._sceneForFrameIndex(frame).offset;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "framesLoaded", {
        get: function () {
            return this._frames.length;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "totalFrames", {
        get: function () {
            return this._totalFrames;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "trackAsMenu", {
        get: function () {
            return this._trackAsMenu;
        },
        set: function (value) {
            this._trackAsMenu = !!value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "scenes", {
        get: function () {
            var scenes = this._scenes ? this._scenes.map(function (x) {
                return x.clone();
            }) : [];
            return this.sec.createArrayUnsafe(scenes);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "currentScene", {
        get: function () {
            var scene = this._sceneForFrameIndex(this._currentFrame);
            return scene.clone();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "currentLabel", {
        get: function () {
            var label = this._labelForFrame(this._currentFrame);
            return label ? label.name : null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "currentLabels", {
        get: function () {
            return this._sceneForFrameIndex(this._currentFrame).labels;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "currentFrameLabel", {
        get: function () {
            var scene = this._sceneForFrameIndex(this._currentFrame);
            var label = scene.getLabelByFrame(this._currentFrame - scene.offset);
            return label && label.name;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "enabled", {
        get: function () {
            return this._enabled;
        },
        set: function (value) {
            this._enabled = !!value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MovieClip.prototype, "isPlaying", {
        get: function () {
            return this._isPlaying;
        },
        enumerable: true,
        configurable: true
    });
    MovieClip.prototype.play = function () {
        if (this._totalFrames > 1) {
            this._isPlaying = true;
        }
        this._stopped = false;
    };
    MovieClip.prototype.stop = function () {
        this._isPlaying = false;
        this._stopped = true;
    };
    /**
     * Resolves frame and scene into absolute frame number. If scene is not specified,
     * the current scene is used. In legacy mode, it might return `undefined` if frame/scene
     * was not found.
     */
    MovieClip.prototype._getAbsFrameNumber = function (frame, sceneName) {
        var navigationModel = MovieClip.axClass.frameNavigationModel;
        var legacyMode = navigationModel !== 10 /* SWF10 */;
        var scene;
        if (sceneName !== null) {
            sceneName = axCoerceString(sceneName);
            var scenes = this._scenes;
            release || assert(scenes.length, "There should be at least one scene defined.");
            for (var i = 0; i < scenes.length; i++) {
                scene = scenes[i];
                if (scene.name === sceneName) {
                    break;
                }
            }
            if (i === scenes.length) {
                if (legacyMode) {
                    return undefined; // noop for SWF9 and below
                }
                this.sec.throwError('ArgumentError', Errors.SceneNotFoundError, sceneName);
            }
        }
        else {
            scene = this._sceneForFrameIndex(this._currentFrame);
        }
        // Amazingly, the `frame` argument, while first coerced to string, is then interpreted as a
        // frame index even if a label with the same name exists.
        /* tslint:disable */
        var frameNum = parseInt(frame, 10);
        if (frameNum != frame) { // TypeScript doesn't like using `==` for number,string vars.
            var label = scene.getLabelByName(frame, legacyMode);
            if (!label) {
                if (legacyMode) {
                    return undefined; // noop for SWF9 and below
                }
                this.sec.throwError('ArgumentError', Errors.FrameLabelNotFoundError, frame, sceneName);
            }
            frameNum = label.frame;
        }
        /* tslint:enable */
        return scene.offset + frameNum;
    };
    /**
     * Implementation for both gotoAndPlay and gotoAndStop.
     *
     * Technically, we should throw all errors from those functions directly so the stack is
     * correct.
     * We might at some point do that by explicitly inlining this function using some build step.
     */
    MovieClip.prototype._gotoFrame = function (frame, sceneName) {
        var frameNum = this._getAbsFrameNumber(frame, sceneName);
        if (frameNum === undefined) {
            return;
        }
        this._gotoFrameAbs(frameNum);
    };
    MovieClip.prototype._gotoFrameAbs = function (frame) {
        if (frame < 1) {
            frame = 1;
        }
        else if (frame > this._totalFrames) {
            frame = this._totalFrames;
        }
        if (frame === this._nextFrame) {
            return;
        }
        this._nextFrame = frame;
        // Frame navigation only happens immediately if not triggered from under a frame script.
        if (this._allowFrameNavigation) {
            if (MovieClip.axClass.frameNavigationModel === 9 /* SWF9 */) {
                // In FP 9, the only thing that happens on inter-frame navigation is advancing the frame
                // and constructing new timeline objects.
                this._advanceFrame();
                this._constructFrame();
            }
            else {
                // Frame navigation in an individual timeline triggers an iteration of the whole
                // frame navigation cycle in FP 10+. This includes broadcasting frame events to *all*
                // display objects.
                DisplayObject.axClass.performFrameNavigation(false, true);
            }
        }
    };
    MovieClip.prototype._advanceFrame = function () {
        var currentFrame = this._currentFrame;
        var nextFrame = this._nextFrame;
        if (nextFrame > this._totalFrames) {
            nextFrame = 1;
        }
        if (currentFrame === nextFrame) {
            // If nextFrame was > this._totalFrames, it has to be written back here, otherwise it'll
            // just be incremented ever further.
            this._nextFrame = nextFrame;
            return;
        }
        if (nextFrame > this.framesLoaded) {
            // If nextFrame was > this._totalFrames, it has to be written back here, otherwise it'll
            // just be incremented ever further.
            this._nextFrame = nextFrame;
            // TODO
            return;
        }
        var currentSwfFrame = this._frames[currentFrame - 1];
        var nextSwfFrame = this._frames[nextFrame - 1];
        if (nextSwfFrame !== currentSwfFrame) {
            this._seekToFrame(nextFrame);
            if (nextSwfFrame.controlTags) {
                var tags = nextSwfFrame.controlTags;
                var soundStarts;
                for (var i = 0; i < tags.length; i++) {
                    var tag = tags[i];
                    // controlTags might contain parsed and unparsed tags.
                    if (tag.tagCode === 15 /* CODE_START_SOUND */ ||
                        tag.tagCode === 61 /* CODE_VIDEO_FRAME */) {
                        var loaderInfo = this._symbol.loaderInfo;
                        tag = loaderInfo._file.getParsedTag(tag);
                    }
                    if (tag.code === 15 /* CODE_START_SOUND */) {
                        if (!soundStarts) {
                            soundStarts = [];
                        }
                        soundStarts.push(new SoundStart(tag.soundId, tag.soundInfo));
                    }
                    else if (tag.tagCode === 61 /* CODE_VIDEO_FRAME */) {
                        // tag.videoData contains a video frame for the video stream symbol with id == tag.streamId.
                        // TODO: Do something useful with that information :).
                    }
                }
                if (soundStarts) {
                    this._registerStartSounds(nextFrame, soundStarts);
                }
            }
        }
        if (this._frameScripts[nextFrame]) {
            this._setFlags(8192 /* HasFrameScriptPending */);
            this._parent && this._propagateFlagsUp(16384 /* ContainsFrameScriptPendingChildren */);
        }
        this._currentFrame = this._nextFrame = nextFrame;
        this._syncSounds(nextFrame);
    };
    MovieClip.prototype._seekToFrame = function (frame) {
        var currentFrame = this._currentFrame;
        var frames = this._frames;
        if (frame === currentFrame + 1) {
            var nextSwfFrame = frames[frame - 1];
            if (nextSwfFrame.controlTags) {
                this._processControlTags(nextSwfFrame.controlTags, false);
            }
            return;
        }
        var currentSwfFrame = frames[currentFrame - 1];
        var loaderInfo = this._symbol.loaderInfo;
        var backwards = frame < currentFrame;
        var controlTags = [];
        var removedObjects;
        // We scan all control tags in reverse order and make sure we only apply those related to
        // objects that exist in the new frame.
        var i = frame;
        var n = backwards ? 0 : currentFrame;
        while (i-- > n) {
            var swfFrame = frames[i];
            if (swfFrame === currentSwfFrame) {
                continue;
            }
            currentSwfFrame = swfFrame;
            var tags = swfFrame.controlTags;
            if (!tags) {
                continue;
            }
            var j = tags.length;
            while (j--) {
                // We may have a mix of the parsed and unparsed tags.
                var parsedOrUnparsedTag = tags[j];
                var tag = parsedOrUnparsedTag.tagCode === undefined ?
                    parsedOrUnparsedTag : loaderInfo._file.getParsedTag(parsedOrUnparsedTag);
                switch (tag.code) {
                    case 5 /* CODE_REMOVE_OBJECT */:
                    case 28 /* CODE_REMOVE_OBJECT2 */:
                        if (!removedObjects) {
                            removedObjects = Object.create(null);
                        }
                        removedObjects[tag.depth] = true;
                        if (!backwards) {
                            controlTags.push(tag);
                        }
                        break;
                    case 4 /* CODE_PLACE_OBJECT */:
                    case 26 /* CODE_PLACE_OBJECT2 */:
                    case 70 /* CODE_PLACE_OBJECT3 */:
                        if (!(removedObjects && removedObjects[tag.depth])) {
                            controlTags.push(tag);
                        }
                        break;
                    default:
                        controlTags.push(tag);
                }
            }
        }
        // Bring the order back to normal.
        controlTags.reverse();
        this._processControlTags(controlTags, backwards);
    };
    /**
     * Because that's how it's mostly used, the current frame is stored as an offset into the
     * entire timeline. Sometimes, we need to know which scene it falls into. This utility
     * function answers that.
     */
    MovieClip.prototype._sceneForFrameIndex = function (frameIndex) {
        var scenes = this._scenes;
        // A gotoAnd* might be invoked by script before the first advanceFrame call. In that case,
        // _currentFrame is 0, which means this function is called with frameIndex being 0.
        // We just return the first scene in that case.
        if (frameIndex === 0) {
            return scenes[0];
        }
        for (var i = 0; i < scenes.length; i++) {
            var scene = scenes[i];
            if (scene.offset < frameIndex && scene.offset + scene.numFrames >= frameIndex) {
                return scene;
            }
        }
        release || assertUnreachable("Must have at least one scene covering all frames.");
    };
    /**
     * Frame indices are stored as offsets into the entire timline, whereas labels are stored
     * in their scenes. This utility function iterates over scenes and their labels to find
     * the label clostest to, but not after the target frame.
     */
    MovieClip.prototype._labelForFrame = function (frame) {
        var scenes = this._scenes;
        var label = null;
        for (var i = 0; i < scenes.length; i++) {
            var scene = scenes[i];
            if (scene.offset > frame) {
                return label;
            }
            var labels = scene.labels.value;
            for (var j = 0; j < labels.length; j++) {
                var currentLabel = labels[j];
                if (currentLabel.frame > frame - scene.offset) {
                    return label;
                }
                label = currentLabel;
            }
        }
        return label;
    };
    MovieClip.prototype.callFrame = function (frame) {
        frame = frame | 0;
        var frameScripts = this._frameScripts[frame];
        if (!frameScripts) {
            return;
        }
        for (var i = 0; i < frameScripts.length; i++) {
            var script = frameScripts[i];
            try {
                script.call(this); // REDUX ? why it was frameScript.$Bgcall(thisArg);
            }
            catch (e) {
                // Telemetry.instance.reportTelemetry({ topic: 'error', error: Telemetry.ErrorTypes.AVM2_ERROR });
                //if ($DEBUG) {
                //  console.error('error ' + e + ', stack: \n' + e.stack);
                //}
                this.stop();
                throw e;
            }
        }
    };
    MovieClip.prototype.queueAvm1FrameScripts = function (frame, queue) {
        // AVM1 action blocks must be executed exactly in the same order they appear in the SWF
        // file. We keep track of the position for each action block at parsing time and carry that
        // information on to their wrapping functions (via the precedence property). Here we queue
        // up all such functions so we can sort and execute them after this loop.
        if (this._hasFlags(1024 /* NeedsLoadEvent */)) {
            this._removeFlags(1024 /* NeedsLoadEvent */);
            release || assert(this._symbol);
            var handler = function () {
                var eventClass = this.sec.flash.events.Event.axClass;
                this.dispatchEvent(eventClass.getInstance(Event.AVM1_LOAD));
            }.bind(this);
            handler.precedence = this._getScriptPrecedence();
            handler.context = this;
            queue.push(handler);
        }
        var frameScripts = this._frameScripts[frame];
        if (frameScripts) {
            for (var j = 0; j < frameScripts.length; j++) {
                var script = frameScripts[j];
                script.context = this;
                queue.push(script);
            }
        }
    };
    MovieClip.prototype.nextFrame = function () {
        this.gotoAndStop(this._currentFrame + 1);
    };
    MovieClip.prototype.prevFrame = function () {
        this.gotoAndStop(this._currentFrame - 1);
    };
    MovieClip.prototype.gotoAndPlay = function (frame, scene) {
        if (scene === void 0) { scene = null; }
        // Argument handling for gotoAnd* is a bit peculiar:
        // - too many arguments throw just as too few do
        // - the `sceneName` argument is coerced first
        // - the `frame` argument is coerced to string, but `undefined` results in `"null"`
        if (arguments.length === 0 || arguments.length > 2) {
            this.sec.throwError('ArgumentError', Errors.WrongArgumentCountError, 'flash.display::MovieClip/gotoAndPlay()', 1, arguments.length);
        }
        scene = axCoerceString(scene);
        frame = axCoerceString(frame) + ''; // The axCoerceString returns `null` for `undefined`.
        this.play();
        this._gotoFrame(frame, scene);
    };
    MovieClip.prototype.gotoAndStop = function (frame, scene) {
        if (scene === void 0) { scene = null; }
        // See comment in gotoAndPlay for an explanation of the arguments handling stuff.
        if (arguments.length === 0 || arguments.length > 2) {
            this.sec.throwError('ArgumentError', Errors.WrongArgumentCountError, 'flash.display::MovieClip/gotoAndPlay()', 1, arguments.length);
        }
        scene = axCoerceString(scene);
        frame = axCoerceString(frame) + ''; // The axCoerceString returns `null` for `undefined`.
        this.stop();
        this._gotoFrame(frame, scene);
    };
    /**
     * Takes pairs of `frameIndex`, `script` arguments and adds the `script`s to the `_frameScripts`
     * Array.
     *
     * Undocumented method used to implement the old timeline concept in AS3.
     */
    MovieClip.prototype.addFrameScript = function (frameIndex, script /*, ...*/) {
        if (!this._currentFrame) {
            return;
        }
        // arguments are pairs of frameIndex and script/function
        // frameIndex is in range 0..totalFrames-1
        var numArgs = arguments.length;
        if (numArgs & 1) {
            this.sec.throwError('ArgumentError', Errors.TooFewArgumentsError, numArgs, numArgs + 1);
        }
        var frameScripts = this._frameScripts;
        var totalFrames = this._totalFrames;
        for (var i = 0; i < numArgs; i += 2) {
            var frameNum = (arguments[i] | 0) + 1;
            if (frameNum < 1 || frameNum > totalFrames) {
                continue;
            }
            var fn = arguments[i + 1];
            var list = frameScripts[frameNum];
            if (list) {
                if (fn.precedence) {
                    list.push(fn);
                }
                else {
                    list[0] = fn;
                }
            }
            else {
                frameScripts[frameNum] = [fn];
            }
            if (frameNum === this._currentFrame) {
                this._setFlags(8192 /* HasFrameScriptPending */);
                this._parent && this._propagateFlagsUp(16384 /* ContainsFrameScriptPendingChildren */);
            }
        }
    };
    Object.defineProperty(MovieClip.prototype, "_isFullyLoaded", {
        get: function () {
            return this.framesLoaded >= this.totalFrames;
        },
        enumerable: true,
        configurable: true
    });
    MovieClip.prototype._registerStartSounds = function (frameNum, soundStartInfo) {
        if (this._sounds === null) {
            this._sounds = new MovieClipSoundsManager(this);
        }
        this._sounds.registerStartSounds(frameNum, soundStartInfo);
    };
    MovieClip.prototype._initSoundStream = function (streamInfo) {
        if (this._sounds === null) {
            this._sounds = new MovieClipSoundsManager(this);
        }
        this._sounds.initSoundStream(streamInfo);
    };
    MovieClip.prototype._addSoundStreamBlock = function (frameNum, streamBlock) {
        if (this._sounds === null) {
            this._sounds = new MovieClipSoundsManager(this);
        }
        this._sounds.addSoundStreamBlock(frameNum, streamBlock);
    };
    MovieClip.prototype._syncSounds = function (frameNum) {
        if (this._sounds !== null) {
            this._sounds.syncSounds(frameNum);
        }
    };
    MovieClip.prototype.addScene = function (name, labels_, offset, numFrames) {
        var labels = this.sec.createArrayUnsafe(labels_);
        this._scenes.push(new Scene(name, labels, offset, numFrames));
    };
    MovieClip.prototype.addFrameLabel = function (name, frame) {
        var scene = this._sceneForFrameIndex(frame);
        if (!scene.getLabelByName(name, false)) {
            scene.labels.value.push(new FrameLabel(name, frame - scene.offset));
        }
    };
    MovieClip.prototype.prevScene = function () {
        var currentScene = this._sceneForFrameIndex(this._currentFrame);
        if (currentScene.offset === 0) {
            return;
        }
        // Since scene offsets are 0-based, the current scene's offset, treated as a frame index,
        // is the previous scene's last frame.
        this._gotoFrameAbs(this._sceneForFrameIndex(currentScene.offset).offset + 1);
    };
    MovieClip.prototype.nextScene = function () {
        var currentScene = this._sceneForFrameIndex(this._currentFrame);
        if (currentScene.offset + currentScene.numFrames === this._totalFrames) {
            return;
        }
        this._gotoFrameAbs(currentScene.offset + currentScene.numFrames + 1);
    };
    MovieClip.prototype._containsPointImpl = function (globalX, globalY, localX, localY, testingType, objects, skipBoundsCheck) {
        var result = _super.prototype._containsPointImpl.call(this, globalX, globalY, localX, localY, testingType, objects, true);
        // In AVM1 SWFs, MovieClips are transparent to the mouse as long as they don't have a handler
        // attached to them for any of the button-related events.
        if (result === 2 /* Shape */ && testingType === 3 /* Mouse */ &&
            '_as2Object' in this && !this.buttonMode && objects[0] === this) {
            objects.length = 0;
        }
        return result;
    };
    // List of static symbols to link.
    MovieClip.classSymbols = null; // [];
    // List of instance symbols to link.
    MovieClip.instanceSymbols = null; // ["currentLabels"];
    return MovieClip;
}(Sprite));
export { MovieClip };
