import { __extends } from "tslib";
import { DisplayObjectContainer } from "./DisplayObjectContainer";
import { release, assert, notImplemented, warning } from "../../base/utilities/Debug";
import { DisplaySymbol } from "../symbol";
import { FrameLabel } from "./FrameLabel";
import { clamp } from "../../base/utilities/NumberUtilities";
import { Mouse } from "../ui/Mouse";
import { ActionScriptVersion } from "./ActionScriptVersion";
import { MovieClip } from "./MovieClip";
/**
 * Copyright 2014 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
// Class: Sprite
var DragMode;
(function (DragMode) {
    DragMode[DragMode["Inactive"] = 0] = "Inactive";
    // Indicates that the dragged object is locked to the center of the mouse position.
    DragMode[DragMode["LockToPointer"] = 1] = "LockToPointer";
    // Indicates that the dragged object is locked to the point where the dragging process started.
    DragMode[DragMode["PreserveDistance"] = 2] = "PreserveDistance";
})(DragMode || (DragMode = {}));
var Sprite = /** @class */ (function (_super) {
    __extends(Sprite, _super);
    function Sprite() {
        var _this = _super.call(this) || this;
        if (_this._symbol && !_this._fieldsInitialized) {
            _this.applySymbol();
        }
        if (!_this._fieldsInitialized) {
            _this._initializeFields();
        }
        _this._constructChildren();
        return _this;
    }
    Sprite.prototype.applySymbol = function () {
        release || assert(this._symbol);
        this._initializeFields();
        var symbol = this._symbol;
        if (symbol.isRoot) {
            this._root = this;
        }
        this._children = [];
        if (symbol.numFrames && symbol.frames.length > 0) {
            // For a SWF's root symbol, all frames are added after initialization, with
            // _initializeChildren called after the first frame is added.
            this._initializeChildren(symbol.frames[0]);
        }
    };
    Sprite.prototype._initializeFields = function () {
        _super.prototype._initializeFields.call(this);
        this._graphics = null;
        this._buttonMode = false;
        this._dropTarget = null;
        this._hitArea = null;
        this._useHandCursor = true;
        this._dragMode = DragMode.Inactive;
        this._dragDeltaX = 0;
        this._dragDeltaY = 0;
        this._dragBounds = null;
        this._hitTarget = null;
    };
    Sprite.prototype._addFrame = function (frame) {
        var frames = this._symbol.frames;
        frames.push(frame);
        if (frames.length === 1) {
            this._initializeChildren(frame);
        }
    };
    Sprite.prototype._initializeChildren = function (frame) {
        if (frame.controlTags) {
            this._processControlTags(frame.controlTags, false);
        }
    };
    Sprite.prototype._processControlTags = function (tags, backwards) {
        // When seeking backwards all timeline objects will be removed unless they are placed again.
        if (backwards) {
            var children = this._children.slice();
            for (var i = 0; i < children.length; i++) {
                var child = children[i];
                if (child._depth < 0) {
                    continue;
                }
                // For AVM1 objects, children with depth 16384 (0 from API point of view)
                // are not removed.
                if (child._depth >= 16384 && '_as2Object' in child) {
                    continue;
                }
                var tag = null;
                // Look for a control tag tag that places an object at the same depth as the current
                // child.
                for (var j = 0; j < tags.length; j++) {
                    if (tags[j].depth === child._depth) {
                        tag = tags[j];
                        break;
                    }
                }
                // If no such tag was found or a different object is placed, remove the current child.
                if (!tag || child._symbol.id !== tag.symbolId || child._ratio !== (tag.ratio | 0)) {
                    this._removeAnimatedChild(child);
                }
            }
        }
        var loaderInfo = this._symbol.loaderInfo;
        for (var i = 0; i < tags.length; i++) {
            // We may have a mix of the parsed and unparsed tags.
            var parsedOrUnparsedTag = tags[i];
            var tag = parsedOrUnparsedTag.tagCode === undefined ?
                parsedOrUnparsedTag : loaderInfo._file.getParsedTag(parsedOrUnparsedTag);
            switch (tag.code) {
                case 5 /* CODE_REMOVE_OBJECT */:
                case 28 /* CODE_REMOVE_OBJECT2 */:
                    var child = this.getTimelineObjectAtDepth(tag.depth | 0);
                    if (child) {
                        this._removeAnimatedChild(child);
                    }
                    break;
                case 4 /* CODE_PLACE_OBJECT */:
                case 26 /* CODE_PLACE_OBJECT2 */:
                case 70 /* CODE_PLACE_OBJECT3 */:
                    var placeObjectTag = tag;
                    var depth = placeObjectTag.depth;
                    var child = this.getTimelineObjectAtDepth(depth);
                    var hasCharacter = placeObjectTag.symbolId > -1;
                    // Check for invalid flag constellations.
                    if (placeObjectTag.flags & 1 /* Move */) {
                        // Invalid case 1: Move flag set but no child found at given depth.
                        if (!child) {
                            //  Ignore the current tag.
                            break;
                        }
                    }
                    else if (!hasCharacter || (child && !(backwards && hasCharacter))) {
                        // Invalid case 2: Neither Move nor HasCharacter flag set.
                        // Invalid case 3: HasCharacter flag set but given depth is already occupied by a
                        // another object (only if seeking forward).
                        warning("Warning: Failed to place object at depth " + depth + ".");
                        break;
                    }
                    var symbol = null;
                    if (hasCharacter) {
                        symbol = loaderInfo.getSymbolById(placeObjectTag.symbolId);
                        // The Flash Player ignores references to undefined symbols here. So should we.
                        if (!symbol) {
                            break;
                        }
                    }
                    if (child) {
                        if (symbol && !symbol.dynamic) {
                            // If the current object is of a simple type (for now Shapes, MorphShapes and
                            // StaticText) only its static content is updated instead of replacing it with a
                            // new instance. TODO: Handle
                            // http://wahlers.com.br/claus/blog/hacking-swf-2-placeobject-and-ratio/.
                            child._setStaticContentFromSymbol(symbol);
                        }
                        // We animate the object only if a user script didn't touch any of the properties
                        // this would affect.
                        if (child._hasFlags(4096 /* AnimatedByTimeline */)) {
                            child._animate(tag);
                        }
                    }
                    else {
                        // Place a new instance of the symbol.
                        child = this.createAnimatedDisplayObject(symbol, placeObjectTag, false);
                        this.addTimelineObjectAtDepth(child, depth);
                        if (symbol.isAVM1Object) {
                            child._placeObjectTag = placeObjectTag;
                            child._setFlags(131072 /* HasPlaceObjectInitPending */);
                        }
                    }
                    break;
            }
        }
    };
    Sprite.prototype._removeAnimatedChild = function (child) {
        this.removeChild(child);
        if (child._name) {
            if (this.axGetPublicProperty(child._name) === child) {
                this.axSetPublicProperty(child._name, null);
            }
            // TODO: Implement proper reference counting.
            // child._removeReference();
        }
        if (child._maskedObject) {
            child._maskedObject.mask = null;
        }
    };
    Sprite.prototype._canHaveGraphics = function () {
        return true;
    };
    Sprite.prototype._getGraphics = function () {
        return this._graphics;
    };
    Object.defineProperty(Sprite.prototype, "graphics", {
        get: function () {
            return this._ensureGraphics();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Sprite.prototype, "buttonMode", {
        get: function () {
            return this._buttonMode;
        },
        set: function (value) {
            this._buttonMode = !!value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Sprite.prototype, "dropTarget", {
        get: function () {
            return this._dropTarget;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Sprite.prototype, "hitArea", {
        get: function () {
            return this._hitArea;
        },
        set: function (value) {
            value = value;
            if (this._hitArea === value) {
                return;
            }
            if (value && value._hitTarget) {
                value._hitTarget._hitArea = null;
            }
            this._hitArea = value;
            if (value) {
                value._hitTarget = this;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Sprite.prototype, "useHandCursor", {
        get: function () {
            return this._useHandCursor;
        },
        set: function (value) {
            this._useHandCursor = !!value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Sprite.prototype, "soundTransform", {
        get: function () {
            release || notImplemented("public flash.display.Sprite::get soundTransform");
            return;
            // return this._soundTransform;
        },
        set: function (sndTransform) {
            sndTransform = sndTransform;
            release || notImplemented("public flash.display.Sprite::set soundTransform");
            return;
            // this._soundTransform = sndTransform;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Returns the current mouse position relative to this object.
     */
    Sprite.prototype._getDragMousePosition = function () {
        var position = Mouse.axClass._currentPosition;
        if (this._parent) {
            position = this._parent.globalToLocal(position);
        }
        return position;
    };
    Sprite.prototype.startDrag = function (lockCenter, bounds) {
        if (lockCenter === void 0) { lockCenter = false; }
        if (bounds === void 0) { bounds = null; }
        lockCenter = !!lockCenter;
        if (lockCenter) {
            this._dragMode = DragMode.LockToPointer;
        }
        else {
            this._dragMode = DragMode.PreserveDistance;
            var mousePosition = this._getDragMousePosition();
            this._dragDeltaX = this.x - mousePosition.x;
            this._dragDeltaY = this.y - mousePosition.y;
        }
        this._dragBounds = bounds;
        // TODO: Our mouse handling logic looks up draggableObject on stage.sec.flash.ui.Mouse.axClass
        // to update its position. Could there be a case where stage.sec !== this.sec?
        Mouse.axClass.draggableObject = this;
    };
    Sprite.prototype.stopDrag = function () {
        if (Mouse.axClass.draggableObject === this) {
            Mouse.axClass.draggableObject = null;
            this._dragMode = DragMode.Inactive;
            this._dragDeltaX = 0;
            this._dragDeltaY = 0;
            this._dragBounds = null;
        }
    };
    Sprite.prototype._updateDragState = function (dropTarget) {
        if (dropTarget === void 0) { dropTarget = null; }
        var mousePosition = this._getDragMousePosition();
        var newX = mousePosition.x;
        var newY = mousePosition.y;
        if (this._dragMode === DragMode.PreserveDistance) {
            // Preserve the distance to the point where the dragging process started.
            newX += this._dragDeltaX;
            newY += this._dragDeltaY;
        }
        if (this._dragBounds) {
            // Clamp new position to constraint bounds.
            var bounds = this._dragBounds;
            newX = clamp(newX, bounds.left, bounds.right);
            newY = clamp(newY, bounds.top, bounds.bottom);
        }
        this.x = newX;
        this.y = newY;
        this._dropTarget = dropTarget;
    };
    Sprite.prototype.startTouchDrag = function (touchPointID /*int*/, lockCenter, bounds) {
        if (lockCenter === void 0) { lockCenter = false; }
        if (bounds === void 0) { bounds = null; }
        touchPointID = touchPointID | 0;
        lockCenter = !!lockCenter;
        bounds = bounds;
        release || notImplemented("public flash.display.Sprite::startTouchDrag");
        return;
    };
    Sprite.prototype.stopTouchDrag = function (touchPointID /*int*/) {
        touchPointID = touchPointID | 0;
        release || notImplemented("public flash.display.Sprite::stopTouchDrag");
        return;
    };
    Sprite.prototype._containsPoint = function (globalX, globalY, localX, localY, testingType, objects) {
        // If looking for a drop target, ignore this object if it is the one being dragged.
        if (testingType === 5 /* Drop */ && this._dragMode > DragMode.Inactive) {
            return;
        }
        var result = this._boundsAndMaskContainPoint(globalX, globalY, localX, localY, testingType);
        if (!result && testingType === 3 /* Mouse */ && this._hitArea && this._mouseEnabled) {
            var matrix = this._hitArea._getInvertedConcatenatedMatrix();
            var hitAreaLocalX = matrix.transformX(globalX, globalY);
            var hitAreaLocalY = matrix.transformY(globalX, globalY);
            result = this._hitArea._boundsAndMaskContainPoint(globalX, globalY, hitAreaLocalX, hitAreaLocalY, testingType);
        }
        if (result === 0 /* None */ || testingType < 2 /* HitTestShape */) {
            return result;
        }
        return this._containsPointImpl(globalX, globalY, localX, localY, testingType, objects, true);
    };
    Sprite.prototype._containsPointDirectly = function (localX, localY, globalX, globalY) {
        if (this._hitArea) {
            return !!this._hitArea._containsGlobalPoint(globalX, globalY, 2 /* HitTestShape */, null);
        }
        var graphics = this._getGraphics();
        return !!graphics && graphics._containsPoint(localX, localY, true, 0);
    };
    Sprite.classInitializer = null;
    // List of static symbols to link.
    Sprite.classSymbols = null; // [];
    // List of instance symbols to link.
    Sprite.instanceSymbols = null; // [];
    return Sprite;
}(DisplayObjectContainer));
export { Sprite };
var SpriteSymbol = /** @class */ (function (_super) {
    __extends(SpriteSymbol, _super);
    function SpriteSymbol(data, loaderInfo) {
        var _this = _super.call(this, data, MovieClip.axClass, true) || this;
        _this.numFrames = 1;
        _this.frames = [];
        _this.labels = [];
        _this.loaderInfo = loaderInfo;
        return _this;
    }
    SpriteSymbol.FromData = function (data, loaderInfo) {
        var symbol = new SpriteSymbol(data, loaderInfo);
        symbol.numFrames = data.frameCount;
        if (loaderInfo.actionScriptVersion === ActionScriptVersion.ACTIONSCRIPT2) {
            symbol.isAVM1Object = true;
            symbol.avm1Context = loaderInfo._avm1Context;
        }
        var frames = data.frames;
        for (var i = 0; i < frames.length; i++) {
            var frame = loaderInfo.getFrame(data, i);
            if (frame.labelName) {
                symbol.labels.push(new FrameLabel(frame.labelName, i + 1));
            }
            symbol.frames.push(frame);
        }
        return symbol;
    };
    return SpriteSymbol;
}(DisplaySymbol));
export { SpriteSymbol };
