import { __extends } from "tslib";
import { DisplayObjectContainer } from "./DisplayObjectContainer";
import { StageScaleMode } from "./StageScaleMode";
import { StageQuality } from "./StageQuality";
import { Rectangle } from "../geom/Rectangle";
import { DisplayObject } from "./DisplayObject";
import { axCoerceString } from "../../avm2/run/axCoerceString";
import { Errors } from "../../avm2/errors";
import { StageAlign } from "./StageAlign";
import { release, assert, notImplemented, somewhatImplemented } from "../../base/utilities/Debug";
import { Event } from "../events/Event";
import { StageDisplayState } from "./StageDisplayState";
import { ColorCorrection } from "./ColorCorrection";
import { ColorCorrectionSupport } from "./ColorCorrectionSupport";
/**
 * Copyright 2014 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
// Class: Stage
var Stage = /** @class */ (function (_super) {
    __extends(Stage, _super);
    function Stage() {
        var _this = _super.call(this) || this;
        _this._root = _this;
        _this._stage = _this;
        _this._frameRate = 24;
        _this._scaleMode = StageScaleMode.SHOW_ALL;
        _this._align = "";
        _this._stageWidth = 0;
        _this._stageHeight = 0;
        _this._showDefaultContextMenu = true;
        _this._focus = null;
        _this._colorCorrection = ColorCorrection.DEFAULT;
        _this._colorCorrectionSupport = ColorCorrectionSupport.DEFAULT_OFF;
        _this._stageFocusRect = true;
        _this._quality = StageQuality.HIGH;
        _this._displayState = null;
        _this._fullScreenSourceRect = null;
        _this._mouseLock = false;
        var objectVectorClass = _this.sec.getVectorClass(_this.sec.AXObject);
        _this._stageVideos = objectVectorClass.axConstruct([0, true]);
        _this._stage3Ds = objectVectorClass.axConstruct([0, true]);
        _this._colorARGB = 0xFFFFFFFF;
        _this._fullScreenWidth = 0;
        _this._fullScreenHeight = 0;
        _this._wmodeGPU = false;
        _this._softKeyboardRect = new Rectangle();
        _this._allowsFullScreen = false;
        _this._allowsFullScreenInteractive = false;
        _this._contentsScaleFactor = 1;
        _this._displayContextInfo = null;
        _this._timeout = -1;
        _this._stageContainerWidth = -1;
        _this._stageContainerHeight = -1;
        _this._setFlags(262144 /* HasPerspectiveProjection */);
        /**
         * Indicates if a Render event was requested by calling the |invalid| function.
         */
        _this._invalidated = false;
        return _this;
    }
    Stage.prototype.setRoot = function (root) {
        this.addTimelineObjectAtDepth(root, 0);
    };
    Object.defineProperty(Stage.prototype, "frameRate", {
        get: function () {
            return this._frameRate;
        },
        set: function (value) {
            this._frameRate = +value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "scaleMode", {
        get: function () {
            return this._scaleMode;
        },
        set: function (value) {
            value = axCoerceString(value);
            if (StageScaleMode.toNumber(value) < 0) {
                this.sec.throwError("ArgumentError", Errors.InvalidEnumError, "scaleMode");
            }
            if (this._scaleMode !== value) {
                this._isDirty = true;
                this._scaleMode = value;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "align", {
        get: function () {
            return this._align;
        },
        set: function (value) {
            value = axCoerceString(value);
            var n = StageAlign.toNumber(value);
            release || assert(n >= 0);
            var newValue = StageAlign.fromNumber(n);
            if (this._align !== newValue) {
                this._isDirty = true;
                this._align = newValue;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "stageWidth", {
        get: function () {
            if (this.scaleMode !== StageScaleMode.NO_SCALE) {
                return this._stageWidth / 20 | 0;
            }
            release || assert(this._stageContainerWidth >= 0);
            return this._stageContainerWidth;
        },
        set: function (value /*int*/) {
            // While the setter doesn't change the stored value, it still coerces the `value` parameter.
            // This is script-visible if the value is something like `{valueOf: function(){throw 1}}`.
            value = value | 0;
        },
        enumerable: true,
        configurable: true
    });
    Stage.prototype._setInitialName = function () {
        this._name = null;
    };
    /**
     * Non-AS3-available setter. In AS3, the `stageWidth` setter is silently ignored.
     */
    Stage.prototype.setStageWidth = function (value) {
        release || assert((value | 0) === value);
        var newValue = (value * 20) | 0;
        if (this._stageWidth !== newValue) {
            this._isDirty = true;
            this._stageWidth = newValue;
        }
    };
    Object.defineProperty(Stage.prototype, "stageHeight", {
        get: function () {
            if (this.scaleMode !== StageScaleMode.NO_SCALE) {
                return this._stageHeight / 20 | 0;
            }
            release || assert(this._stageContainerHeight >= 0);
            return this._stageContainerHeight;
        },
        set: function (value /*int*/) {
            // While the setter doesn't change the stored value, it still coerces the `value` parameter.
            // This is script-visible if the value is something like `{valueOf: function(){throw 1}}`.
            value = value | 0;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Non-AS3-available setter. In AS3, the `stageHeight` setter is silently ignored.
     */
    Stage.prototype.setStageHeight = function (value) {
        release || assert((value | 0) === value);
        var newValue = (value * 20) | 0;
        if (this._stageHeight !== newValue) {
            this._isDirty = true;
            this._stageHeight = newValue;
        }
    };
    /**
     * Almost the same as color setter, except it preserves alpha channel.
     * @param value
     */
    Stage.prototype.setStageColor = function (value) {
        if (this._colorARGB !== value) {
            this._isDirty = true;
            this._colorARGB = value;
        }
    };
    /**
     * Non-AS3-available setter. Notifies the stage that the dimensions of the stage container have changed.
     */
    Stage.prototype.setStageContainerSize = function (width, height, pixelRatio) {
        // Flash doesn't fire a resize event if the pixel ratio has changed, but it needs to be set if
        // a resize event gets dispatched as a result of a size change.
        this._contentsScaleFactor = pixelRatio;
        var sizeHasChanged = this._stageContainerWidth !== width || this._stageContainerHeight !== height;
        if (sizeHasChanged) {
            this._stageContainerWidth = width;
            this._stageContainerHeight = height;
            if (this.scaleMode === StageScaleMode.NO_SCALE) {
                this.dispatchEvent(Event.axClass.getInstance(Event.RESIZE));
            }
        }
    };
    Object.defineProperty(Stage.prototype, "showDefaultContextMenu", {
        get: function () {
            return this._showDefaultContextMenu;
        },
        set: function (value) {
            this._showDefaultContextMenu = !!value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "focus", {
        get: function () {
            return this._focus;
        },
        set: function (newFocus) {
            this._focus = newFocus;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "colorCorrection", {
        get: function () {
            return this._colorCorrection;
        },
        set: function (value) {
            //this._colorCorrection = axCoerceString(value);
            release || notImplemented("public flash.display.Stage::set colorCorrection");
            return;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "colorCorrectionSupport", {
        get: function () {
            return this._colorCorrectionSupport;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "stageFocusRect", {
        get: function () {
            return this._stageFocusRect;
        },
        set: function (on) {
            this._stageFocusRect = !!on;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "quality", {
        get: function () {
            return this._quality.toUpperCase(); // Return value is always uppercase
        },
        set: function (value) {
            // TODO: The *linear versions return just *, stripping the "linear" part
            // Value is compared case-insensitively, and has default handling, so '' is ok.
            value = (axCoerceString(value) || '').toLowerCase();
            if (StageQuality.toNumber(value) < 0) {
                value = StageQuality.HIGH;
            }
            this._quality = value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "displayState", {
        get: function () {
            return this._displayState;
        },
        set: function (value) {
            value = axCoerceString(value);
            // TODO: This should only be allowed if the embedding page allows full screen mode.
            if (StageDisplayState.toNumber(value) < 0) {
                value = StageDisplayState.NORMAL;
            }
            if (this._displayState !== value) {
                this._isDirty = true;
                this._displayState = value;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "fullScreenSourceRect", {
        get: function () {
            return this._fullScreenSourceRect;
        },
        set: function (value) {
            //this._fullScreenSourceRect = value;
            release || notImplemented("public flash.display.Stage::set fullScreenSourceRect");
            return;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "mouseLock", {
        get: function () {
            return this._mouseLock;
        },
        set: function (value) {
            release || somewhatImplemented("public flash.display.Stage::set mouseLock");
            this._mouseLock = !!value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "stageVideos", {
        get: function () {
            release || somewhatImplemented("public flash.display.Stage::get stageVideos");
            return this._stageVideos;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "stage3Ds", {
        get: function () {
            release || somewhatImplemented("public flash.display.Stage::get stage3Ds");
            return this._stage3Ds;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "color", {
        get: function () {
            return this._colorARGB;
        },
        set: function (rgb /*uint*/) {
            var newValue = rgb | 0xff000000;
            // Flash player forces the alpha channel to 0xff.
            if (this._colorARGB !== newValue) {
                this._isDirty = true;
                this._colorARGB = newValue;
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "alpha", {
        get: function () {
            return this._colorTransform.alphaMultiplier;
        },
        set: function (alpha) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "fullScreenWidth", {
        get: function () {
            return this._fullScreenWidth;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "fullScreenHeight", {
        get: function () {
            return this._fullScreenHeight;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "wmodeGPU", {
        get: function () {
            return this._wmodeGPU;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "softKeyboardRect", {
        get: function () {
            return this._softKeyboardRect;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "allowsFullScreen", {
        get: function () {
            return this._allowsFullScreen;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "allowsFullScreenInteractive", {
        get: function () {
            return this._allowsFullScreenInteractive;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "contentsScaleFactor", {
        get: function () {
            return this._contentsScaleFactor;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "displayContextInfo", {
        get: function () {
            return this._displayContextInfo;
        },
        enumerable: true,
        configurable: true
    });
    Stage.prototype.removeChildAt = function (index) {
        this.requireOwnerPermissions();
        return _super.prototype.removeChildAt.call(this, index);
    };
    Stage.prototype.swapChildrenAt = function (index1 /*int*/, index2 /*int*/) {
        this.requireOwnerPermissions();
        _super.prototype.swapChildrenAt.call(this, index1, index2);
    };
    Object.defineProperty(Stage.prototype, "width", {
        get: function () {
            this.requireOwnerPermissions();
            return this._getWidth();
        },
        set: function (value) {
            this.requireOwnerPermissions();
            this._setWidth(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "height", {
        get: function () {
            this.requireOwnerPermissions();
            return this._getHeight();
        },
        set: function (value) {
            this.requireOwnerPermissions();
            this._setHeight(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "mouseChildren", {
        get: function () {
            this.requireOwnerPermissions();
            return this._getMouseChildren();
        },
        set: function (value) {
            this.requireOwnerPermissions();
            this._setMouseChildren(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "numChildren", {
        get: function () {
            this.requireOwnerPermissions();
            return this._getNumChildren();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "tabChildren", {
        get: function () {
            this.requireOwnerPermissions();
            return this._getTabChildren();
        },
        set: function (value) {
            this.requireOwnerPermissions();
            this._setTabChildren(value);
        },
        enumerable: true,
        configurable: true
    });
    Stage.prototype.addChild = function (child) {
        this.requireOwnerPermissions();
        return _super.prototype.addChild.call(this, child);
    };
    Stage.prototype.addChildAt = function (child, index) {
        this.requireOwnerPermissions();
        return _super.prototype.addChildAt.call(this, child, index);
    };
    Stage.prototype.setChildIndex = function (child, index) {
        this.requireOwnerPermissions();
        _super.prototype.setChildIndex.call(this, child, index);
    };
    Stage.prototype.addEventListener = function (type, listener, useCapture, priority, useWeakReference) {
        this.requireOwnerPermissions();
        _super.prototype.addEventListener.call(this, type, listener, useCapture, priority, useWeakReference);
    };
    Stage.prototype.hasEventListener = function (type) {
        this.requireOwnerPermissions();
        return _super.prototype.hasEventListener.call(this, type);
    };
    Stage.prototype.willTrigger = function (type) {
        this.requireOwnerPermissions();
        return _super.prototype.willTrigger.call(this, type);
    };
    Stage.prototype.dispatchEvent = function (event) {
        this.requireOwnerPermissions();
        return _super.prototype.dispatchEvent.call(this, event);
    };
    Stage.prototype.invalidate = function () {
        this._invalidated = true;
    };
    Stage.prototype.isFocusInaccessible = function () {
        release || notImplemented("public flash.display.Stage::isFocusInaccessible");
        return;
    };
    Stage.prototype.requireOwnerPermissions = function () {
        // TODO: implement requireOwnerPermissions
    };
    Stage.prototype.render = function () {
        if (!this._invalidated) {
            return;
        }
        DisplayObject.axClass._broadcastFrameEvent(Event.RENDER);
        this._invalidated = false;
    };
    Object.defineProperty(Stage.prototype, "name", {
        get: function () {
            return this._name;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "mask", {
        get: function () {
            return this._mask;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "visible", {
        get: function () {
            return this._hasFlags(1 /* Visible */);
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "x", {
        get: function () {
            return this._getX();
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "y", {
        get: function () {
            return this._getY();
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "z", {
        get: function () {
            return this._z;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "scaleX", {
        get: function () {
            return Math.abs(this._scaleX);
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "scaleY", {
        get: function () {
            return this._scaleY;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "scaleZ", {
        get: function () {
            return this._scaleZ;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "rotation", {
        get: function () {
            return this._rotation;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "rotationX", {
        get: function () {
            return this._rotationX;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "rotationY", {
        get: function () {
            return this._rotationX;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "rotationZ", {
        get: function () {
            return this._rotationX;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "cacheAsBitmap", {
        get: function () {
            return this._getCacheAsBitmap();
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "opaqueBackground", {
        get: function () {
            return this._opaqueBackground;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "scrollRect", {
        get: function () {
            return this._getScrollRect();
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "filters", {
        get: function () {
            return this._getFilters();
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "blendMode", {
        get: function () {
            return this._blendMode;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "transform", {
        get: function () {
            return this._getTransform();
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "accessibilityProperties", {
        get: function () {
            return this._accessibilityProperties;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "scale9Grid", {
        get: function () {
            return this._getScale9Grid();
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "tabEnabled", {
        get: function () {
            return this._tabEnabled;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "tabIndex", {
        get: function () {
            return this._tabIndex;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "focusRect", {
        get: function () {
            return this._focusRect;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "mouseEnabled", {
        get: function () {
            return this._mouseEnabled;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "accessibilityImplementation", {
        get: function () {
            return this._accessibilityImplementation;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "textSnapshot", {
        get: function () {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
            return null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Stage.prototype, "contextMenu", {
        get: function () {
            return this._contextMenu;
        },
        set: function (value) {
            this.sec.throwError('IllegalOperationError', Errors.InvalidStageMethodError);
        },
        enumerable: true,
        configurable: true
    });
    Stage.classInitializer = null;
    Stage.classSymbols = null; // [];
    Stage.instanceSymbols = null;
    return Stage;
}(DisplayObjectContainer));
export { Stage };
