import { __extends } from "tslib";
import { error, somewhatImplemented, release, assert } from "../../base/utilities/Debug";
import { EventDispatcher } from "../events/EventDispatcher";
import { SoundMixer } from "./SoundMixer";
import { SoundTransform } from "./SoundTransform";
import { isNullOrUndefined } from "../../base/utilities";
import { Event } from "../events/Event";
function createAudioChannel(sampleRate, channels) {
    if (WebAudioChannel.isSupported) {
        return new WebAudioChannel(sampleRate, channels);
    }
    else {
        error('PCM data playback is not supported by the browser');
    }
}
var AudioResampler = /** @class */ (function () {
    function AudioResampler(sourceRate, targetRate) {
        this._sourceRate = sourceRate;
        this._targetRate = targetRate;
        this._tail = [];
        this._sourceOffset = 0;
    }
    AudioResampler.prototype.getData = function (channelsData, count) {
        var k = this._sourceRate / this._targetRate;
        var offset = this._sourceOffset;
        var needed = Math.ceil((count - 1) * k + offset) + 1;
        var sourceData = [];
        for (var channel = 0; channel < channelsData.length; channel++) {
            sourceData.push(new Float32Array(needed));
        }
        var e = { data: sourceData, count: needed };
        this.ondatarequested(e);
        for (var channel = 0; channel < channelsData.length; channel++) {
            var data = channelsData[channel];
            var source = sourceData[channel];
            for (var j = 0; j < count; j++) {
                var i = j * k + offset;
                var i1 = i | 0, i2 = Math.ceil(i) | 0;
                var source_i1 = i1 < 0 ? this._tail[channel] : source[i1];
                if (i1 === i2) {
                    data[j] = source_i1;
                }
                else {
                    var alpha = i - i1;
                    data[j] = source_i1 * (1 - alpha) + source[i2] * alpha;
                }
            }
            this._tail[channel] = source[needed - 1];
        }
        this._sourceOffset = ((count - 1) * k + offset) - (needed - 1);
    };
    return AudioResampler;
}());
var WebAudioChannel = /** @class */ (function () {
    function WebAudioChannel(sampleRate, channels) {
        var context = WebAudioChannel._cachedContext;
        if (!context) {
            context = new AudioContext();
            WebAudioChannel._cachedContext = context;
        }
        this._context = context;
        this._contextSampleRate = context.sampleRate || 44100;
        this._channels = channels;
        this._sampleRate = sampleRate;
        if (this._contextSampleRate !== sampleRate) {
            this._resampler = new AudioResampler(sampleRate, this._contextSampleRate);
            this._resampler.ondatarequested = function (e) {
                this.requestData(e.data, e.count);
            }.bind(this);
        }
    }
    WebAudioChannel.prototype.setVolume = function (value) {
        // TODO set volume on this._source via gain node
    };
    WebAudioChannel.prototype.start = function () {
        var source = this._context.createScriptProcessor(2048, 0, this._channels);
        var self = this;
        source.onaudioprocess = function (e) {
            var channelsData = [];
            for (var i = 0; i < self._channels; i++) {
                channelsData.push(e.outputBuffer.getChannelData(i));
            }
            var count = channelsData[0].length;
            if (self._resampler) {
                self._resampler.getData(channelsData, count);
            }
            else {
                self.requestData(channelsData, count);
            }
        };
        source.connect(this._context.destination);
        this._source = source;
    };
    WebAudioChannel.prototype.stop = function () {
        this._source.disconnect(this._context.destination);
    };
    WebAudioChannel.prototype.requestData = function (channelsData, count) {
        var channels = this._channels;
        var buffer = new Float32Array(count * channels);
        var e = { data: buffer, count: buffer.length };
        this.ondatarequested(e);
        for (var j = 0, p = 0; j < count; j++) {
            for (var i = 0; i < channels; i++) {
                channelsData[i][j] = buffer[p++];
            }
        }
    };
    WebAudioChannel.isSupported = function () {
        return typeof AudioContext !== 'undefined';
    };
    return WebAudioChannel;
}());
var SoundChannel = /** @class */ (function (_super) {
    __extends(SoundChannel, _super);
    function SoundChannel() {
        var _this = _super.call(this) || this;
        _this._element = null;
        _this._position = 0;
        _this._leftPeak = 0;
        _this._rightPeak = 0;
        _this._pcmData = null;
        _this._soundTransform = new SoundTransform();
        _this._playing = false;
        _this._element = null;
        return _this;
    }
    SoundChannel.initializeFromAudioElement = function (sec, element) {
        var channel = new SoundChannel();
        channel._element = element;
        SoundMixer._registerSoundSource(channel);
        return channel;
    };
    Object.defineProperty(SoundChannel.prototype, "position", {
        get: function () {
            return this._position;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SoundChannel.prototype, "soundTransform", {
        get: function () {
            return this._soundTransform;
        },
        set: function (sndTransform) {
            release || somewhatImplemented("public flash.media.SoundChannel::set soundTransform");
            this._soundTransform = isNullOrUndefined(sndTransform) ?
                new SoundTransform() :
                sndTransform;
            SoundMixer._updateSoundSource(this);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SoundChannel.prototype, "leftPeak", {
        get: function () {
            return this._leftPeak;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SoundChannel.prototype, "rightPeak", {
        get: function () {
            return this._rightPeak;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SoundChannel.prototype, "playing", {
        get: function () {
            return this._playing;
        },
        enumerable: true,
        configurable: true
    });
    SoundChannel.prototype.stop = function () {
        if (this._element) {
            SoundMixer._unregisterSoundSource(this);
            this._element.loop = false;
            this._element.pause();
            this._element.removeAttribute('src');
            this._playing = false;
        }
        if (this._audioChannel) {
            SoundMixer._unregisterSoundSource(this);
            this._audioChannel.stop();
            this._playing = false;
        }
    };
    SoundChannel.prototype._playSoundDataViaAudio = function (soundData, startTime, loops) {
        if (!soundData.mimeType) {
            return;
        }
        SoundMixer._registerSoundSource(this);
        this._position = startTime;
        var self = this;
        var lastCurrentTime = 0;
        var element = document.createElement('audio');
        if (!element.canPlayType(soundData.mimeType)) {
            console.error('ERROR: \"' + soundData.mimeType + '\" ' +
                'type playback is not supported by the browser');
            return;
        }
        element.preload = 'metadata'; // for mobile devices
        element.loop = loops > 0; // starts loop played if at least one is specified
        var blob = new Blob([soundData.data], { type: soundData.mimeType });
        element.src = URL.createObjectURL(blob);
        element.addEventListener("loadeddata", function loaded() {
            element.currentTime = startTime / 1000;
            element.play();
        });
        element.addEventListener("timeupdate", function timeupdate() {
            var currentTime = element.currentTime;
            if (loops && lastCurrentTime > currentTime) {
                --loops;
                if (!loops) { // checks if we need to stop looping
                    element.loop = false;
                }
                if (currentTime < startTime / 1000) {
                    element.currentTime = startTime / 1000;
                }
            }
            self._position = (lastCurrentTime = currentTime) * 1000;
        });
        element.addEventListener("ended", function ended() {
            SoundMixer._unregisterSoundSource(self);
            self._element = null;
            self._playing = false;
            self.dispatchEvent(new Event("soundComplete", false, false));
        });
        this._element = element;
        this._playing = true;
        SoundMixer._updateSoundSource(this);
    };
    SoundChannel.prototype._playSoundDataViaChannel = function (soundData, startTime, loops) {
        release || assert(soundData.pcm, 'no pcm data found');
        SoundMixer._registerSoundSource(this);
        var self = this;
        var startPosition = Math.round(startTime / 1000 * soundData.sampleRate) *
            soundData.channels;
        var position = startPosition;
        this._position = startTime;
        this._audioChannel = createAudioChannel(soundData.sampleRate, soundData.channels);
        this._audioChannel.ondatarequested = function (e) {
            var end = soundData.end;
            if (position >= end && soundData.completed) {
                // end of buffer
                SoundMixer._unregisterSoundSource(this);
                self._audioChannel.stop();
                self._playing = false;
                self.dispatchEvent(new Event("soundComplete", false, false));
                return;
            }
            var left = e.count;
            var data = e.data;
            var source = soundData.pcm;
            do {
                var count = Math.min(end - position, left);
                for (var j = 0; j < count; j++) {
                    data[j] = source[position++];
                }
                left -= count;
                if (position >= end) {
                    if (!loops) {
                        break;
                    }
                    loops--;
                    position = startPosition;
                }
            } while (left > 0);
            self._position = position / soundData.sampleRate / soundData.channels * 1000;
        };
        this._audioChannel.start();
        this._playing = true;
        SoundMixer._updateSoundSource(this);
    };
    SoundChannel.prototype.stopSound = function () {
        this.stop();
    };
    SoundChannel.prototype.updateSoundLevels = function (volume) {
        if (this._element) {
            this._element.volume = volume <= 0 ? 0 : volume >= 1.0 ? 1.0 : volume;
        }
        if (this._audioChannel) {
            this._audioChannel.setVolume(volume);
        }
    };
    SoundChannel.classInitializer = null;
    return SoundChannel;
}(EventDispatcher));
export { SoundChannel };
